/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz> (original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef WIN32
#include <assert.h>
#include <setjmp.h>
#endif

static void ComputeDifferentials(double y, const double *x, double *dx);
static void EvaluateSimple(double y, double *x);
static void EvaluateSimpleOnce(double *x);
static void read_input_command();
static void init_helper();
static void cmd_start(void *data, PRUint32 length);
static void cmd_suspend(void *data, PRUint32 length);
static void cmd_stop(void *data, PRUint32 length);
static void integration_done();
static void integration_failed(const char* error);

/* + 1 just to avoid 0 length array. */
static double constants[NC + 1];

static enum
{
  SUSPENDED,
  RUNNING,
  DYING
} helper_state;

/* Milliseconds between reads in suspend mode... */
#define SUSPEND_MILLIS 500
/* Cycles before poll in running mode... */
#define CYCLES_BEFORE_POLL 100

/* Packets which can be queued before updating the client... */
#define PACKET_QUEUE_SIZE 20

/* The upper and lower bounds on the order count(1, 3, 5, or 7) */
#define MINIMAL_ORDER 3
#define MAXIMAL_ORDER 7
/* How many order at the moment? */
PRInt32 current_order;
PRUint32 n_current_order;

#ifndef WIN32
static jmp_buf sigrecover;
#endif

/* The maximal number of Newton iteractions per step. */
#define MAXIMAL_NEWTON 7
/* Dimension of index 1 variables(order of system for ODEs) */
#define NIND1 ND
#define NIND2 0
#define NIND3 0

#define t311 0.09123239487089294279155
#define t312 -0.141255295020954208428
#define t313 -0.03002919410514742449186
#define t321 0.2417179327071070189575
#define t322 0.204129352293799931996
#define t323 0.3829421127572619377954
#define t331 0.9660481826150929361906
#define ti311 4.325579890063155351024
#define ti312 0.3391992518158098695428
#define ti313 0.5417705399358748711865
#define ti321 -4.178718591551904727346
#define ti322 -0.3276828207610623870825
#define ti323 0.4766235545005504519601
#define ti331 -0.5028726349457868759512
#define ti332 2.571926949855605429187
#define ti333 -0.5960392048282249249688
#define t511 -0.01251758622050104589014
#define t512 -0.01024204781790882707009
#define t513 0.04767387729029572386318
#define t514 -0.01147851525522951470794
#define t515 -0.01401985889287541028108
#define t521 -0.001491670151895382429004
#define t522 0.05017286451737105816299
#define t523 -0.09433181918161143698066
#define t524 -0.007668830749180162885157
#define t525 0.02470857842651852681253
#define t531 -0.07298187638808714862266
#define t532 -0.2305395340434179467214
#define t533 0.1027030453801258997922
#define t534 0.01939846399882895091122
#define t535 0.08180035370375117083639
#define t541 -0.3800914400035681041264
#define t542 0.3778939022488612495439
#define t543 0.4667441303324943592896
#define t544 0.4076011712801990666217
#define t545 0.1996824278868025259365
#define t551 -0.9219789736812104884883
#define ti511 -30.04156772154440162771
#define ti512 -13.86510785627141316518
#define ti513 -3.480002774795185561828
#define ti514 1.032008797825263422771
#define ti515 -0.8043030450739899174753
#define ti521 5.344186437834911598895
#define ti522 4.593615567759161004454
#define ti523 -3.036360323459424298646
#define ti524 1.05066019023145886386
#define ti525 -0.2727786118642962705386
#define ti531 3.748059807439804860051
#define ti532 -3.984965736343884667252
#define ti533 -1.044415641608018792942
#define ti534 1.184098568137948487231
#define ti535 -0.4499177701567803688988
#define ti541 -33.04188021351900000806
#define ti542 -17.37695347906356701945
#define ti543 -0.1721290632540055611515
#define ti544 -0.09916977798254264258817
#define ti545 0.5312281158383066671849
#define ti551 -8.6114439798752919777
#define ti552 9.699991409528808231336
#define ti553 1.914728639696874284851
#define ti554 2.418692006084940026427
#define ti555 -1.047463487935337418694
#define t711 -0.002153754627310526422828
#define t712  .02156755135132077338691
#define t713  .008783567925144144407326
#define t714  -.004055161452331023898198
#define t715  .004427232753268285479678
#define t716  -.001238646187952874056377
#define t717  -.002760617480543852499548
#define t721  .001600025077880428526831
#define t722  -.03813164813441154669442
#define t723  -.02152556059400687552385
#define t724  .008415568276559589237177
#define t725  -.004031949570224549492304
#define t726  -6.666635339396338181761e-5
#define t727  .003185474825166209848748
#define t731  -.00405910730194768309165
#define t732  .05739650893938171539757
#define t733  .05885052920842679105612
#define t734  -.008560431061603432060177
#define t735  -.006923212665023908924141
#define t736  -.002352180982943338340535
#define t737  4.169077725297562691409e-4
#define t741  -.01575048807937684420346
#define t742  -.03821469359696835048464
#define t743  -.1657368112729438512412
#define t744  -.03737124230238445741907
#define t745  .008239007298507719404499
#define t746  .003115071152346175252726
#define t747  .02511660491343882192836
#define t751  -.1129776610242208076086
#define t752  -.2491742124652636863308
#define t753  .2735633057986623212132
#define t754  .005366761379181770094279
#define t755  .1932111161012620144312
#define t756  .1017177324817151468081
#define t757  .09504502035604622821039
#define t761  -.4583810431839315010281
#define t762  .5315846490836284292051
#define t763  .4863228366175728940567
#define t764  .5265742264584492629141
#define t765  .2755343949896258141929
#define t766  .5217519452747652852946
#define t767  .1280719446355438944141
#define t771  -.8813915783538183763135
#define ti711  -258.1319263199822292761
#define ti712  -189.073763081398508952
#define ti713  -49.08731481793013119445
#define ti714  -4.110647469661428418112
#define ti715  -4.053447889315563304175
#define ti716  3.112755366607346076554
#define ti717  -1.646774913558444650169
#define ti721  -3.007390169451292131731
#define ti722  -11.01586607876577132911
#define ti723  1.487799456131656281486
#define ti724  2.130388159559282459432
#define ti725  -1.816141086817565624822
#define ti726  1.134325587895161100083
#define ti727  -.414699045943303531993
#define ti731  -8.441963188321084681757
#define ti732  -.6505252740575150028169
#define ti733  6.940670730369876478804
#define ti734  -3.205047525597898431565
#define ti735  1.071280943546478589783
#define ti736  -.354850749121622187973
#define ti737  .09198549132786554154409
#define ti741  74.67833223502269977153
#define ti742  87.40858897990081640204
#define ti743  4.024158737379997877014
#define ti744  -3.714806315158364186639
#define ti745  -3.430093985982317350741
#define ti746  2.696604809765312378853
#define ti747  -.9386927436075461933568
#define ti751  58.35652885190657724237
#define ti752  -10.06877395780018096325
#define ti753  -30.36638884256667120811
#define ti754  -1.020020865184865985027
#define ti755  -.1124175003784249621267
#define ti756  1.8906408310003776228
#define ti757  -.9716486393831482282172
#define ti761  -299.1862480282520966786
#define ti762  -243.0407453687447911819
#define ti763  -48.77710407803786921219
#define ti764  -2.03867190574193440528
#define ti765  1.673560239861084944268
#define ti766  -1.087374032057106164456
#define ti767  .9019382492960993738427
#define ti771  -93.07650289743530591157
#define ti772  23.88163105628114427703
#define ti773  39.2788807308138438271
#define ti774  14.38891568549108006988
#define ti775  -3.510438399399361221087
#define ti776  4.863284885566180701215
#define ti777  -2.2464827295912399164

/* No banding supported.
#define M1 0
#define M2 ND
#define NM1 (ND-M1)
 */
/* The rounding unit... */
#define ROUNDING_UNIT 1e-310
/* The safety factor in the step-size prediction... */
#define SAFETY_FACTOR 0.9
/* Cost factor for Jacobian recomputation... */
#define THET 0.001
/*
 * The ratio used to determine when to change the
 * step size(which can be costly).
 */
#define STEP_HOLDRATIO_MIN 1.0
#define STEP_HOLDRATIO_MAX 1.2
static double maximal_step_size, step_size, step_size_old, step_size_opt,
              step_size_new, step_size_acc;
/* L/R factors for stepsize selection...*/
#define L_FACTOR 5.0
#define R_FACTOR 0.125
#define INCREASE_ORDER_CONTRACTIVITY 0.002
#define REDUCE_ORDER_CONTRACTIVITY 0.8
#define INCREASE_ORDER_HOLDRATIO_MIN 0.8
#define INCREASE_ORDER_HOLDRATIO_MAX 1.2
#define INITIAL_STEP_SIZE 1.0E-6

/* Some constants */
/* sqrt(6) */
#define SQRT6 2.44948974278318
#define C31 ((4.0-SQRT6)/10.0)
#define C32 ((SQRT6 + 4.0)/10.0)
#define C31M1 (C31-1.0)
#define C32M1 (C32-1.0)
#define C31MC2 (C31-C32)
/* ST9 = 9^(1/3) */
#define ST9 2.08008382305190
#define DD1 (-(SQRT6 * 7.0 + 13.0) / 3.0)
#define DD2 ((SQRT6 * 7.0 - 13.0) / 3.0)
#define DD3 (-1.0/3.0)
#define N2 (ND<<1)
#define N3 (ND*3)
#define N4 (ND<<2)
#define N5 (ND*5)
#define N6 (ND*6)

static PRUint32 pointcount, maxpoint;
static double x, xend, xold, faccon, xstart;

#define STATIC_ARRAYS
#ifdef STATIC_ARRAYS
static double zz[MAXIMAL_ORDER * ND];
static double all_yvalues[NS+ND];
static double y_0[ND];
static double scal[ND];
static double ff[MAXIMAL_ORDER * ND];
static double jac[ND][ND];
static double e1[ND][ND];
static double e2r[(MAXIMAL_ORDER - 1) / 2][ND][ND];
static double e2i[(MAXIMAL_ORDER - 1) / 2][ND][ND];
static double cont[(MAXIMAL_ORDER + 1) * (ND + NS) + ND + NS];
static double expo, expmns, quott, hhfac, err, dynold, thqold,
              erracc;
#else
double* zz;
double* all_yvalues;
double* y_0;
double* scal;
double* ff;
typedef double (*double_array_ND_t)[ND];
typedef double (*double_array_ND_ND_t)[ND][ND];
double_array_ND_t jac;
double_array_ND_t e1;
double_array_ND_ND_t e2r;
double_array_ND_ND_t e2i;
double* cont;
static double expo, expmns, quott, hhfac, err, dynold, thqold,
              erracc;

void
allocate_arrays()
{
  zz = malloc(MAXIMAL_ORDER * ND * sizeof(double));
  all_yvalues = malloc((NS+ND) * sizeof(double));
  y_0 = malloc(ND * sizeof(double));
  scal = malloc(ND * sizeof(double));
  ff = malloc(MAXIMAL_ORDER * ND * sizeof(double));
  jac = (double_array_ND_t)malloc(sizeof(double[ND]) * ND);
  e1 = (double_array_ND_t)malloc(sizeof(double[ND]) * ND);
  e2r = (double_array_ND_ND_t)malloc(sizeof(double[ND][ND]) * (MAXIMAL_ORDER-1)/2);
  e2i = (double_array_ND_ND_t)malloc(sizeof(double[ND][ND]) * (MAXIMAL_ORDER-1)/2);
  cont = malloc(sizeof(double) * ((MAXIMAL_ORDER + 1) * (ND + NS) + ND + NS));
}
#endif

static PRUint32 num_queued_packets = 0;

/* This is just so we can use the same code in functions in which y is not
   available.
*/
static double y = 0.0;

static void add_points(double xval, double* values);

static PRUint32 ip1[ND];
static PRUint32 ip2[ND * (MAXIMAL_ORDER - 1) / 2];
/* It should be size ND - 1, but we avoid a zero length array. */
static PRUint32 iphes[ND /* - 1 */ ];
static PRBool unexp, unexn, change, reject, first, last, caljac, calhes,
              skipjac, got_nan;
static PRInt32 posneg;
static PRUint32 ikeep, ichan;
static double theta, thetat, fac1, quot;
static PRUint32 lrc, nstep;
static PRUint32 newton_count, naccpt;
static double weights[8], u1, dd[7], alpha[7], beta[7], alphan[7],
              betan[7];

static PRPollDesc polldesc;
static PRFileDesc *input, *output;
static double abs_tolerance, abs_tolerance_1,
              rel_tolerance, rel_tolerance_1, resolution;

static char* active_cmeta_id = "debug mode off";
static int disable_debug = 1;

static struct
{
  PRUint32 length;
  char *data;
  PRBool needsfree;
} QueuedPackets[PACKET_QUEUE_SIZE];

static void
setup_io()
{
  PRNetAddr addr;
  PRSocketOptionData sod;

#ifdef DEBUG_MODE
  printf("setup_io called.\n");
  fflush(stdout);
#endif

  input = PR_NewTCPSocket();
  output = input;

  if (input == NULL)
  {
#ifdef DEBUG_MODE
    printf("PR_NewTCPSocket() failed!\n");
    fflush(stdout);
#endif
    PR_ProcessExit(-1);
  }
  addr.inet.family = PR_AF_INET;
  addr.inet.port = PR_ntohs(CONNECT_PORT);
  addr.inet.ip = PR_ntohl(PR_INADDR_LOOPBACK);

  sod.option = PR_SockOpt_Nonblocking;
  sod.value.non_blocking = PR_FALSE;
  PR_SetSocketOption(input, &sod);

  if (PR_Connect(input, &addr, PR_INTERVAL_NO_TIMEOUT) != PR_SUCCESS)
  {
#define ERRMSG(P) "Connect to port " #P " failed.\n"
    printf(ERRMSG(CONNECT_PORT));
    fflush(stdout);
    PR_ProcessExit(-2);
  }

  /* Setup the input poll descriptor... */
  polldesc.fd = input;
  polldesc.in_flags = PR_POLL_READ;
  polldesc.out_flags = 0;
#ifdef DEBUG_MODE
  printf("I/O setup OK\n");
  fflush(stdout);
#endif
}

static double
take_numeric_derivative(double y, double* x, double (*f)(double y, double* x),
                        PRUint32 wrt)
{
  double saved_x, value0, value1, value2, delta, slope1, slope2, ratio;
  saved_x = x[wrt];
  /*
   * Since we have only 52 bits of mantissa, we need delta to flip only the
   * last couple of bits.
   */
  delta = saved_x * 1E-10;
  value0 = f(y, x);
  x[wrt] -= delta;
  value1 = f(y, x);
  x[wrt] = saved_x + delta;
  value2 = f(y, x);
  x[wrt] = delta;
  /* We take two numeric derivatives, so we can compare them. This stops us
   * from getting trapped at around discontinuities(which otherwise produce
   * very step trapping slopes).
   */
  slope1 = (value0 - value1) / delta;
  slope2 = (value2 - value0) / delta;

  /* If one slope is zero, return the other... */
  if (slope1 == 0.0 || !isfinite(slope1))
    return slope2;
  if (slope2 == 0.0 || !isfinite(slope2))
    return slope1;

  ratio = slope1 / slope2;
  /* If the slopes are similar, return the average... */
  if (ratio > 0.5 && ratio < 2)
    return (slope1 + slope2) / 2.0;
  /* Otherwise, return the least steep of the two... */
  if (abs_d(slope1) < abs_d(slope2))
    return slope1;
  return slope2;
}

static double
random_double_logUniform()
{
  union
  {
    double asDouble;
    PRUint32 asIntegers[2];
  } X;

  PRUint32 spareRand;

  do
  {
    spareRand = rand();
    X.asIntegers[0] = rand() | ((spareRand & 0x1) << 31);
    X.asIntegers[1] = rand() | ((spareRand & 0x2) << 30);
  }
  while (!finite(X.asDouble));

  return X.asDouble;
}

#define NR_RANDOM_STARTS 100
#define NR_MAX_STEPS 1000
#define NR_MAX_STEPS_INITIAL 10

static void
solve_by_minimisation(double y, double* x, double (*f)(double y, double* x),
                      PRUint32 storeWhere)
{
  double best_X, best_fX = INFINITY;
  double current_X, current_fX, current_dfX_dX;
  PRUint32 steps, maxsteps;
  PRUint32 i;

  current_X = x[storeWhere];

  SKIP_TRAP_INCREMENT;
  /* We use a 100 random start Newton-Raphson algorithm... */
  for (i = 0; i < NR_RANDOM_STARTS; i++)
  {
    /* Choose a random X as a starting point, except for the first run. */
    if (i > 0 || !isfinite(current_X))
      current_X = random_double_logUniform();

    maxsteps = NR_MAX_STEPS_INITIAL;
    for (steps = 0; steps < maxsteps; steps++)
    {
      x[storeWhere] = current_X;
      current_fX = f(y, x);
      if (!finite(current_fX))
        break;

      if (best_fX > current_fX)
      {
        /* We can go past NR_MAX_STEPS_INITIAL steps up to NR_MAX_STEPS steps,
         * but only if we keep improving on our previous answer.
         */
        maxsteps += 2;
        if (maxsteps > NR_MAX_STEPS)
          maxsteps = NR_MAX_STEPS;
        best_fX = current_fX;
        best_X = current_X;
        /* This is quite far from 0(relatively speaking for a double) to avoid
         * numerical instability issues when the minimum doesn't exist but the
         * limit at a point from a particular direction would otherwise be the
         * minimum.
         */
        if (best_fX <= 1E-250)
          break;
      }

      current_dfX_dX = take_numeric_derivative(y, x, f, storeWhere);
      /* If it is completely flat, or infinite, we are done with this round. */
      if (!isfinite(current_dfX_dX) || current_dfX_dX == 0.0)
        break;

      current_X -= current_fX / current_dfX_dX;
      if (!finite(current_X))
        break;
    }
    if (best_fX <= 1E-250)
      break;
  }

  SKIP_TRAP_DECREMENT;

  x[storeWhere] = best_X;
}

static void
flush_queue()
{
  PRUint32 i;
  for (i = 0; i < num_queued_packets; i++)
  {
    PR_Write(output, QueuedPackets[i].data, QueuedPackets[i].length);
    if (QueuedPackets[i].needsfree)
      PR_Free(QueuedPackets[i].data);
  }
  num_queued_packets = 0;
}

static void
queue_packet(char *data, PRUint32 length, PRBool mustfree)
{
  QueuedPackets[num_queued_packets].length = length;
  QueuedPackets[num_queued_packets].data = data;
  QueuedPackets[num_queued_packets].needsfree = mustfree;
  if (++num_queued_packets == PACKET_QUEUE_SIZE)
    flush_queue();
}

static void
send_packet(char *data, PRUint32 length, PRBool mustfree)
{
  flush_queue();
  PR_Write(output, data, length);
  if (mustfree)
    PR_Free(data);
}

static void
do_coercv(PRUint32 order)
{
  weights[0] = 0.0;
  weights[order] = 1.0;
  switch (order)
  {
  case 1:
    weights[1] = 1.0;
    u1 = 1.0;
    dd[0] = -1.0;
    break;
  case 3:
    weights[1] = C31;
    weights[2] = C32;
    /* The 1/ thing seems to make a big difference(numerical stability). */
    /* u1 = 1 / ((6.0 + ST9 * (ST9-1.0)) / 30.0); */
    /* The FORTRAN code gets a slightly different number that makes a big
     * difference, so to enable comparisons, we need to use the exact same
     * figure...
     */
    u1 = 0.27488882959567734;
    alpha[0] = 2.68108287362775;
    beta[0] = 3.05043019924741;
    break;
  case 5:
    weights[1] = .05710419611451768219312;
    weights[2] = .27684301363812382768;
    weights[3] = .5835904323689168200567;
    weights[4] = .8602401356562194478479;
    dd[0] = -27.78093394406463730479;
    dd[1] = 3.641478498049213152712;
    dd[2] = -1.252547721169118720491;
    dd[3] = .5920031671845428725662;
    dd[4] = -.2;
    u1 = 6.286704751729276645173;
    alpha[0] = 3.655694325463572258243;
    beta[0] = 6.543736899360077294021;
    alpha[1] = 5.70095329867178941917;
    beta[1] = 3.210265600308549888425;
    break;
  case 7:
    weights[1] = .02931642715978489197205;
    weights[2] = .14807859966848429185;
    weights[3] = .3369846902811542990971;
    weights[4] = .5586715187715501320814;
    weights[5] = .7692338620300545009169;
    weights[6] = .9269456713197411148519;
    dd[0] = -54.37443689412861451458;
    dd[1] = 7.000024004259186512041;
    dd[2] = -2.355661091987557192256;
    dd[3] = 1.132289066106134386384;
    dd[4] = -.6468913267673587118673;
    dd[5] = .3875333853753523774248;
    dd[6] = -.1428571428571428571429;
    u1 = 8.936832788405216337302;
    alpha[0] = 4.378693561506806002523;
    beta[0] = 10.16969328379501162732;
    alpha[1] = 7.141055219187640105775;
    beta[1] = 6.623045922639275970621;
    alpha[2] = 8.511834825102945723051;
    beta[2] = 3.281013624325058830036;
    break;
  }
}

static PRBool
is_there_input()
{
  return (PR_Poll(&polldesc, 1, 0) > 0);
}

static void
check_input()
{
  while (PR_TRUE)
  {
    /*
     * input is blocking, so we have to keep checking if there is a command
     * left...
     */
    if (PR_Poll(&polldesc, 1, 0) <= 0)
      return;
    read_input_command();
  }
}

static void
do_interpolation(double at_x, double* out_y)
{
  PRInt32 i, k;
  double s = (at_x - x) / step_size_old + 1.0;
  /* Just copy the simple variables, rather than interpolate... */
  for (i = 0; i < NS; i++)
    out_y[i] = all_yvalues[i];
  for (; i < ND + NS; i++)
  {
    out_y[i] = cont[i + current_order * (ND + NS)];
    for (k = current_order - 1; k >= 0; k--)
      out_y[i] = cont[i + k * (ND + NS)] +
                 (s - weights[current_order - k]) * out_y[i];
  }
  EvaluateSimple(at_x, out_y);
}

static void
HessianEliminateJacobian()
{
  PRUint32 i, j, m;
  double x, y;
  PRUint32 la;
  PRUint32 mm1, mp1;
  PRUint32 kp1;

  la = ND - 1;
  kp1 = 2;
  for (m = 2; m <= ND - 1; m++)
  {
    mm1 = m - 1;
    x = 0.0;
    i = m;
    for (j = m; j <= ND; j++)
    {
      if (abs_d(jac[j - 1][mm1 - 1]) > abs_d(x))
      {
        x = jac[j - 1][mm1 - 1];
        i = j;
      }
    }

    iphes[m - 1] = i;

    if (i != m)
    {
      /* interchange rows and columns of a */
      for (j = mm1; j <= ND; j++)
      {
        y = jac[i - 1][j - 1];
        jac[i - 1][j - 1] = jac[m - 1][j - 1];
        jac[m - 1][j - 1] = y;
      }

      for (j = 1; j <= ND; ++j)
      {
        y = jac[j - 1][i - 1];
        jac[j - 1][i - 1] = jac[j - 1][m - 1];
        jac[j - 1][m - 1] = y;
      }
    }

    if (x != 0.0)
    {
      mp1 = m + 1;
      for (i = mp1; i <= ND; i++)
      {
        y = jac[i - 1][mm1 - 1];
        if (y != 0.0)
        {
          y /= x;
          jac[i - 1][mm1 - 1] = y;
          for (j = m; j <= ND; j++)
            jac[i - 1][j - 1] -= y * jac[m - 1][j - 1];
          for (j = 1; j <= ND; j++)
            jac[j - 1][m - 1] += y * jac[j - 1][i - 1];
        }
      }
    }
  }
}

static void dech(
                 PRUint32* singular
                )
{
  PRUint32 j, k, m, kp1;
  double t;

  *singular = 0;
  ip1[ND - 1] = 1;
  if (ND != 1)
  {
    for (k = 1; k <= ND - 1 ; ++k)
    {
      kp1 = k + 1;
      m = k;
      if (abs_d(e1[k][k - 1]) > abs_d(e1[m - 1][k]))
        m = kp1;
      ip1[k - 1] = m;
      t = e1[m - 1][k - 1];
      if (m != k)
      {
        ip1[ND - 1] = -ip1[ND - 1];
        e1[m - 1][k - 1] = e1[k - 1][k - 1];
        e1[k - 1][k - 1] = t;
      }
      if (t == 0.0)
      {
        *singular = k;
        ip1[ND - 1] = 0;
        return;
      }
      t = 1.0 / t;
      e1[k][k - 1] = -e1[k][k - 1] * t;
      for (j = kp1; j <= ND; ++j)
      {
        t = e1[m - 1][j - 1];
        e1[m - 1][j - 1] = e1[k - 1][j - 1];
        e1[k - 1][j - 1] = t;
        if (t != 0.0)
        {
          e1[k][j - 1] += e1[k][k - 1] * t;
        }
      }
    }
  }
  if (e1[ND - 1][ND - 1] != 0.0)
    return;
  *singular = ND;
  ip1[ND - 1] = 0;
  return;
}

static void
dechc(PRUint32* singular, PRUint32 km1)
{
  PRUint32 i, j, k, m, na;
  double ti, tr;
  PRUint32 kp1;
  double den, prodi, prodr;
  PRUint32* xip2 = ip2 + km1 * ND;

  *singular = 0;
  xip2[ND - 1] = 1;

  for (k = 1; k <= ND - 1; k++)
  {
    kp1 = k + 1;
    m = k;
    na = imin(ND, kp1);
    for (i = kp1; i <= na; i++)
    {
      if (abs_d(e2r[km1][i - 1][k - 1]) + abs_d(e2i[km1][i - 1][k - 1]) >
          abs_d(e2r[km1][m - 1][k - 1]) + abs_d(e2i[km1][m - 1][k - 1]))
        m = i;
    }
    xip2[k - 1] = m;
    tr = e2r[km1][m - 1][k - 1];
    ti = e2i[km1][m - 1][k - 1];
    if (m != k)
    {
      xip2[ND - 1] = -xip2[ND - 1];
      e2r[km1][m - 1][k - 1] = e2r[km1][k - 1][k - 1];
      e2i[km1][m - 1][k - 1] = e2i[km1][k - 1][k - 1];
      e2r[km1][k - 1][k - 1] = tr;
      e2i[km1][k - 1][k - 1] = ti;
    }

    if (abs_d(tr) + abs_d(ti) == 0.0)
    {
      *singular = k;
      ip1[ND - 1] = 0;
      return;
    }

    den = tr * tr + ti * ti;
    tr /= den;
    ti = -ti / den;

    for (i = kp1; i <= na; i++)
    {
      prodr = e2r[km1][i - 1][k - 1] * tr - e2i[km1][i - 1][k - 1] * ti;
      prodi = e2i[km1][i - 1][k - 1] * tr + e2r[km1][i - 1][k - 1] * ti;
      e2r[km1][i - 1][k - 1] = -prodr;
      e2i[km1][i - 1][k - 1] = -prodi;
    }
    for (j = kp1; j <= ND; j++)
    {
      tr = e2r[km1][m - 1][j - 1];
      ti = e2i[km1][m - 1][j - 1];
      e2r[km1][m - 1][j - 1] = e2r[km1][k - 1][j - 1];
      e2i[km1][m - 1][j - 1] = e2i[km1][k - 1][j - 1];
      e2r[km1][k - 1][j - 1] = tr;
      e2i[km1][k - 1][j - 1] = ti;

      if (abs_d(tr) + abs_d(ti) == 0.0)
        continue;
      
      if (ti == 0.0)
      {
        for (i = kp1; i <= na; i++)
        {
          prodr = e2r[km1][i - 1][k - 1] * tr;
          prodi = e2i[km1][i - 1][k - 1] * tr;
          e2r[km1][i - 1][j - 1] += prodr;
          e2i[km1][i - 1][j - 1] += prodi;
        }
        continue;
      }
      
      if (tr == 0.0)
      {
        for (i = kp1; i <= na; i++)
        {
          prodr = -e2i[km1][i - 1][k - 1] * ti;
          prodi = e2r[km1][i - 1][k - 1] * ti;
          e2r[km1][i - 1][j - 1] += prodr;
          e2i[km1][i - 1][j - 1] += prodi;
        }
        continue;
      }
      for (i = kp1; i <= na; i++)
      {
        prodr = e2r[km1][i - 1][k - 1] * tr - e2i[km1][i - 1][k - 1] *  ti;
        prodi = e2i[km1][i - 1][k - 1] * tr + e2r[km1][i - 1][k - 1] * ti;
        e2r[km1][i - 1][j - 1] += prodr;
        e2i[km1][i - 1][j - 1] += prodi;
      }
    }
  }

  k = ND;
  if (abs_d(e2r[km1][ND - 1][ND - 1]) + abs_d(e2i[km1][ND - 1][ND - 1]) != 0.0)
    return;
  *singular = k;
  ip1[ND - 1] = 0;
  return;
}

static void
do_decomr(double fac1, PRUint32* singular)
{
  PRUint32 i, j;
  
  if (calhes)
    HessianEliminateJacobian();
  calhes = PR_FALSE;

  for (j = 1; j <= ND - 1; j++)
  {
    e1[j][j - 1] = -jac[j][j - 1];
  }
  for (j = 1; j <= ND; j++)
  {
    for (i = 1; i <= j; i++)
      e1[i - 1][j - 1] = -jac[i - 1][j - 1];
    e1[j - 1][j - 1] += fac1;
  }
  dech(singular);
}

/* Solve AX = B for X, with A, B ND*ND matrices, after decomposition.
 * Clobbers B
 */
static void
do_solve(double A[][ND], double* B, PRUint32* X)
{
  PRUint32 i, k, m, na;
  double t;
  PRUint32 kb, km1, kp1;

  for (k = 1; k <= ND - 1; k++)
  {
    kp1 = k + 1;
    m = X[k - 1];
    t = B[m - 1];
    B[m - 1] = B[k - 1];
    B[k - 1] = t;
    na = imin(ND, kp1);
    for (i = kp1; i <= na; i++)
      B[i - 1] += A[i - 1][k - 1] * t;
  }
  for (kb = 1; kb <= ND - 1; kb++)
  {
    km1 = ND - kb;
    B[km1] /= A[km1][km1];
    t = -B[km1];
    for (i = 0; i < km1; i++)
      B[i] += A[i][km1] * t;
  }
  B[0] /= A[0][0];
}

static void
do_solve_complex(double Ar[][ND], double Ai[][ND],
                 double *Br, double *Bi, PRUint32 *X)
{
  PRUint32 i, k, m, kb;
  double ti, tr;
  PRUint32 km1, nm1, kp1;
  double den, prodi, prodr;

  for (k = 1; k <= ND - 1; k++)
  {
    kp1 = k + 1;
    m = X[k - 1];
    tr = Br[m - 1];
    ti = Bi[m - 1];
    Br[m - 1] = Br[k - 1];
    Bi[m - 1] = Bi[k - 1];
    Br[k - 1] = tr;
    Bi[k - 1] = ti;
    nm1 = imin(ND, kp1);
    for (i = kp1; i <= nm1; i++)
    {
      prodr = Ar[i - 1][k - 1] * tr - Ai[i - 1][k - 1] * ti;
      prodi = Ai[i - 1][k - 1] * tr + Ar[i - 1][k - 1] * ti;
      Br[i - 1] += prodr;
      Bi[i - 1] += prodi;
    }
  }
  for (kb = 1; kb <= ND - 1; kb++)
  {
    km1 = ND - kb;
    den = Ar[km1][km1] * Ar[km1][km1] + Ai[km1][km1] * Ai[km1][km1];
    prodr = Br[km1] * Ar[km1][km1] + Bi[km1] * Ai[km1][km1];
    prodi = Bi[km1] * Ar[km1][km1] - Br[km1] * Ai[km1][km1];
    Br[km1] = prodr / den;
    Bi[km1] = prodi / den;
    tr = -Br[km1];
    ti = -Bi[km1];
    for (i = 1; i <= km1; i++)
    {
      prodr = Ar[i - 1][km1] * tr - Ai[i - 1][km1] * ti;
      prodi = Ai[i - 1][km1] * tr + Ar[i - 1][km1] * ti;
      Br[i - 1] += prodr;
      Bi[i - 1] += prodi;
    }
  }
  den = Ar[0][0] * Ar[0][0] + Ai[0][0] * Ai[0][0];
  prodr = Br[0] * Ar[0][0] + Bi[0] * Ai[0][0];
  prodi = Bi[0] * Ar[0][0] - Br[0] * Ai[0][0];
  Br[0] = prodr / den;
  Bi[0] = prodi / den;
}

static void
do_slvrad(double fac1, double alphn, double betan)
{
  PRUint32 i;
  double s2, s3;
  PRInt32 mm, mp, mp1;
  double e1imp;
  double zsafe;

  for (i = 1; i <= ND; i++)
  {
    s2 = -ff[i + ND - 1];
    s3 = -ff[i + N2 - 1];
    zz[i - 1] -= ff[i - 1] * fac1;
    zz[i + ND - 1] += s2 * alphn - s3 * betan;
    zz[i + N2 - 1] += s3 * alphn + s2 * betan;
  }
  if (ND > 2)
  {
    for (mm = ND - 2; mm >= 1; mm--)
    {
      mp = ND - mm;
      mp1 = mp - 1;
      i = iphes[mp1];
      if (i != mp)
      {
        zsafe = zz[mp1];
        zz[mp1] = zz[i - 1];
        zz[i - 1] = zsafe;
        zsafe = zz[mp + ND - 1];
        zz[mp + ND - 1] = zz[i + ND - 1];
        zz[i + ND - 1] = zsafe;
        zsafe = zz[mp + N2 - 1];
        zz[mp + N2 - 1] = zz[i + N2 - 1];
        zz[i + N2 - 1] = zsafe;
      }
      for (i = mp + 1; i <= ND; i++)
      {
        e1imp = jac[i - 1][mp1 - 1];
        zz[i - 1] -= e1imp * zz[mp1];
        zz[i + ND - 1] -= e1imp * zz[mp + ND - 1];
        zz[i + N2 - 1] -= e1imp * zz[mp + N2 - 1];
      }
    }
  }
  do_solve(e1, zz, ip1);
  do_solve_complex(e2r[0], e2i[0], zz + ND, zz + N2, ip2);
  for (mm = 1; mm <= ND - 2; mm++)
  {
    mp = ND - mm;
    mp1 = mp - 1;
    for (i = mp + 1; i <= ND; i++)
    {
      e1imp = jac[i - 1][mp1 - 1];
      zz[i - 1] += e1imp * zz[mp - 1];
      zz[i + ND - 1] += e1imp * zz[mp + ND - 1];
      zz[i + N2 - 1] += e1imp * zz[mp + N2 - 1];
    }
    i = iphes[mp1];
    if (i != mp)
    {
      zsafe = zz[mp - 1];
      zz[mp - 1] = zz[i - 1];
      zz[i - 1] = zsafe;
      zsafe = zz[mp + ND - 1];
      zz[mp + ND - 1] = zz[i + ND - 1];
      zz[i + ND - 1] = zsafe;
      zsafe = zz[mp + N2 - 1];
      zz[mp + N2 - 1] = zz[i + N2 - 1];
      zz[i + N2 - 1] = zsafe;
    }
  }
}

static void
do_slvrar(double fac1)
{
  PRUint32 i;
  PRInt32 mm, mp, mp1;
  double zsafe;

  for (i = 0; i < ND; i++)
    zz[i] -= ff[i] * fac1;

  if (ND > 2)
  {
    for (mm = ND - 2; mm >= 1; mm--)
    {
      mp = ND - mm;
      mp1 = mp - 1;
      i = iphes[mp1];
      if (i != mp)
      {
        zsafe = zz[mp1];
        zz[mp1] = zz[i - 1];
        zz[i - 1] = zsafe;
      }
      for (i = mp; i < ND; i++)
        zz[i] -= jac[i][mp1 - 1] * zz[mp1];
    }
  }
  do_solve(e1, zz, ip1);
  if (ND > 2)
  {
    for (mm = 1; mm <= ND - 2; mm++)
    {
      mp = ND - mm;
      mp1 = mp - 1;
      for (i = mp; i < ND; i++)
        zz[i] += jac[i][mp1 - 1] * zz[mp1];
      i = iphes[mp1];
      if (i != mp)
      {
        zsafe = zz[mp1];
        zz[mp1] = zz[i - 1];
        zz[i - 1] = zsafe;
      }
    }
  }
}

static void
do_slvrai(double fac1, PRUint32 k)
{
  PRUint32 i;
  double s2, s3;
  PRInt32 mm, mp, mp1;
  double e1imp, zsafe;
  double alphn, betn;
  double *z2, *z3, *f2, *f3;
  PRUint32 *xip2;
  double (*xe2r)[ND], (*xe2i)[ND];

  alphn = alphan[k - 1];
  betn = betan[k - 1];
  z2 = zz + (2 * k - 1) * ND;
  z3 = zz + (2 * k * ND);
  f2 = ff + (2 * k - 1) * ND;
  f3 = ff + (2 * k * ND);
  xip2 = ip2 + (k - 1) * ND;
  xe2r = e2r[k - 1];
  xe2i = e2i[k - 1];

  for (i = 1; i <= ND; i++)
  {
    s2 = -f2[i - 1];
    s3 = -f3[i - 1];
    z2[i - 1] = z2[i - 1] + s2 * alphn - s3 * betn;
    z3[i - 1] = z3[i - 1] + s3 * alphn + s2 * betn;
  }
  if (ND > 2)
  {
    for (mm = ND - 2; mm >= 1; mm--)
    {
      mp = ND - mm;
      mp1 = mp - 1;
      i = iphes[mp1];
      if (i != mp)
      {
        zsafe = z2[mp1];
        z2[mp1] = z2[i - 1];
        z2[i - 1] = zsafe;
        zsafe = z3[mp1];
        z3[mp1] = z3[i - 1];
        z3[i - 1] = zsafe;
      }
      for (i = mp + 1; i <= ND; i++)
      {
        e1imp = jac[i - 1][mp1 - 1];
        z2[i - 1] -= e1imp * z2[mp1];
        z3[i - 1] -= e1imp * z3[mp1];
      }
    }
  }

  do_solve_complex(xe2r, xe2i, z2, z3, xip2);

  if (ND > 2)
  {
    for (mm = 1; mm <= ND - 2; mm++)
    {
      mp = ND - mm;
      mp1 = mp - 1;
      for (i = mp + 1; i <= ND; i++)
      {
        e1imp = jac[i - 1][mp1 - 1];
        z2[i - 1] += e1imp * z2[mp1];
        z3[i - 1] += e1imp * z3[mp1];
      }
      i = iphes[mp1];
      if (i != mp)
      {
        zsafe = z2[mp1];
        z2[mp1] = z2[i - 1];
        z2[i - 1] = zsafe;
        zsafe = z3[mp1];
        z3[mp1] = z3[i - 1];
        z3[i - 1] = zsafe;
      }
    }
  }
}

static void
do_decomc(double alphan, double betan,
	  PRUint32 km1, PRUint32 *singular)
{
  PRUint32 i, j;

  for (j = 1; j < ND; j++)
  {
    e2r[km1][j][j - 1] = -jac[j][j - 1];
    e2i[km1][j][j - 1] = 0.0;
  }
  for (j = 1; j <= ND; j++)
  {
    for (i = 1; i <= j; i++)
    {
      e2i[km1][i - 1][j - 1] = 0.0;
      e2r[km1][i - 1][j - 1] = -jac[i - 1][j - 1];
    }
    e2r[km1][j - 1][j - 1] += alphan;
    e2i[km1][j - 1][j - 1] = betan;
  }
  dechc(singular, km1);
  return;
}

static void
do_estrad()
{
  PRUint32 i;
  double hee1, hee2, hee3, t;
  PRUint32 mm, mp;
  double zsafe;

  hee1 = DD1 / step_size;
  hee2 = DD2 / step_size;
  hee3 = DD3 / step_size;

  for (i = 1; i <= ND; i++)
  {
    ff[i + ND - 1] = hee1 * zz[i - 1] + hee2 * zz[i + ND - 1] + hee3 * zz[i + N2 - 1];
    cont[i + NS - 1] = ff[i + ND - 1] + y_0[i - 1];
  }
  if (ND > 2)
  {
    for (mm = ND - 2; mm >= 1; mm--)
    {
      mp = ND - mm;
      i = iphes[mp - 1];
      if (i != mp)
      {
        zsafe = cont[mp + NS - 1];
        cont[mp + NS - 1] = cont[i + NS - 1];
        cont[i + NS - 1] = zsafe;
      }
      for (i = mp; i < ND; i++)
        cont[i + NS] -= jac[i][mp - 2] * cont[mp + NS - 1];
    }
  }
  do_solve(e1, cont + NS, ip1);
  if (ND > 2)
  {
    for (mm = 1; mm <= ND - 2; mm++)
    {
      mp = ND - mm;
      for (i = mp; i < ND; i++)
        cont[i + NS] += jac[i][mp - 2] * cont[mp + NS - 1];
      
      i = iphes[mp - 1];
      if (i != mp)
      {
        zsafe = cont[mp + NS - 1];
        cont[mp + NS - 1] = cont[i + NS - 1];
        cont[i + NS - 1] = zsafe;
      }
    }
  }

  err = 0.0;
  got_nan = PR_FALSE;
  for (i = 0; i < ND; i++)
  {
    t = cont[i + NS] / scal[i];
    err += t * t;
  }

  err = max(sqrt(err / ND), 1e-10);
  if (!finite(err))
  {
    got_nan = PR_TRUE;
  }
  if (!got_nan && err < 1.0)
    return;

  if (!first && !reject)
    return;

  for (i = 0; i < NS; i++)
    cont[i] = all_yvalues[i];
  for (i = NS; i < NS+ND; i++)
    cont[i] += all_yvalues[i];
  EvaluateSimple(x, cont);
  ComputeDifferentials(x, cont, ff);
  
  for (i = 0; i < ND; i++)
    cont[i + NS] = ff[i] + ff[i + ND];

  if (ND > 2)
  {
    for (mm = ND - 2; mm >= 1; mm--)
    {
      mp = ND - mm;
      i = iphes[mp - 1];
      if (i != mp)
      {
        zsafe = cont[mp + NS - 1];
        cont[mp + NS - 1] = cont[i + NS - 1];
        cont[i + NS - 1] = zsafe;
      }
    
      for (i = mp; i < ND; i++)
        cont[i + NS] -= jac[i][mp - 2] * cont[mp + NS - 1];
    }
  }
  do_solve(e1, cont + NS, ip1);

  if (ND > 2)
  {
    for (mm = 1; mm <= ND - 2; mm++)
    {
      mp = ND - mm;
      for (i = mp + 1; i <= ND; i++)
        cont[i + NS - 1] += jac[i - 1][mp - 2] * cont[mp + NS - 1];
      i = iphes[mp - 1];
      if (i != mp)
      {
        zsafe = cont[mp + NS - 1];
        cont[mp + NS - 1] = cont[i + NS - 1];
        cont[i + NS - 1] = zsafe;
      }
    }
  }
  got_nan = PR_FALSE;
  err = 0;
  
  for (i = 0; i < ND; i++)
  {
    t = cont[i + NS] / scal[i];
    err += t * t;
  }
  err = max(sqrt(err / ND), 1e-10);
  if (!finite(err))
  {
    got_nan = PR_TRUE;
  }
}

static void
do_estrav()
{
  PRUint32 i, k;
  double sum, t;
  PRUint32 mm, mp;
  double zsafe;

  for (i = 0; i < ND; i++)
  {
    sum = 0.0;
    for (k = 0; k < current_order; k++)
      sum += dd[k] * zz[i + k * ND];
    ff[i + ND] = sum / step_size;
    cont[i + NS] = ff[i + ND] + y_0[i];
  }
  if (ND > 2)
  {
    for (mm = ND - 2; mm >= 1; mm--)
    {
      mp = ND - mm;
      i = iphes[mp - 1];
      if (i != mp)
      {
        zsafe = cont[mp + NS - 1];
        cont[mp + NS - 1] = cont[i + NS - 1];
        cont[i + NS - 1] = zsafe;
      }
      for (i = mp; i < ND; i++)
        cont[i + NS] -= jac[i][mp - 2] * cont[mp + NS - 1];
    }
  }
  do_solve(e1, cont + NS, ip1);
  if (ND > 2)
  {
    for (mm = 1; mm <= ND - 2; mm++)
    {
      mp = ND - mm;
      for (i = mp; i < ND; i++)
        cont[i + NS] += jac[i][mp - 2] * cont[mp + NS - 1];
      i = iphes[mp - 1];
      if (i != mp)
      {
        zsafe = cont[mp + NS - 1];
        cont[mp + NS - 1] = cont[i + NS - 1];
        cont[i + NS - 1] = zsafe;
      }
    }
  }
  got_nan = PR_FALSE;
  err = 0.0;
  for (i = 1; i <= ND; i++)
  {
    t = cont[i + NS - 1] / scal[i - 1];
    err += t * t;
  }

  if (!finite(err))
  {
    got_nan = PR_TRUE;
  }
  if (err < 1.0)
    return;

  if (!first && !reject)
    return;

  for (i = 0; i < NS; i++)
    cont[i] = all_yvalues[i];
  for (i = NS; i < ND + NS; i++)
    cont[i] += all_yvalues[i];

  EvaluateSimple(x, cont);
  ComputeDifferentials(x, cont, ff);

  for (i = 0; i < ND; i++)
    cont[i + NS] = ff[i] + ff[i + ND];

  if (ND > 2)
  {
    for (mm = ND - 2; mm >= 1; mm--)
    {
      mp = ND - mm;
      i = iphes[mp - 1];
      if (i != mp)
      {
        zsafe = cont[mp + NS - 1];
        cont[mp + NS - 1] = cont[i + NS - 1];
        cont[i + NS - 1] = zsafe;
      }
      for (i = mp; i < ND; i++)
        cont[i + NS] -= jac[i][mp - 2] * cont[mp + NS - 1];
    }
  }

  do_solve(e1, cont + NS, ip1);

  if (ND > 2)
  {
    for (mm = 1; mm <= ND - 2; mm++)
    {
      mp = ND - mm;
      for (i = mp; i < ND; i++)
        cont[i + NS] += jac[i][mp - 2] * cont[mp + NS - 1];
      
      i = iphes[mp - 1];
      if (i != mp)
      {
        zsafe = cont[mp + NS - 1];
        cont[mp + NS - 1] = cont[i + NS - 1];
        cont[i + NS - 1] = zsafe;
      }
    }
  }
  err = 0.0;
  for (i = 0; i < ND; i++)
  {
    t = cont[i + NS] / scal[i];
    err += t * t;
  }
  err = max(sqrt(err / ND), 1e-10);
  if (!finite(err))
  {
    got_nan = PR_TRUE;
  }
}

static void
step_rejection()
{
  unexp = PR_TRUE;
  reject = PR_TRUE;
  last = PR_FALSE;
  step_size *= 0.5;
  hhfac = 0.5;
}

static void
do_integrator_cycle()
{
  double hquot, xph, dyno, fac;
  PRUint32 i, j, k, l, m, new_order;
  PRUint32 singular;
  PRUint32 newton_its;
  
  if (ikeep != 1)
  {
    if (!caljac && !skipjac)
    {
      PRUint32 i, j;
      /* Compute the Jacobian matrix numerically... */
      for (i = NS; i < NS+ND; i++)
      {
        double ysafe = all_yvalues[i];
        double delt = sqrt(ROUNDING_UNIT * max(1.0E-5, abs_d(ysafe)));
        all_yvalues[i] = ysafe + delt;
        EvaluateSimple(x, all_yvalues);
        ComputeDifferentials(x, all_yvalues, cont + NS);
        for (j = 0; j < ND; j++)
        {
          jac[j][i - NS] = (cont[j + NS] - y_0[j]) / delt;
        }
        all_yvalues[i] = ysafe;
      }
      caljac = PR_TRUE;
      calhes = PR_TRUE;
      EvaluateSimple(x, all_yvalues);
    }
    skipjac = PR_FALSE;

    /* Consider switching order... */
    ichan++;
    hquot = step_size / step_size_old;
    new_order = current_order;
    thetat = min(10, max(theta, thetat * 0.5));
    if (newton_count > 1 && thetat <= INCREASE_ORDER_CONTRACTIVITY &&
        hquot < INCREASE_ORDER_HOLDRATIO_MAX &&
        hquot > INCREASE_ORDER_HOLDRATIO_MIN)
      new_order = imin(MAXIMAL_ORDER, current_order + 2);
    if (thetat >= REDUCE_ORDER_CONTRACTIVITY || unexp)
      new_order = imax(MINIMAL_ORDER, current_order - 2);
    if (ichan >= 1 && unexn)
      new_order = imax(MINIMAL_ORDER, current_order - 2);
    if (ichan <= 10)
      new_order = imin(current_order, new_order);
    
    change = (current_order != new_order);
    unexn = PR_FALSE;
    unexp = PR_FALSE;
    
    /* If we decided to change order, update everything... */
    if (change)
    {
      PRUint32 i;
      current_order = new_order;
      ichan = 1;
      n_current_order = current_order * ND;
      lrc = n_current_order + ND;
      do_coercv(current_order);
      expo = 1.0 / (current_order + 1.0);
      expmns = (current_order + 1.0) / (current_order * 2.0);
      rel_tolerance_1 = pow(rel_tolerance, expmns) * 0.1;
      abs_tolerance_1 = rel_tolerance_1 * quott;
      for (i = 0; i < ND; i++)
        scal[i] = abs_tolerance_1 + rel_tolerance_1 * abs_d(all_yvalues[i + NS]);
    }
    
    /* Calculate and decompose E1 and E2 */
    fac1 = u1 / step_size;
    do_decomr(fac1, &singular);
    if (singular)
    {
      step_rejection();
      return;
    }

    m = (current_order - 1) / 2;
    for (k = 1; k <= m; k++)
    {
      alphan[k - 1] = alpha[k - 1] / step_size;
      betan[k - 1] = beta[k - 1] / step_size;
      do_decomc(alphan[k - 1], betan[k - 1], k - 1,
                &singular);
      if (singular)
      {
        step_rejection();
        return;
      }
    }
  }
  else /* if (ikeep == 1) */
  {
    ichan++;
    ikeep = 0;
    /* See if we need to restart this step...*/
    if (ichan >= 10 && current_order < MAXIMAL_ORDER)
      return;
  }

  if (0.1 * abs_d(step_size) <= abs_d(x) * ROUNDING_UNIT)
  {
    integration_failed("stepSizeTooSmall");
    return;
  }

  /* Find the next X value... */
  xph = x + step_size;
  if (!finite(xph))
  {
    integration_failed("boundNotFinite");
    return;
  }

  if (current_order == 3)
  {
    double c3q, c1q, c2q, ak1, ak2, ak3, z1i, z2i, z3i, expmi, fnewt, xp;
    if (first || change)
    {
      for (i = 0; i < ND * 3; i++)
      {
        zz[i] = 0.0;
        ff[i] = 0.0;
      }
    }
    else
    {
      hquot = step_size / step_size_old;
      c3q = hquot;
      c1q = C31 * c3q;
      c2q = C32 * c3q;
      for (i = 1; i <= ND; i++)
      {
        ak1 = cont[i + (NS + ND) + NS - 1];
        ak2 = cont[i + NS + (NS + ND) * 2 - 1];
        ak3 = cont[i + NS + (NS + ND) * 3 - 1];
        z1i = c1q * (ak1 + (c1q - C32M1) * (ak2 + (c1q - C31M1) * ak3));
        z2i = c2q * (ak1 + (c2q - C32M1) * (ak2 + (c2q - C31M1) * ak3));
        z3i = c3q * (ak1 + (c3q - C32M1) * (ak2 + (c3q - C31M1) * ak3));
        zz[i - 1] = z1i;
        zz[i + ND - 1] = z2i;
        zz[i + N2 - 1] = z3i;
        ff[i - 1] = ti311 * z1i + ti312 * z2i + ti313 * z3i;
        ff[i + ND - 1] = ti321 * z1i + ti322 * z2i +
                         ti323 * z3i;
        ff[i + N2 - 1] = ti331 * z1i + ti332 * z2i + 
                         ti333 * z3i;
      }
    }
    newton_count = 0;
    newton_its = MAXIMAL_NEWTON;
    expmi = 1.0 / expmns;
    fnewt = max(ROUNDING_UNIT * 10 / rel_tolerance_1,
                min(0.03, pow(rel_tolerance_1, expmi - 1.0)));
    faccon = pow(max(faccon, ROUNDING_UNIT), 0.8);
    theta = THET;

    while (PR_TRUE)
    {
      if (newton_count >= newton_its)
      {
        step_rejection();
        return;
      }

      for (i = 0; i < NS; i++)
        cont[i] = all_yvalues[i];

      /* 3 point evaluation... */
      for (i = NS; i < NS+ND; i++)
      {
        cont[i] = all_yvalues[i] + zz[i - NS];
      }
      xp = x + C31 * step_size;
      EvaluateSimple(xp, cont);
      ComputeDifferentials(xp, cont, zz);
      for (i = NS; i < NS+ND; i++)
      {
        cont[i] = all_yvalues[i] + zz[i + ND - NS];
      }
      xp = x + C32 * step_size;
      EvaluateSimple(xp, cont);
      ComputeDifferentials(xp, cont, zz + ND);
      for (i = NS; i < NS+ND; i++)
      {
        cont[i] = all_yvalues[i] + zz[i + N2 - NS];
      }
      EvaluateSimple(xph, cont);
      ComputeDifferentials(xph, cont, zz + N2);
      
      for (i = 1; i <= ND; i++)
      {
        double a1, a2, a3;
        a1 = zz[i - 1];
        a2 = zz[i + ND - 1];
        a3 = zz[i + N2 - 1];
        zz[i - 1] = ti311 * a1 + ti312 * a2 + ti313 * a3;
        zz[i + ND - 1] = ti321 * a1 + ti322 * a2 + ti323 * a3;
        zz[i + N2 - 1] = ti331 * a1 + ti332 * a2 + ti333 * a3;
      }
      
      do_slvrad(fac1, alphan[0], betan[0]);
      newton_count++;
      dyno = 0.0;
      for (i = 1; i <= ND; i++)
      {
        double t1, t2, t3, denom;
        denom = scal[i - 1];
        t1 = zz[i - 1] / denom;
        t2 = zz[i - 1 + ND] / denom;
        t3 = zz[i - 1 + N2] / denom;
        dyno += t1 * t1 + t2 * t2 + t3 * t3;
      }
      dyno = sqrt(dyno / (ND * 3));
      if (newton_count > 1 && newton_count < newton_its)
      {
        double thq = dyno / dynold;
        if (newton_count == 2)
          theta = thq;
        else
          theta = sqrt(thq * thqold);
        thqold = thq;
        if (theta < 0.99)
        {
          double dyth;
          faccon = theta / (1.0 - theta);
          dyth = faccon * dyno * pow(theta, newton_its - 1 - newton_count) / fnewt;
          if (dyth >= 1.0)
          {
            double qnewt = max(1e-4, min(20, dyth));
            hhfac = 0.8 * pow(qnewt, -1.0 / (newton_its + 4.0 - 1 - newton_count));
            step_size *= hhfac;
            reject = PR_TRUE;
            last = PR_FALSE;
            if (hhfac <= 0.5)
              unexn = PR_TRUE;
            return;
          }
        }
        else
        {
          step_rejection();
          return;
        }
      }
      dynold = max(dyno, ROUNDING_UNIT);

      for (i = 1; i <= ND; i++)
      {
        PRUint32 in, in2;
        double f1i, f2i, f3i;
        in = i + ND - 1;
        in2 = i + N2 - 1;
        f1i = ff[i - 1] + zz[i - 1];
        f2i = ff[in] + zz[in];
        f3i = ff[in2] + zz[in2];
        ff[i - 1] = f1i;
        ff[in] = f2i;
        ff[in2] = f3i;
        zz[i - 1] = t311 * f1i + t312 * f2i + t313 * f3i;
        zz[in] = t321 * f1i + t322 * f2i + t323 * f3i;
        zz[in2] = t331 * f1i + f2i;
      }
      
      if (faccon * dyno > fnewt)
        continue;
      break;
    }
    
    do_estrad();

    if (!got_nan && err < 1.0)
    {
      for (i = 1; i <= ND; i++)
      {
        double ak, acont3;
        all_yvalues[i + NS - 1] += zz[i + N2 - 1];
        z2i = zz[i + ND - 1];
        z1i = zz[i - 1];
        cont[i + (NS + ND) + NS - 1] = (z2i - zz[i + N2 - 1]) / C32M1;
        ak = (z1i - z2i) / C31MC2;
        acont3 = (ak - (z1i / C31)) / C32;
        cont[i + NS + (NS + ND) * 2 - 1] =
          (ak - cont[i + (NS + ND) + NS - 1]) / C31M1;
        cont[i + NS + (NS + ND) * 3 - 1] =
          cont[i + NS + (NS + ND) * 2 - 1] - acont3;
      }
    }
  }
  else if (current_order == 5)
  {
    double expmi, fnewt;
    if (first || change)
    {
      for (i = 0; i < ND * 5; i++)
      {
        zz[i] = 0.0;
        ff[i] = 0.0;
      }
    }
    else
    {
      hquot = step_size / step_size_old;
      for (k = 1; k <= 5; k++)
      {
        double ccq;
        ccq = weights[k] * hquot;
        for (i = 1; i <= ND; i++)
        {
          PRUint32 l;
          double val = cont[i + NS + (NS + ND) * 5 - 1];
          for (l = 4; l >= 1; l--)
            val = cont[i + NS + l * (NS + ND) - 1] + (ccq - weights[5 - l] + 1.0) * val;
          zz[i + (k - 1) * ND - 1] = ccq * val;
        }
      }
      for (i = 1; i <= ND; i++)
      {
        double z1i, z2i, z3i, z4i, z5i;
        z1i = zz[i - 1];
        z2i = zz[i + ND - 1];
        z3i = zz[i + N2 - 1];
        z4i = zz[i + N3 - 1];
        z5i = zz[i + N4 - 1];
        ff[i - 1] = ti511 * z1i + ti512 * z2i + ti513 * z3i +
                    ti514 * z4i + ti515 * z5i;
        ff[i + ND - 1] = ti521 * z1i + ti522 * z2i +
                         ti523 * z3i + ti524 * z4i +
                         ti525 * z5i;
        ff[i + N2 - 1] = ti531 * z1i + ti532 * z2i +
                         ti533 * z3i + ti534 * z4i +
                         ti535 * z5i;
        ff[i + N3 - 1] = ti541 * z1i + ti542 * z2i +
                         ti543 * z3i + ti544 * z4i +
                         ti545 * z5i;
        ff[i + N4 - 1] = ti551 * z1i + ti552 * z2i +
                         ti553 * z3i + ti554 * z4i +
                         ti555 * z5i;
      }
    }
    newton_count = 0;
    newton_its = MAXIMAL_NEWTON + 5;
    expmi = 1.0 / expmns;
    fnewt = max(ROUNDING_UNIT * 10 / rel_tolerance_1,
                min(0.03, pow(rel_tolerance_1, expmi - 1.0)));
    faccon = pow(max(faccon, ROUNDING_UNIT), 0.8);
    theta = THET;

    while (PR_TRUE)
    {
      if (newton_count >= newton_its)
      {
        step_rejection();
        return;
      }

      for (k = 0; k <= 4; k++)
      {
        double xp;
        PRUint32 iadd = k * ND;
        for (i = 0; i < NS; i++)
          cont[i] = all_yvalues[i];
        for (i = NS; i < NS + ND; i++)
          cont[i] = all_yvalues[i] + zz[iadd + i - NS];
        xp = x + weights[k + 1] * step_size;
        EvaluateSimple(xp, cont);
        ComputeDifferentials(xp, cont, zz + iadd);
      }
      for (i = 1; i <= ND; i++)
      {
        double z1i, z2i, z3i, z4i, z5i;
        z1i = zz[i - 1];
        z2i = zz[i + ND - 1];
        z3i = zz[i + N2 - 1];
        z4i = zz[i + N3 - 1];
        z5i = zz[i + N4 - 1];
        zz[i - 1] = ti511 * z1i + ti512 * z2i + ti513 * z3i +
                    ti514 * z4i + ti515 * z5i;
        zz[i + ND - 1] = ti521 * z1i + ti522 * z2i + ti523 * z3i +
                         ti524 * z4i + ti525 * z5i;
        zz[i + N2 - 1] = ti531 * z1i + ti532 * z2i + ti533 * z3i +
                         ti534 * z4i + ti535 * z5i;
        zz[i + N3 - 1] = ti541 * z1i + ti542 * z2i + ti543 * z3i +
                         ti544 * z4i + ti545 * z5i;
        zz[i + N4 - 1] = ti551 * z1i + ti552 * z2i + ti553 * z3i +
                         ti554 * z4i + ti555 * z5i;
      }
      do_slvrar(fac1);
    
      for (k = 1; k <= 2; k++)
        do_slvrai(fac1, k);
      
      newton_count++;
      dyno = 0.0;
      for (i = 1; i <= ND; i++)
      {
        double denom = scal[i - 1];
        for (k = 0; k < 5; k++)
        {
          double t = zz[i + k * ND - 1] / denom;
          dyno += t * t;
        }
      }
      dyno = sqrt(dyno / (5.0 * ND));
      if (newton_count > 1 && newton_count < newton_its)
      {
        double thq = dyno / dynold;
        if (newton_count == 2)
          theta = thq;
        else
          theta = sqrt(thq * thqold);
        thqold = thq;
        if (theta < 0.99)
        {
          double dyth;
          faccon = theta / (1.0 - theta);
          dyth = faccon * dyno * pow(theta,
                     (newton_its - 1 - newton_count)) / fnewt;
          if (dyth >= 1.0)
          {
            hhfac = 0.8 * pow(max(1.0E-4, min(20.0, dyth)),
                        -1.0 / (4.0 + newton_its - 1 - newton_count));
            step_size *= hhfac;
            reject = PR_TRUE;
            last = PR_FALSE;
            if (hhfac < 0.5)
              unexn = PR_TRUE;
            return;
          }
        }
      }
      dynold = max(dyno, ROUNDING_UNIT);
      for (i = 1; i <= ND; i++)
      {
        double z1i, z2i, z3i, z4i, z5i;
        z1i = ff[i - 1] + zz[i - 1];
        z2i = ff[i + ND - 1] + zz[i + ND - 1];
        z3i = ff[i + N2 - 1] + zz[i + N2 - 1];
        z4i = ff[i + N3 - 1] + zz[i + N3 - 1];
        z5i = ff[i + N4 - 1] + zz[i + N4 - 1];
        ff[i - 1] = z1i;
        ff[i + ND - 1] = z2i;
        ff[i + N2 - 1] = z3i;
        ff[i + N3 - 1] = z4i;
        ff[i + N4 - 1] = z5i;
        zz[i - 1] = t511 * z1i + t512 * z2i + t513 * z3i +
                    t514 * z4i + t515 * z5i;
        zz[i + ND - 1] = t521 * z1i + t522 * z2i + t523 * z3i +
                         t524 * z4i + t525 * z5i;
        zz[i + N2 - 1] = t531 * z1i + t532 * z2i + t533 * z3i +
                         t534 * z4i + t535 * z5i;
        zz[i + N3 - 1] = t541 * z1i + t542 * z2i + t543 * z3i +
                         t544 * z4i + t545 * z5i;
        zz[i + N4 - 1] = t551 * z1i + z2i + z4i;
      }
      if (faccon * dyno > fnewt)
        continue;
      break;
    }
    
    do_estrav();

    if (!got_nan && err < 1.0)
    {
      for (i = NS; i < NS+ND; i++)
      {
        all_yvalues[i] += zz[i + N4 - NS];
        cont[i + (NS+ND)*5] = zz[i - NS] / weights[1];
      }
      for (k = 1; k < 5; k++)
      {
        double fact = 1.0 / (weights[5 - k] - weights[5 - k + 1]);
        for (i = 1; i <= ND; i++)
          cont[i + k * (ND + NS) + NS - 1] = (zz[i + (5-1-k) * ND - 1] -
                                              zz[i + (5-k)*ND - 1]) * fact;
      }
      for (j = 2; j <= 5; j++)
        for (k = 5; k >= j; k--)
        {
          double fact = 1.0 / (weights[5 - k] - weights[5 - k + j]);
          for (i = 1; i <= ND; i++)
            cont[i + k * (ND + NS) + NS - 1] = (cont[i + k * (NS + ND) + NS - 1] -
                                                cont[i + (k - 1)*(NS + ND) + NS - 1]) * fact;
        }
    }
  }
  else if (current_order == 7)
  {
    double expmi, fnewt;
    if (first || change)
      for (i = 1; i <= 7 * ND; i++)
      {
        zz[i - 1] = 0.0;
        ff[i - 1] = 0.0;
      }
    else
    {
      hquot = step_size / step_size_old;
      for (k = 1; k <= 7; k++)
      {
        double ccq = weights[k] * hquot;
        for (i = 1; i <= ND; i++)
        {
          double val = cont[i + 7 * (ND + NS) + NS - 1];
          for (l = 7-1; l >= 1; l--)
            val = cont[i + l * (NS + ND) + NS - 1] +
                  (ccq - weights[7 - l] + 1.0) * val;
          zz[i + (k-1) * ND - 1] = ccq * val;
        }
      }
      for (i = 1; i <= ND; i++)
      {
        double z1i, z2i, z3i, z4i, z5i, z6i, z7i;
        z1i = zz[i - 1];
        z2i = zz[i + ND - 1];
        z3i = zz[i + N2 - 1];
        z4i = zz[i + N3 - 1];
        z5i = zz[i + N4 - 1];
        z6i = zz[i + N5 - 1];
        z7i = zz[i + N6 - 1];
        ff[i - 1] = ti711 * z1i + ti712 * z2i + ti713 * z3i +
                    ti714 * z4i + ti715 * z5i + ti716 * z6i +
                    ti717 * z7i;
        ff[i + ND - 1] = ti721 * z1i + ti722 * z2i +
                         ti723 * z3i + ti724 * z4i +
                         ti725 * z5i + ti726 * z6i +
                         ti727 * z7i;
        ff[i + N2 - 1] = ti731 * z1i + ti732 * z2i + ti733 * z3i +
                         ti734 * z4i + ti735 * z5i + ti736 * z6i +
                         ti737 * z7i;
        ff[i + N3 - 1] = ti741 * z1i + ti742 * z2i +
                         ti743 * z3i + ti744 * z4i +
                         ti745 * z5i + ti746 * z6i +
                         ti747 * z7i;
        ff[i + N4 - 1] = ti751 * z1i + ti752 * z2i + ti753 * z3i +
                    ti754 * z4i + ti755 * z5i + ti756 * z6i +
                    ti757 * z7i;
        ff[i + N5 - 1] = ti761 * z1i + ti762 * z2i +
                         ti763 * z3i + ti764 * z4i +
                         ti765 * z5i + ti766 * z6i +
                         ti767 * z7i;
        ff[i + N6 - 1] = ti771 * z1i + ti772 * z2i + ti773 * z3i +
                         ti774 * z4i + ti775 * z5i + ti776 * z6i +
                         ti777 * z7i;
      }
    }
    newton_count = 0;
    newton_its = MAXIMAL_NEWTON + 10;
    expmi = 1.0 / expmns;
    fnewt = max(10 * ROUNDING_UNIT / rel_tolerance_1,
                min(0.03, pow(rel_tolerance_1, expmi - 1.0)));
    faccon = pow(max(faccon, ROUNDING_UNIT), 0.8);
    theta = THET;
    while (PR_TRUE)
    {
      double xp;
      if (newton_count > newton_its)
      {
        step_rejection();
        return;
      }
      for (k = 0; k < 7; k++)
      {
        PRUint32 iadd = k * ND;
        for (i = 0; i < NS; i++)
          cont[i] = all_yvalues[i];
        for (i = NS; i < NS+ND; i++)
          cont[i] = all_yvalues[i] + zz[iadd + i - NS];
        xp = x + weights[k + 1] * step_size;
        EvaluateSimple(xp, cont);
        ComputeDifferentials(xp, cont, zz + iadd);
      }
      for (i = 1; i <= ND; i++)
      {
        double z1i, z2i, z3i, z4i, z5i, z6i, z7i;
        z1i = zz[i - 1];
        z2i = zz[i + ND - 1];
        z3i = zz[i + N2 - 1];
        z4i = zz[i + N3 - 1];
        z5i = zz[i + N4 - 1];
        z6i = zz[i + N5 - 1];
        z7i = zz[i + N6 - 1];
        zz[i - 1] = ti711 * z1i + ti712 * z2i + ti713 * z3i +
                    ti714 * z4i + ti715 * z5i + ti716 * z6i +
                    ti717 * z7i;
        zz[i + ND - 1] = ti721 * z1i + ti722 * z2i +
                         ti723 * z3i + ti724 * z4i +
                         ti725 * z5i + ti726 * z6i +
                         ti727 * z7i;
        zz[i + N2 - 1] = ti731 * z1i + ti732 * z2i + ti733 * z3i +
                         ti734 * z4i + ti735 * z5i + ti736 * z6i +
                         ti737 * z7i;
        zz[i + N3 - 1] = ti741 * z1i + ti742 * z2i +
                         ti743 * z3i + ti744 * z4i +
                         ti745 * z5i + ti746 * z6i +
                         ti747 * z7i;
        zz[i + N4 - 1] = ti751 * z1i + ti752 * z2i + ti753 * z3i +
                    ti754 * z4i + ti755 * z5i + ti756 * z6i +
                    ti757 * z7i;
        zz[i + N5 - 1] = ti761 * z1i + ti762 * z2i +
                         ti763 * z3i + ti764 * z4i +
                         ti765 * z5i + ti766 * z6i +
                         ti767 * z7i;
        zz[i + N6 - 1] = ti771 * z1i + ti772 * z2i + ti773 * z3i +
                         ti774 * z4i + ti775 * z5i + ti776 * z6i +
                         ti777 * z7i;        
      }
      do_slvrar(fac1);
      for (k = 1; k <= 3; k++)
        do_slvrai(fac1, k);
      newton_count++;
      dyno = 0.0;
      for (i = 1; i <= ND; i++)
      {
        double denom = scal[i - 1];
        for (k = 0; k < 7; k++)
        {
          double t = zz[i + k * ND - 1] / denom;
          dyno += t * t;
        }
      }
      dyno = sqrt(dyno / (7 * ND));
      if (newton_count > 1 && newton_count < newton_its)
      {
        double thq = dyno / dynold;
        if (newton_count == 2)
          theta = thq;
        else
          theta = sqrt(thq * thqold);
        thqold = thq;
        if (theta < 0.99)
        {
          double dyth;
          faccon = theta / (1.0 - theta);
          dyth = faccon * dyno * pow(theta,
                                     newton_its - 1 - newton_count) / fnewt;
          if (dyth >= 1.0)
          {
            double qnewt = max(1.0e-4, min(20.0, dyth));
            hhfac = 0.8 * pow(qnewt, -1.0 / (4.0 + newton_its - 1 - newton_count));
            step_size *= hhfac;
            reject = PR_TRUE;
            last = PR_FALSE;
            if (hhfac < 0.5)
              unexn = PR_TRUE;
            return;
          }
        }
        else
        {
          step_rejection();
          return;
        }
      }
      dynold = max(dyno, ROUNDING_UNIT);
      for (i = 1; i <= ND; i++)
      {
        double z1i, z2i, z3i, z4i, z5i, z6i, z7i;
        z1i = zz[i - 1];
        z2i = zz[i + ND - 1];
        z3i = zz[i + N2 - 1];
        z4i = zz[i + N3 - 1];
        z5i = zz[i + N4 - 1];
        z6i = zz[i + N5 - 1];
        z7i = zz[i + N6 - 1];
        zz[i - 1] = t711 * z1i + t712 * z2i + t713 * z3i +
                    t714 * z4i + t715 * z5i + t716 * z6i +
                    t717 * z7i;
        zz[i + ND - 1] = t721 * z1i + t722 * z2i +
                         t723 * z3i + t724 * z4i +
                         t725 * z5i + t726 * z6i +
                         t727 * z7i;
        zz[i + N2 - 1] = t731 * z1i + t732 * z2i + t733 * z3i +
                         t734 * z4i + t735 * z5i + t736 * z6i +
                         t737 * z7i;
        zz[i + N3 - 1] = t741 * z1i + t742 * z2i +
                         t743 * z3i + t744 * z4i +
                         t745 * z5i + t746 * z6i +
                         t747 * z7i;
        zz[i + N4 - 1] = t751 * z1i + t752 * z2i + t753 * z3i +
                    t754 * z4i + t755 * z5i + t756 * z6i +
                    t757 * z7i;
        zz[i + N5 - 1] = t761 * z1i + t762 * z2i +
                         t763 * z3i + t764 * z4i +
                         t765 * z5i + t766 * z6i +
                         t767 * z7i;
        zz[i + N6 - 1] = t771 * z1i + z2i + z4i + z6i;
      }
      if (faccon * dyno > fnewt)
        continue;
      break;
    }

    do_estrav();
    
    if (!got_nan && err < 1.0)
    {
      for (i = NS; i <= ND + NS; i++)
      {
        all_yvalues[i] += zz[i + 6 * ND - NS];
        cont[i + 7 * (NS + ND)] = zz[i - NS] / weights[1];
      }
      for (k = 1; k < 7; k++)
      {
        double fact = 1.0 / (weights[7 - k] - weights[7 - k + 1]);
        for (i = 1; i <= ND; i++)
          cont[i + k * (NS + ND) + NS - 1] = (zz[i + (7 - 1 - k) * ND - 1] -
                                              zz[i + (7 - k) * ND - 1]) * fact;
      }
      for (j = 2; j <= 5; j++)
        for (k = 7; k >= j; k--)
        {
          double fact = 1.0 / (weights[7 - k] - weights[7 - k + j]);
          for (i = 1; i <= ND; i++)
            cont[i + k * (ND + NS) + NS - 1] =
               (cont[i + k * (ND + NS) + NS - 1] -
                cont[i + (k - 1) * (ND + NS) + NS - 1]) * fact;
        }
    }
  }
  else /* if current_order == 1 */
  {
    double expmi, fnewt;
    if (first || change)
    {
      zz[0] = 0.0;
      ff[0] = 0.0;
    }
    else
    {
      hquot = step_size / step_size_old;
      for (i = 1; i <= ND; i++)
      {
        double z1i = hquot * cont[i + (NS + ND) + NS - 1];
        zz[i - 1] = z1i;
        ff[i - 1] = z1i;
      }
    }
    
    newton_count = 0;
    newton_its = MAXIMAL_NEWTON - 3;
    expmi = 1.0 / expmns;
    fnewt = max(10 * ROUNDING_UNIT / rel_tolerance_1, 0.03);
    faccon = pow(max(faccon, ROUNDING_UNIT), 0.8);
    theta = THET;
    while (PR_TRUE)
    {
      if (newton_count >= newton_its)
      {
        step_rejection();
        return;
      }
      for (i = 0; i < NS; i++)
        cont[i] = all_yvalues[i];
      for (i = NS; i < ND + NS; i++)
        cont[i] = all_yvalues[i] + zz[i - NS];
      EvaluateSimple(xph, cont);
      ComputeDifferentials(xph, cont, zz);
      do_slvrar(fac1);
      newton_count++;
      dyno = 0.0;
      for (i = 1; i <= ND; i++)
      {
        double denom = scal[i - 1], t;
        t = zz[i - 1] / denom;
        dyno += t * t;
      }
      dyno = sqrt(dyno / ND);
      if (newton_count > 1 && newton_count < newton_its)
      {
        double thq = dyno / dynold;
        if (newton_count == 2)
          theta = thq;
        else
          theta = sqrt(thq * thqold);
        thqold = thq;
        if (theta < 0.99)
        {
          double dyth;
          faccon = theta / (1.0 - theta);
          dyth = faccon * dyno *
                 pow(theta, (newton_its - 1 - newton_count)) / fnewt;
          if (dyth >= 1.0)
          {
            double qnewt = max(1.0E-4, min(20.0, dyth));
            hhfac = 0.8 * pow(qnewt, -1.0 / (4.0 + newton_its - 1 - newton_count));
            step_size *= hhfac;
            reject = PR_TRUE;
            last = PR_FALSE;
            if (hhfac <= 0.5)
              unexn = PR_TRUE;
            return;
          }
        }
        else
        {
          step_rejection();
          return;
        }
      }
      dynold = max(dyno, ROUNDING_UNIT);
      for (i = 0; i < ND; i++)
      {
        double f1i = ff[i] + zz[i];
        ff[i] = f1i;
        zz[i] = f1i;
      }
      if (faccon * dyno > fnewt)
        continue;
      break;
    }
    
    do_estrav();

    if (!got_nan && err < 1.0)
      for (i = NS; i < ND + NS; i++)
      {
        all_yvalues[i] += zz[i - NS];
        cont[i + NS + ND] = zz[i - NS];
      }
  }

  fac = min(SAFETY_FACTOR,
            (1 + 2 * newton_its) * SAFETY_FACTOR /
            (newton_count + 2 * newton_its));

  if (!got_nan)
  {
    quot = max(R_FACTOR, min(L_FACTOR, pow(err, expo) / fac));

    step_size_new = step_size / quot;
  }
  else
    step_size_new = step_size / L_FACTOR;

  if (!got_nan && err < 1.0)
  {
    /* Step was accepted... */
    first = PR_FALSE;
    if (!change)
    {
      if (naccpt++ > 1)
      {
        double facgus = (step_size_acc / step_size) *
                         pow((err * err) / erracc, expo) / SAFETY_FACTOR;
        facgus = max(R_FACTOR, min(L_FACTOR, facgus));
        quot = max(quot, facgus);
        step_size_new = step_size / quot;
      }
      step_size_acc = step_size;
      erracc = max(0.01, err);
    }
    xold = x;
    step_size_old = step_size;
    x = xph;
    
    /* Update scaling... */
    for (i = 1; i <= ND; i++)
      scal[i - 1] = abs_tolerance_1 + rel_tolerance_1 * abs_d(all_yvalues[i + NS - 1]);

    EvaluateSimple(x, all_yvalues);
    for (i = 0; i < NS + ND; i++)
      cont[i] = all_yvalues[i];
    /* Use xend instead of x to overcome rounding issues on the last point. */
    add_points(last ? xend : x, all_yvalues);

    caljac = PR_FALSE;
    if (last)
    {
      step_size = step_size_opt;
      integration_done();
      return;
    }

    ComputeDifferentials(x, all_yvalues, y_0);
    step_size_new = posneg * min(abs_d(step_size_new), maximal_step_size);
    step_size_opt = min(step_size, step_size_new);
    if (reject)
      step_size_new = posneg * min(abs_d(step_size_new), abs_d(step_size));
    reject = PR_FALSE;
    if ((x + step_size_new / STEP_HOLDRATIO_MIN - xend) * posneg >= 0.0)
    {
      step_size = xend - x;
      last = PR_TRUE;
    }
    else
    {
      double qt = step_size_new / step_size;
      hhfac = step_size;
      if (theta <= THET && qt >= STEP_HOLDRATIO_MIN &&
          qt <= STEP_HOLDRATIO_MAX)
      {
        ikeep = 1;
        return;
      }
      step_size = step_size_new;
    }
    hhfac = step_size;
    if (theta <= THET)
    {
      skipjac = PR_TRUE;
      return;
    }
  }
  else
  {
    /* Step is rejected... */
    reject = PR_TRUE;
    last = PR_FALSE;
    if (first)
    {
      step_size *= 0.1;
      hhfac = 0.1;
    }
    else
    {
      hhfac = step_size_new / step_size;
      step_size = step_size_new;
    }
  }
}

#ifndef WIN32

int
matherr(struct exception *exc)
{
  char buf[1024];

  if (disable_debug)
    return 0;

  printf("In the matherr handler.\n");

  strcpy(buf, "mathError/");
  strncpy(buf + sizeof("mathError/") - 1,
          active_cmeta_id,
          1024 - sizeof("mathError/")
         );
  buf[1023] = 0;
  integration_failed(buf);
  longjmp(sigrecover, 1);
}
#endif

#ifndef WIN32
#include <signal.h>

static void
handle_trap(int signum, siginfo_t* si, void* context)
{
  char buf[1024];

  printf("In the math trap handler.\n");

  const char* msg = "mathError/";

  switch (si->si_code)
  {
  case FPE_INTDIV:
  case FPE_FLTDIV:
    msg = "divisionByZero/";
    break;
  case FPE_FLTUND:
    msg = "underflow/";
    break;
  case FPE_FLTOVF:
  case FPE_INTOVF:
    msg = "overflow/";
    break;
  default:
    msg = "unexpectedFloatingPointError/";
  }

  strcpy(buf, msg);
  strncpy(buf + strlen(msg),
          active_cmeta_id,
          1023 - strlen(msg)
         );
  buf[1023] = 0;
#ifdef DEBUG_MODE
  printf("In handle_trap\n");
#endif
  printf("Sending error: %s\n", buf);
  integration_failed(buf);
  longjmp(sigrecover, 1);
}
#endif

int
main(int argc, char** argv)
{
  PRUint32 i;
  PRUint32 sleep_time = PR_MillisecondsToInterval(SUSPEND_MILLIS);

  setvbuf(stderr, NULL, _IONBF, 0);

#ifndef STATIC_ARRAYS
  allocate_arrays();
#endif

  setup_io();

  init_helper();

#ifndef WIN32
  setjmp(sigrecover);
#endif

  if (!disable_debug)
  {
#ifndef WIN32
    sigset_t blocked;
    struct sigaction sa;
    sa.sa_sigaction = handle_trap;
    sa.sa_flags = SA_SIGINFO;
    sigemptyset(&sa.sa_mask);
    sigaddset(&sa.sa_mask, SIGFPE);
    feenableexcept(FE_DIVBYZERO | FE_OVERFLOW);
    sigaction(SIGFPE, &sa, NULL);
    sigemptyset(&blocked);
    sigprocmask(SIG_UNBLOCK, &blocked, NULL);
#endif
  }

  while (helper_state != DYING)
  {
    if (helper_state == SUSPENDED)
      PR_Sleep(sleep_time);
    else if (helper_state == RUNNING)
    {
      for (i = 0;
           i < CYCLES_BEFORE_POLL && helper_state == RUNNING;
           i++)
        CYCLE_FUNCTION();
    }
    check_input();
  }
  return 0;
}

static void
read_input_command()
{
  PRUint32 length;
  struct HelperCommandHeader *hdr;
  char* data;
  if (PR_Read(input, &length, 4) <= 0)
  {
    PR_ProcessExit(-3);
  }
  if (length > 16777216)
    length = 16777216;
  if (length < sizeof(struct HelperCommandHeader))
    length = sizeof(struct HelperCommandHeader);

  data = (char*)PR_Malloc(length);
  /* XXX alignment issues on some platforms? */
  *(PRUint32*)data = length;
  if (PR_Read(input, data + 4, length - 4) <= 0)
  {
    PR_ProcessExit(-3);
  }
  hdr = (struct HelperCommandHeader*)data;

  hdr->length -= sizeof(struct HelperCommandHeader);
  if (hdr->cmd == START_HELPER)
    cmd_start(hdr + 1, hdr->length);
  else if (hdr->cmd == SUSPEND_HELPER)
    cmd_suspend(hdr + 1, hdr->length);
  else if (hdr->cmd == STOP_HELPER)
    cmd_stop(hdr + 1, hdr->length);
  PR_Free(data);
}
