/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellML.h"
#include "mozCellMLVariable.h"
#include "mozCellMLVariableResolution.h"
#include "mozCellMLIntegrator.h"
#include "mozCellMLFitter.h"
#include "mozCellMLChannel.h"
#include "MathMLUtilities.h"
#include <nsIURL.h>
#include <necko/nsNetCID.h>
#include <nsIServiceManager.h>
#include <nsCOMArray.h>
#include <nsArray.h>
#include <nsArrayEnumerator.h>
#include <nsIDOMDocument.h>
#include <necko/nsIServerSocket.h>
#include <nsDirectoryService.h>
#include <nsPrintfCString.h>
#include <nsContentCID.h>
#include <nsISyncLoadDOMService.h>
#include <necko/nsNetUtil.h>
#include <nsIContent.h>
#include <nsCRT.h>
#include <nsClassHashtable.h>
#include "Icorba4xpcom.h"
#ifdef XP_WIN32
#include <windows.h>
#include <nsIThread.h>
#include "mozWinFixup.h"
#endif
#include <nsIRunnable.h>

NS_IMPL_ISUPPORTS2(mozCellML, mozICellML, mozICellMLLocal)

#define COMPLAIN(args) \
{\
  char* result = PR_smprintf args; \
  nsCAutoString complainCString(result); \
  nsAutoString complainString; \
  CopyUTF8toUTF16(complainCString, complainString); \
  PR_Free(result); \
  return ValidationError(complainString); \
}

#define T(msg) Translate(msg)

mozCellML::mozCellML(nsCAutoString ititle, PRBool ivalidate)
  : title(ititle),
    localValidationError(PR_FALSE),
    validationErrorMessage(NS_LITERAL_STRING("")),
    validate(ivalidate)
{
  LEAKTRACE_CREATED(mozCellML)
  NS_NewArray(getter_AddRefs(variableList));
  NS_NewArray(getter_AddRefs(allComponents));
  allVariables.Init();
  allUnits.Init();
  allVariableRes.Init();
  nonMainVariables.Init();
  importNumbers.Init();
  loadingImports.Init();
  componentAliases.Init();

  AddImplicitUnits();

  functionNumber = 0;
#ifdef XP_WIN
  corbaThread = new mozCellMLCORBAThread();
#endif
}

mozCellML::mozCellML()
  : localValidationError(PR_FALSE),
    validationErrorMessage(NS_LITERAL_STRING(""))
{
  LEAKTRACE_CREATED(mozCellML)
  NS_NewArray(getter_AddRefs(variableList));
  NS_NewArray(getter_AddRefs(allComponents));
  allVariables.Init();
  allUnits.Init();
  allVariableRes.Init();
  nonMainVariables.Init();
  importNumbers.Init();
  loadingImports.Init();
  componentAliases.Init();

  AddImplicitUnits();

  functionNumber = 0;
#ifdef XP_WIN
  corbaThread = new mozCellMLCORBAThread();
#endif
}

mozCellML::~mozCellML()
{
  LEAKTRACE_DELETED(mozCellML);
#ifdef XP_WIN32
  corbaThread->AutoDestruct();
#endif
}

NS_IMETHODIMP
mozCellML::GetTitle(char** titleOut)
{
  *titleOut = ToNewCString(title);
  if (!*titleOut)
    return NS_ERROR_OUT_OF_MEMORY;
  return NS_OK;
}

void
mozCellML::AddImplicitUnits()
{
  // Base units do not go on unitsList.
#define BASE_UNIT(x) \
  nsCOMPtr<mozCellMLUnits> x = \
    new mozCellMLUnits(); \
  allUnits.Put(NS_LITERAL_STRING(#x), x);

  BASE_UNIT(ampere);
  BASE_UNIT(candela);
  BASE_UNIT(kelvin);
  BASE_UNIT(kilogram);
  BASE_UNIT(metre);
  BASE_UNIT(mole);
  BASE_UNIT(second);
  BASE_UNIT(dimensionless);
#undef BASE_UNIT

  // American spelling...
  allUnits.Put(NS_LITERAL_STRING("meter"), metre);

  // Dimensionless bases...
  allUnits.Put(NS_LITERAL_STRING("radian"), dimensionless);
  allUnits.Put(NS_LITERAL_STRING("steradian"), dimensionless);
  nsCOMPtr<mozCellMLUnits> celsius = new mozCellMLUnits(kelvin, 1.0, 273.15);
  allUnits.Put(NS_LITERAL_STRING("celsius"), celsius);

  nsCOMPtr<mozCellMLUnits> hertz = new mozCellMLUnits(second, -1.0);
  allUnits.Put(NS_LITERAL_STRING("becquerel"), hertz);
  allUnits.Put(NS_LITERAL_STRING("hertz"), hertz);

  nsCOMPtr<mozCellMLUnits> newton
    = new mozCellMLUnits(kilogram, 1.0, metre, 1.0, second, -2.0);
  allUnits.Put(NS_LITERAL_STRING("newton"), newton);

  nsCOMPtr<mozCellMLUnits> upascal
    = new mozCellMLUnits(kilogram, 1.0, metre, -1.0, second, -2.0);
  allUnits.Put(NS_LITERAL_STRING("pascal"), upascal);

  nsCOMPtr<mozCellMLUnits> joule
    = new mozCellMLUnits(metre, 2.0, kilogram, 1.0, second, -2.0);
  allUnits.Put(NS_LITERAL_STRING("joule"), joule);

  nsCOMPtr<mozCellMLUnits> watt
    = new mozCellMLUnits(metre, 2.0, kilogram, 1.0, second, -3.0);
  allUnits.Put(NS_LITERAL_STRING("watt"), watt);

  nsCOMPtr<mozCellMLUnits> coulomb
    = new mozCellMLUnits(ampere, 1.0, second, 1.0);
  allUnits.Put(NS_LITERAL_STRING("coulomb"), coulomb);

  nsCOMPtr<mozCellMLUnits> volt
    = new mozCellMLUnits(metre, 2.0, kilogram, 1.0, second, -3.0,
                         ampere, -1.0);
  allUnits.Put(NS_LITERAL_STRING("volt"), volt);

  nsCOMPtr<mozCellMLUnits> farad
    = new mozCellMLUnits(metre, -2.0, kilogram, -1.0, second, 4.0,
                         ampere, 2.0);
  allUnits.Put(NS_LITERAL_STRING("farad"), farad);

  nsCOMPtr<mozCellMLUnits> ohm
    = new mozCellMLUnits(metre, 2.0, kilogram, 1.0, second, -3.0, ampere, -2.0);
  allUnits.Put(NS_LITERAL_STRING("ohm"), ohm);

  nsCOMPtr<mozCellMLUnits> siemens
    = new mozCellMLUnits(metre, -2.0, kilogram, -1.0, second, 3.0, ampere, 2.0);
  allUnits.Put(NS_LITERAL_STRING("siemens"), siemens);

  nsCOMPtr<mozCellMLUnits> weber
    = new mozCellMLUnits(metre, 2.0, kilogram, 1.0, second, -2.0, ampere, -1.0);
  allUnits.Put(NS_LITERAL_STRING("weber"), weber);

  nsCOMPtr<mozCellMLUnits> tesla
    = new mozCellMLUnits(kilogram, 1.0, second, -2.0, ampere, -1.0);
  allUnits.Put(NS_LITERAL_STRING("tesla"), tesla);

  nsCOMPtr<mozCellMLUnits> henry
    = new mozCellMLUnits(metre, 2.0, kilogram, 1.0, second, -2.0, ampere, -2.0);
  allUnits.Put(NS_LITERAL_STRING("henry"), henry);

  allUnits.Put(NS_LITERAL_STRING("lumen"), candela);
  
  nsCOMPtr<mozCellMLUnits> lux
    = new mozCellMLUnits(metre, -2.0, candela, 1.0);
  allUnits.Put(NS_LITERAL_STRING("lux"), lux);
  
  nsCOMPtr<mozCellMLUnits> gray
    = new mozCellMLUnits(metre, 2.0, second, -2.0);
  allUnits.Put(NS_LITERAL_STRING("gray"), gray);
  allUnits.Put(NS_LITERAL_STRING("sievert"), gray);

  nsCOMPtr<mozCellMLUnits> katal
    = new mozCellMLUnits(mole, 1.0, second, -1.0);
  allUnits.Put(NS_LITERAL_STRING("katal"), katal);

  nsCOMPtr<mozCellMLUnits> gram
    = new mozCellMLUnits(kilogram, 1.0, 0.0, 1E-3);
  allUnits.Put(NS_LITERAL_STRING("gram"), gram);

  nsCOMPtr<mozCellMLUnits> litre
    = new mozCellMLUnits(metre, 3.0, 0.0, 1E-3);
  allUnits.Put(NS_LITERAL_STRING("litre"), litre);
  allUnits.Put(NS_LITERAL_STRING("liter"), litre);

  nsCOMPtr<mozCellMLUnits> boolean = new mozCellMLUnits();
  allUnits.Put(NS_LITERAL_STRING("cellml:boolean"), boolean);
}

PRUint32
mozCellML::GetFunctionNumber()
{
  return functionNumber++;
}

nsresult
mozCellML::SetEarlyError(mozCellMLLoadError* le)
{
  if (early_error != nsnull)
  {
    /* delete le; */
    nsCOMPtr<mozCellMLLoadError> xle = le;
  }
  else
    early_error = le;
  return NS_ERROR_INVALID_ARG;
}

mozCellMLLoadError*
mozCellML::GetEarlyError()
{
  return early_error;
}

void
mozCellML::FindImportedComponent(nsAutoString& imp, nsAutoString& comp)
{
  while (PR_TRUE)
  {
    ComponentAlias* ca = nsnull;
    componentAliases.Get(imp + comp, &ca);
    if (ca == nsnull)
      return;
#ifdef DEBUG_MODE
    nsCAutoString coimp, cocomp, cnimp, cncomp;
    CopyUTF16toUTF8(imp, coimp);
    CopyUTF16toUTF8(comp, cocomp);
    CopyUTF16toUTF8(ca->importPrefix, cnimp);
    CopyUTF16toUTF8(ca->componentName, cncomp);
    printf("Following component alias: %s%s => %s%s\n",
           coimp.get(), cocomp.get(), cnimp.get(), cncomp.get());
#endif
    imp.Assign(ca->importPrefix);
    comp.Assign(ca->componentName);
  }
}

NS_IMETHODIMP
mozCellML::InitialiseUnits()
{
  PRUint32 i, l;
  nsCOMPtr<nsISimpleEnumerator> allEnum;

  // Go through every unit and initialise it. The initialisation actually
  // occurs in a depth first manner, so we are only looking at the roots...
  for (i = 0, l = unitsList.Count(); i < l; i++)
    if (!unitsList[i]->IsDone())
    {
      nsresult rv = unitsList[i]->Initialise(this, importNumbers, allUnits);
      NS_ENSURE_SUCCESS(rv, NS_OK);
    }

  variableList->Enumerate(getter_AddRefs(allEnum));
  variableList->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> varSupports;
    allEnum->GetNext(getter_AddRefs(varSupports));
    nsCOMPtr<mozCellMLVariable> var(do_QueryInterface(varSupports));
    nsresult rv = var->SetupUnits(this, importNumbers, allUnits);
    NS_ENSURE_SUCCESS(rv, NS_OK);
  }
  for (i = 0, l = nonMainList.Count(); i < l; i++)
  {
    nsresult rv = nonMainList[i]->SetupUnits(this, importNumbers, allUnits);
    NS_ENSURE_SUCCESS(rv, NS_OK);
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::InitialiseLocalConnections()
{
  PRUint32 i, l;

  // We need to search connections to link variables across components...
  nsCOMPtr<nsIDOMNodeList> dnl;
  nsCOMPtr<nsIArray> connections;
  docRoot->GetChildNodes(getter_AddRefs(dnl));
  nsCOMPtr<nsISimpleEnumerator> connectionsE;
  FilterNodeListByNameAndTypeNS2(
                                 dnl,
                                 getter_AddRefs(connections),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("connection")
                                );
  connections->GetLength(&l);
  connections->Enumerate(getter_AddRefs(connectionsE));

  for (
       i = 0;
       i < l;
       i++
      )
  {
    nsCOMPtr<nsISupports> connSupports;
    connectionsE->GetNext(getter_AddRefs(connSupports));
    nsCOMPtr<nsIDOMElement> conn(do_QueryInterface(connSupports));
    allConnections.AppendObject(conn);
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::MatchConnectedVariables()
{
  PRUint32 i, l;

  // Set up the promise tables and lists. These store variables that we cannot
  // find immediately, but which we need to find eventually.
  nsInterfaceHashtable<nsStringHashKey, mozCellMLVariablePromise>
    PromiseTable;
  PromiseTable.Init();
  nsCOMArray<mozCellMLVariablePromise> PromiseList;

  nsCOMPtr<nsIDOMNodeList> dnl;

  l = allConnections.Count();
  for (
       i = 0;
       i < l;
       i++
      )
  {
    nsIDOMElement* conn = allConnections[i];
    nsCOMPtr<nsIArray> mapcomps;
    conn->GetChildNodes(getter_AddRefs(dnl));
    FilterNodeListByNameAndTypeNS2(
                                   dnl,
                                   getter_AddRefs(mapcomps),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("map_components")
                                  );
    PRUint32 nmc, j;
    mapcomps->GetLength(&nmc);
    if (nmc == 0)
      continue;
    nsCOMPtr<nsIDOMElement> mapel;
    mapcomps->QueryElementAt(0, NS_GET_IID(nsIDOMElement),
                             getter_AddRefs(mapel));
    nsAutoString comp1, comp2;
    mapel->GetAttribute(NS_LITERAL_STRING("component_1"), comp1);
    mapel->GetAttribute(NS_LITERAL_STRING("component_2"), comp2);

    nsAutoString inp1, inp2;
    GetImportNumberPrefix(mapel, nsnull, inp1);
    inp2.Assign(inp1);
    // This will ensure we map imported components to the correct name...
    FindImportedComponent(inp1, comp1);
    FindImportedComponent(inp2, comp2);

    FilterNodeListByNameAndTypeNS2(
                                   dnl,
                                   getter_AddRefs(mapcomps),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("map_variables")
                                  );
    mapcomps->GetLength(&nmc);
    nsCOMPtr<nsISimpleEnumerator> varenum;
    mapcomps->Enumerate(getter_AddRefs(varenum));

    for (
         j = 0;
         j < nmc;
         j++
        )
    {
      nsCOMPtr<nsISupports> varSupports;
      varenum->GetNext(getter_AddRefs(varSupports));
      nsCOMPtr<nsIDOMElement> varel(do_QueryInterface(varSupports));
      nsAutoString var1, var2;
      varel->GetAttribute(NS_LITERAL_STRING("variable_1"), var1);
      varel->GetAttribute(NS_LITERAL_STRING("variable_2"), var2);
      nsAutoString map1(inp1 + comp1 + NS_LITERAL_STRING(":") + var1),
                   map2(inp2 + comp2 + NS_LITERAL_STRING(":") + var2);

#ifdef DEBUG_MODE
      {
        nsCAutoString cmap1, cmap2;
        CopyUTF16toUTF8(map1, cmap1);
        CopyUTF16toUTF8(map2, cmap2);
        printf("Found mapping from %s to %s\n", cmap1.get(), cmap2.get());
      }
#endif

      nsCOMPtr<mozCellMLVariableResolution> var;
      allVariableRes.Get(map1, getter_AddRefs(var));
      if (var == nsnull)
      {
        allVariableRes.Get(map2, getter_AddRefs(var));
        if (var != NULL)
        {
          allVariables.Put(map1, var->GetVariable());
          nsCOMPtr<mozCellMLVariableResolution> varConv;
          nsCOMPtr<mozCellMLVariable> nmvar;
          nonMainVariables.Get(map1, getter_AddRefs(nmvar));
          if (nmvar == nsnull)
          {
            nsCAutoString cvar, ccomp;
            CopyUTF16toUTF8(var1, cvar);
            CopyUTF16toUTF8(comp1, ccomp);
            COMPLAIN((T("missingVariable"), cvar.get(), ccomp.get()));
          }
          nsresult rv =
            var->GetConversion(this, nmvar, allResolutions,
                               getter_AddRefs(varConv));
          NS_ENSURE_SUCCESS(rv, NS_OK);
          allVariableRes.Put(map1, varConv);
        }
        else
        {
          nsCOMPtr<mozCellMLVariable> nmvar;
          nonMainVariables.Get(map1, getter_AddRefs(nmvar));

          if (nmvar == nsnull)
          {
            nsCAutoString cvar, ccomp;
            CopyUTF16toUTF8(var1, cvar);
            CopyUTF16toUTF8(comp1, ccomp);
            COMPLAIN((T("missingVariable"), cvar.get(), ccomp.get()));
          }
          nonMainVariables.Get(map2, getter_AddRefs(nmvar));
          if (nmvar == nsnull)
          {
            nsCAutoString cvar, ccomp;
            CopyUTF16toUTF8(var2, cvar);
            CopyUTF16toUTF8(comp2, ccomp);
            COMPLAIN((T("missingVariable"), cvar.get(), ccomp.get()));
          }

          // See if we have a promise...
          nsCOMPtr<mozCellMLVariablePromise> promise1, promise2;
          PromiseTable.Get(map1, getter_AddRefs(promise1));
          PromiseTable.Get(map2, getter_AddRefs(promise2));
          if (promise1 != nsnull && promise2 != nsnull)
            promise1->MergeIn(promise2);
          else if (promise1 == nsnull && promise2 == nsnull)
          {
            promise1 = new mozCellMLVariablePromise(map1, map2);
            PromiseTable.Put(map1, promise1);
            PromiseTable.Put(map2, promise1);
            PromiseList.AppendObject(promise1);
          }
          else if (promise1 != nsnull)
          {
            promise1->AddParty(map2, map1);
            PromiseTable.Put(map2, promise1);
          }
          else if (promise2 != nsnull)
          {
            promise2->AddParty(map1, map2);
            PromiseTable.Put(map1, promise2);
          }
        }
      }
      else
      {
        nsCOMPtr<mozCellMLVariableResolution> xvar2;
        allVariableRes.Get(map2, getter_AddRefs(xvar2));
        if (xvar2 != nsnull)
        {
          nsCAutoString cvar1, ccomp1, cvar2, ccomp2;
          CopyUTF16toUTF8(var1, cvar1);
          CopyUTF16toUTF8(comp1, ccomp1);
          CopyUTF16toUTF8(var2, cvar2);
          CopyUTF16toUTF8(comp2, ccomp2);
          COMPLAIN((T("duplicateConnection"), cvar1.get(), ccomp1.get(),
                     cvar2.get(), ccomp2.get()));
        }
        allVariables.Put(map2, var->GetVariable());
        nsCOMPtr<mozCellMLVariableResolution> varConv;
        nsCOMPtr<mozCellMLVariable> nmvar;
        nonMainVariables.Get(map2, getter_AddRefs(nmvar));
        if (nmvar == nsnull)
        {
          nsCAutoString cvar, ccomp;
          CopyUTF16toUTF8(var2, cvar);
          CopyUTF16toUTF8(comp2, ccomp);
          COMPLAIN((T("missingVariable"), cvar.get(), ccomp.get()));
        }
        nsresult rv = var->GetConversion(this, nmvar, allResolutions,
                                         getter_AddRefs(varConv));
        NS_ENSURE_SUCCESS(rv, NS_OK);
        allVariableRes.Put(map2, varConv);
      }
    }
  }

  // Now resolve all connection promises we made...
  nsresult rv;
  for (i = 0, l = PromiseList.Count(); i < l; i++)
  {
    rv = PromiseList.ObjectAt(i)
          ->Resolve(this, nonMainVariables,
                    allVariableRes, allVariables,
                    allResolutions);
    NS_ENSURE_SUCCESS(rv, NS_OK);
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::BuildEquationList(nsCOMArray<mozCellMLEquation>& eql,
                             nsCOMArray<mozCellMLEquation>& deql)
{
  PRUint32 i, j, k, l, m, n;

  nsCOMPtr<nsISimpleEnumerator> enComp;
  allComponents->GetLength(&l);
  allComponents->Enumerate(getter_AddRefs(enComp));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> cnSupports;
    enComp->GetNext(getter_AddRefs(cnSupports));
    nsCOMPtr<nsIDOMElement> cnEl(do_QueryInterface(cnSupports));

    nsAutoString inp;
    GetImportNumberPrefix(cnEl, nsnull, inp);
    nsAutoString compName;
    cnEl->GetAttribute(NS_LITERAL_STRING("name"), compName);
    compName = inp + compName;

    nsCOMPtr<nsIDOMNodeList> cnElAllChildren;
    cnEl->GetChildNodes(getter_AddRefs(cnElAllChildren));
    nsCOMPtr<nsIArray> mathNodes;
    FilterNodeListByNameAndTypeNS(
                                  cnElAllChildren,
                                  getter_AddRefs(mathNodes),
                                  nsIDOMNode::ELEMENT_NODE,
                                  MATHML_NS,
                                  NS_LITERAL_STRING("math")
                                 );
    nsCOMPtr<nsISimpleEnumerator> enMathNodes;
    mathNodes->Enumerate(getter_AddRefs(enMathNodes));
    mathNodes->GetLength(&m);
    for (j = 0; j < m; j++)
    {
      nsCOMPtr<nsISupports> mnSupports;
      enMathNodes->GetNext(getter_AddRefs(mnSupports));
      nsCOMPtr<nsIDOMElement> mn(do_QueryInterface(mnSupports));
      // We now have to loop over apply nodes...
      nsCOMPtr<nsIDOMNodeList> mathElAllChildren;
      mn->GetChildNodes(getter_AddRefs(mathElAllChildren));
      nsCOMPtr<nsIArray> applyNodes;
      FilterNodeListByNameAndTypeNS(
                                    mathElAllChildren,
                                    getter_AddRefs(applyNodes),
                                    nsIDOMNode::ELEMENT_NODE,
                                    MATHML_NS,
                                    NS_LITERAL_STRING("apply")
                                   );
      nsCOMPtr<nsISimpleEnumerator> enApplyNodes;
      applyNodes->Enumerate(getter_AddRefs(enApplyNodes));
      applyNodes->GetLength(&n);
      for (k = 0; k < n; k++)
      {
        nsCOMPtr<nsISupports> applySupports;
        enApplyNodes->GetNext(getter_AddRefs(applySupports));
        nsCOMPtr<nsIDOMElement> applyNode(do_QueryInterface(applySupports));
        nsCOMPtr<mozCellMLEquation> eq =
          new mozCellMLEquation(this, applyNode, allVariableRes,
                                compName);
        if (eq->GetEquationType() == mozCellMLEquation::MATHEMATICAL)
          eql.AppendObject(eq);
        else
          deql.AppendObject(eq);
      }
    }
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::ComputeAssignmentOrder
(
 nsCOMArray<mozCellMLEquation>& allEquations,
 nsCOMArray<mozCellMLEquation>& allDEquations,
 nsCOMArray<mozCellMLVariable>& varUndef,
 nsCOMArray<mozCellMLVariable>& varBlacklist
)
{
  PRUint32 i, l;
  // Next build up a list of "computed constants", i.e. values which can be
  // computed using only constants and computed constants.
  PRBool didWork;
  do
  {
    didWork = PR_FALSE;
    // Search the equation list for equations that involve only one blacklisted
    // variable...
    for (i = 0, l = allEquations.Count(); i < l;)
    {
      PRUint32 count = allEquations[i]->CountInvolving(varBlacklist);
      if (count == 0)
      {
        // We have an equation we don't need.
        nsAutoString eqid;
        allEquations[i]->GetIdentifier(eqid);
        nsCAutoString ceqid;
        CopyUTF16toUTF8(eqid, ceqid);
        COMPLAIN((T("overconstrainedModel"), ceqid.get()));
      }
      else if (count == 1)
      {
        // We have an equation involving only one variable we don't yet know.
        // See if this is a simple variable...
        mozCellMLVariable* v = allEquations[i]->GetInvolving(varBlacklist);
        if (v->type != mozICellMLVariable::SIMPLE)
        {
          i++;
          continue;
        }
        if (i > 0)
          didWork = PR_TRUE;
        // We have found a simple variable to solve.
        nsCOMPtr<mozCellMLEvaluation> ev;
        nsresult ret =
          allEquations[i]->ProduceEvaluation(v, getter_AddRefs(ev));
        NS_ENSURE_SUCCESS(ret, ret);
        computedConstantEvaluations.AppendObject
        (
         ev
        );
        // The equation has now been used, remove it...
        allEquations.RemoveObjectAt(i);
        l--;

        // The variable is now known, and so can be removed from the undef and
        // blacklist lists...
        varUndef.RemoveObject(v);
        varBlacklist.RemoveObject(v);

        continue;
      }
      else
        i++;
    }
  }
  while (didWork);

  // Next, we expand the list to include those that need the bound variable, or
  // one of the differentials, to be computed...
  do
  {
    didWork = PR_FALSE;
    // Search the equation list for equations that involve only one undefined
    // variable.
    for (i = 0, l = allEquations.Count(); i < l;)
    {
      PRUint32 count = allEquations[i]->CountInvolving(varUndef);
      if (count == 0)
      {
        // We have an equation we don't need.
        nsAutoString eqid;
        allEquations[i]->GetIdentifier(eqid);
        nsCAutoString ceqid;
        CopyUTF16toUTF8(eqid, ceqid);
        COMPLAIN((T("overconstrainedModel"), ceqid.get()));
      }
      else if (count == 1)
      {
        // We have an equation involving only one variable we don't yet know.
        mozCellMLVariable* v = allEquations[i]->GetInvolving(varUndef);
        if (i > 0)
          didWork = PR_TRUE;
        // We have found a simple variable to solve.
        nsCOMPtr<mozCellMLEvaluation> ev;
        nsresult ret = allEquations[i]->ProduceEvaluation(v, getter_AddRefs(ev));
        NS_ENSURE_SUCCESS(ret, ret);
        perStepEvaluations.AppendObject(ev);
        // The equation has now been used, remove it...
        allEquations.RemoveObjectAt(i);
        l--;

        // The variable is now known, and so can be removed from the undef list...
        varUndef.RemoveObject(v);
        continue;
      }
      else
        i++;
    }
  }
  while (didWork);

  // Differentials order is arbitrary, so do it the simple way...
  nsCOMPtr<mozCellMLEvaluation> ev;
  nsCOMPtr<mozCellMLVariableResolution> vr;
  PRUint32 j;
  for (i = 0; i < (PRUint32)differentials.Count(); i++)
  {
    for (j = 0; j < (PRUint32)allDEquations.Count(); j++)
    {
      mozCellMLEquation::EquationVariableRelation evr;
      nsresult ret =
        allDEquations[j]->GetVariableRelation
        (
         differentials[i], evr, getter_AddRefs(vr)
        );
      NS_ENSURE_SUCCESS(ret, ret);
      if (evr == mozCellMLEquation::DIFF_VAR)
      {
        nsresult ret = allDEquations[j]->ProduceEvaluation(differentials[i], getter_AddRefs(ev));
        NS_ENSURE_SUCCESS(ret, ret);
        differentialEvaluations.AppendObject(ev);        
        break;
      }
    }
  }

  return NS_OK;
}

NS_IMETHODIMP
mozCellML::Initialise()
{
  PRUint32 i, l;
  nsCOMPtr<nsISimpleEnumerator> allEnum;
  if (docRoot == nsnull)
  {
#ifdef DEBUG_MODE
    printf("Initialise called but document is null.\n");
#endif
    return NS_ERROR_FAILURE;
  }

  nsresult ret;

  ret = FindVariableElements(docRoot);
  NS_ENSURE_SUCCESS(ret, ret);

  ret = InitialiseUnits();
  NS_ENSURE_SUCCESS(ret, ret);

  ret = InitialiseLocalConnections();
  NS_ENSURE_SUCCESS(ret, ret);

  ret = MatchConnectedVariables();
  NS_ENSURE_SUCCESS(ret, ret);

  nsCOMArray<mozCellMLEquation> allNondiffEquations, allDiffEquations;
  ret = BuildEquationList(allNondiffEquations, allDiffEquations);
  NS_ENSURE_SUCCESS(ret, ret);

  // Next initialise all variables...
  variableList->Enumerate(getter_AddRefs(allEnum));
  variableList->GetLength(&l);
  nsCOMPtr<nsIArray> allComponentsA(do_QueryInterface(allComponents));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> vSupports;
    allEnum->GetNext(getter_AddRefs(vSupports));
    nsCOMPtr<mozCellMLVariable> v(do_QueryInterface(vSupports));
    v->Initialise(allDiffEquations, allNondiffEquations);
  }

  nsCOMPtr<nsISimpleEnumerator> varEnum;
  nsCOMPtr<mozCellMLVariable> var;

  // Notify all variables of the units on the bound variable in that component.
  // This is needed because the bound variable may not be in the same units in
  // all equations. This information is used to compute differentials in the
  // correct units.
  for (i = 0, l = allResolutions.Count(); i < l; i++)
  {
    allResolutions[i]->BoundInformSiblings(this, allVariableRes);
  }

  variableList->Enumerate(getter_AddRefs(varEnum));
  PRBool hme;
  nsCOMPtr<mozCellMLVariable> bound;
  while (varEnum->HasMoreElements(&hme), hme)
  {
    nsCOMPtr<nsISupports> varSupports;
    varEnum->GetNext(getter_AddRefs(varSupports));
    var = do_QueryInterface(varSupports);
    PRUint32 type;
    var->GetType(&type);
    if (type == mozICellMLVariable::BOUND)
    {
      bound = var;
      var->SetVariableName(NS_LITERAL_CSTRING("y"));
      var->SetCode(-1);
    }
    else if (type == mozICellMLVariable::DIFFERENTIAL)
      differentials.AppendObject(var);
    else if (type == mozICellMLVariable::SIMPLE)
      simples.AppendObject(var);
    else if (type == mozICellMLVariable::CONSTANT)
      constants.AppendObject(var);
  }

  // Build an initial array of variables that need to be computed...
  nsCOMArray<mozCellMLVariable> varUndef = simples;
  nsCOMArray<mozCellMLVariable> varBlacklist = simples;
  for (i = 0, l = differentials.Count(); i < l; i++)
    varBlacklist.AppendObject(differentials[i]);
  if (bound != nsnull)
    varBlacklist.AppendObject(bound);

  ComputeAssignmentOrder(allNondiffEquations, allDiffEquations, varUndef,
                         varBlacklist);

  // If the maths were right, we should have used every equation, and left no
  // variables undefined.
  PRBool allEquationsUsed = (allNondiffEquations.Count() == 0);
  PRBool allVariablesDefined = (varUndef.Count() == 0);

  if (!allEquationsUsed && !allVariablesDefined)
  {
    nsXPIDLString varName, compName;
    varUndef[0]->GetName(getter_Copies(varName));
    varUndef[0]->GetComponentName(getter_Copies(compName));
    nsCAutoString cvarName, ccompName;
    CopyUTF16toUTF8(varName, cvarName);
    CopyUTF16toUTF8(compName, ccompName);
    COMPLAIN((T("cantSolveSystems"), cvarName.get(), ccompName.get()));
  }
  // Impossible: else if (!allEquationsUsed && allVariablesDefined)
  else if (allEquationsUsed && !allVariablesDefined)
  {
    nsXPIDLString varName, compName;
    varUndef[0]->GetName(getter_Copies(varName));
    varUndef[0]->GetComponentName(getter_Copies(compName));
    nsCAutoString cvarName, ccompName;
    CopyUTF16toUTF8(varName, cvarName);
    CopyUTF16toUTF8(compName, ccompName);
    COMPLAIN((T("underconstrainedModel"), cvarName.get(), ccompName.get()));
  }

  // If we get here, there is a feasible way to evaluate the mathematics.
  // Assign each variable a name...
  PRUint32 lt;
  for (
       i = 0, l = simples.Count();
       i < l;
       i++
      )
  {
    nsPrintfCString name("x[%u]", i);
    simples[i]->SetVariableName(name);
    simples[i]->SetCode(i);
  }
  for (lt = l + differentials.Count();
       i < lt;
       i++)
  {
    nsPrintfCString name("x[%u]", i);
    differentials[i - l]->SetVariableName(name);

    nsPrintfCString rname("dx[%u]", (i - l));
    differentials[i - l]->SetVariableRateName(rname);

    differentials[i - l]->SetCode(-2 - i + l);
  }

  for (i = 0, l = constants.Count();
       i < l;
       i++)
  {
    nsPrintfCString name("constants[%u]", i);
    constants[i]->SetVariableName(name);
  }

  // Finally, set up for higher derivatives...
  // We have to find all the "hidden higher" variables...
  PRUint32 higherDifferentialSpaceDX = differentials.Count();
  PRUint32 higherDifferentialSpaceX =
    higherDifferentialSpaceDX + simples.Count();
  l = differentialEvaluations.Count();
  for (i = 0; i < l; i++)
  {
    differentialEvaluations[i]->
      AllocateDifferentialSpace(higherDifferentialSpaceDX,
                                higherDifferentialSpaceX);
  }
  mHiddenHigher = higherDifferentialSpaceDX - differentials.Count();

  return NS_OK;
}

nsresult
mozCellML::PopulateModelDescription(nsIDOMElement* modelDescription)
{
  if (docRoot == nsnull)
    return NS_ERROR_NOT_INITIALIZED;

  if (modelDescription == nsnull)
    return NS_ERROR_INVALID_ARG;

  nsAutoString ctitle;
  CopyUTF8toUTF16(title, ctitle);
  modelDescription->SetAttribute(NS_LITERAL_STRING("title"), ctitle);

  nsCOMPtr<nsIDOMNode> drc;
  docRoot->CloneNode(true, getter_AddRefs(drc));
  nsCOMPtr<nsIDOMNode> drcc;
  modelDescription->AppendChild(drc, getter_AddRefs(drcc));

  PRUint32 i;

  for (i = 0; i < (PRUint32)importedRootList.Count(); i++)
  {
    nsCOMPtr<nsIDOMNode> irc;
    importedRootList.ObjectAt(i)->CloneNode(true, getter_AddRefs(irc));

    modelDescription->AppendChild(irc, getter_AddRefs(drcc));
  }

  return NS_OK;
}

nsresult
mozCellML::RestoreModelDescription(nsIDOMElement* modelDescription)
{
  if (docRoot != nsnull)
    return NS_ERROR_FAILURE;

  if (modelDescription == nsnull)
    return NS_ERROR_FAILURE;

  nsCOMPtr<nsIDOMElement> mainElement;
  nsInterfaceHashtable<nsStringHashKey,nsIDOMElement> importElements;

  nsAutoString ctitle;
  modelDescription->GetAttribute(NS_LITERAL_STRING("title"), ctitle);
  CopyUTF16toUTF8(ctitle, title);

  nsCOMPtr<nsIDOMNodeList> nl;
  modelDescription->GetChildNodes(getter_AddRefs(nl));
  nsCOMPtr<nsIArray> allModels;
  FilterNodeListByNameAndTypeNS2(
                                 nl,
                                 getter_AddRefs(allModels),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("model")
                                );
  PRUint32 l;
  nsCOMPtr<nsISimpleEnumerator> modelE;
  allModels->GetLength(&l);
  allModels->Enumerate(getter_AddRefs(modelE));

  if (l < 1)
    return NS_ERROR_FAILURE;

  nsCOMPtr<nsISupports> modelSupports;
  modelE->GetNext(getter_AddRefs(modelSupports));
  mainElement = do_QueryInterface(modelSupports);

  PRUint32 i;
  for (i = 1; i < l; i++)
  {
    modelE->GetNext(getter_AddRefs(modelSupports));
    nsCOMPtr<nsIDOMElement> model(do_QueryInterface(modelSupports));
    nsAutoString str;
    model->GetAttributeNS(
                          NS_LITERAL_STRING("http://www.w3.org/1999/xlink"),
                          NS_LITERAL_STRING("href"),
                          str
                         );
    importElements.Put(str, model);
  }

  // Now we simply need to load the model...
  SetDOM(mainElement);
  nsCOMPtr<nsIMutableArray> allNeededImports;

  nsresult rv;


  while (true)
  {
    rv = FetchImports(getter_AddRefs(allNeededImports));
    NS_ENSURE_SUCCESS(rv, rv);

    PRUint32 nremaining = 0;
    allNeededImports->GetLength(&nremaining);
    if (nremaining == 0)
      break;

    nsCOMPtr<nsISimpleEnumerator> importE;
    allNeededImports->Enumerate(getter_AddRefs(importE));

    for (i = 0; i < nremaining; i++)
    {
      nsCOMPtr<nsISupports> importSupport;
      importE->GetNext(getter_AddRefs(importSupport));
      nsISupports* iS = importSupport;
      mozCellMLImportRequest* mcir =
        NS_STATIC_CAST(mozCellMLImportRequest*,
                       NS_STATIC_CAST(nsIChannel*, iS));
      nsCOMPtr<nsIDOMElement> node;
      mcir->GetImportNode(getter_AddRefs(node));

      nsAutoString str;
      node->GetAttributeNS(
                           NS_LITERAL_STRING("http://www.w3.org/1999/xlink"),
                           NS_LITERAL_STRING("href"),
                           str
                          );
      nsCOMPtr<nsIDOMElement> importModel;
      rv = importElements.Get(str, getter_AddRefs(importModel));
      NS_ENSURE_SUCCESS(rv, rv);

      ImportLoaded(node, importModel);
    }
  }

  return NS_OK;
}

nsresult
mozCellML::ReadDocument(const PRUnichar* docURL, PRUnichar** errorMessage)
{
  nsCOMPtr<nsIURI> uri;
  nsCOMPtr<nsIChannel> chan;
#ifdef DEBUG_MODE
  printf("Making a new URI...\n");
#endif
  nsAutoString docURLc(docURL);
  nsresult rv = NS_NewURI(getter_AddRefs(uri), docURLc);
  NS_ENSURE_SUCCESS(rv, rv);
#ifdef DEBUG_MODE
  printf("Making a new channel...\n");
#endif
  rv = NS_NewChannel(getter_AddRefs(chan), uri);
  nsCOMPtr<nsISyncLoadDOMService>
    slds(do_GetService("@mozilla.org/content/syncload-dom-service;1"));
#ifdef DEBUG_MODE
  printf("Loading document...\n");
#endif
  nsCOMPtr<nsIDOMDocument> doc;
  rv = slds->LoadDocument(chan, nsnull, getter_AddRefs(doc));
  doc->GetDocumentElement(getter_AddRefs(docRoot));

  nsCOMPtr<nsIDOMDocument> idoc;

  *errorMessage = NULL;
  if (rv != NS_OK)
  {
#ifdef DEBUG_MODE
    printf("Loading document failed: %08X\n", rv);
#endif
    if (rv != NS_OK)
    {
      switch (rv)
      {
      case NS_ERROR_FAILURE:
        *errorMessage = ToNewUnicode(NS_LITERAL_STRING("Could not load file(generic failure)"));
        break;
      default:
        *errorMessage = ToNewUnicode(NS_LITERAL_STRING("Could not load file."));
      }
      return NS_OK;
    }
    return rv;
  }

#ifdef DEBUG_MODE
  printf("Calling SetDOM...\n");
#endif
  SetDOM(docRoot);

#ifdef USE_CELLML_API
  model.fromNode(docRoot);
#endif

  nsCOMPtr<nsILoadGroup> lg;
  chan->GetLoadGroup(getter_AddRefs(lg));
  if (lg == nsnull)
  {
    lg = do_CreateInstance("@mozilla.org/network/load-group;1");
    chan->SetLoadGroup(lg);
  }

  nsCOMPtr<nsIMutableArray> allImports;
  do
  {
#ifdef DEBUG_MODE
    printf("Calling FetchImports...\n");
#endif

    rv = FetchImports(getter_AddRefs(allImports));
    if (rv != NS_OK)
    {
      *errorMessage = ToNewUnicode(NS_LITERAL_STRING("Could not fetch imports."));
      return NS_OK;
    }

    PRUint32 remainingImports;
    allImports->GetLength(&remainingImports);
    if (remainingImports == 0)
      break;

    nsCOMPtr<nsISimpleEnumerator> importE;
    allImports->Enumerate(getter_AddRefs(importE));
    
    PRBool moreEls;
    while (importE->HasMoreElements(&moreEls), moreEls)
    {
      nsCOMPtr<nsISupports> importSupport;
      importE->GetNext(getter_AddRefs(importSupport));

      nsISupports* iS = importSupport;
      mozCellMLImportRequest* mcir =
        NS_STATIC_CAST(mozCellMLImportRequest*,
                       NS_STATIC_CAST(nsIChannel*, iS));
      nsCOMPtr<nsIDOMElement> node;
      mcir->GetImportNode(getter_AddRefs(node));
      
      mcir->SetLoadGroup(lg);
#ifdef DEBUG_MODE
      printf("Loading import...\n");
#endif
      rv = slds->LoadDocument(mcir, nsnull, getter_AddRefs(idoc));
      if (rv != NS_OK)
      {
#ifdef DEBUG_MODE
        printf("Loading import document failed: %08X\n", rv);
#endif
        return rv;
      }
      nsCOMPtr<nsIDOMElement> idocRoot;
      idoc->GetDocumentElement(getter_AddRefs(idocRoot));
      if (idocRoot == nsnull)
        return NS_ERROR_FAILURE;
      else
        rv = ImportLoaded(node, idocRoot);
      NS_ENSURE_SUCCESS(rv, rv);
    }
    allImports = nsnull;
  }
  while (PR_TRUE);

#ifdef DEBUG_MODE
  printf("Calling Initialise...\n");
#endif
  rv = Initialise();
  NS_ENSURE_SUCCESS(rv, rv);

#ifdef DEBUG_MODE
  printf("Load complete.\n");
#endif
  *errorMessage = ToNewUnicode(NS_LITERAL_STRING(""));

  return NS_OK;
}

nsresult
mozCellML::SetDOM(nsIDOMElement* domRoot)
{
  docRoot = domRoot;
  if (docRoot == nsnull)
    return NS_ERROR_FAILURE;
#ifdef USE_CELLML_API
  model.fromNode(docRoot);
#endif

  ExtractUnitsFrom(0, docRoot);

  allImports = nsnull;

  NS_NewArray(getter_AddRefs(allImports));

  importURLTable.AppendString(NS_LITERAL_STRING("Base"));
  nextHashNumber = 1;
  importNumbers.Put(docRoot, 0);

  ImportAll(docRoot);
  return NS_OK;
}

nsresult
mozCellML::DescribeImportNumber(PRUint32 impNo, PRUnichar** description)
{
  *description = ToNewUnicode(*importURLTable[impNo]);
  return NS_OK;
}

nsresult
mozCellML::GetDOM(nsIDOMElement** dom)
{
  CallQueryInterface(docRoot, dom);
  return NS_OK;
}

void
mozCellML::RequestLoadImport(nsIDOMElement* importNode)
{
  PRBool isloading = PR_FALSE;
  loadingImports.Get(importNode, &isloading);
  if (isloading)
    return;
  loadingImports.Put(importNode, PR_TRUE);

  nsCOMPtr<nsISupports> mcir =
    NS_STATIC_CAST(nsIChannel*,
                   new mozCellMLImportRequest(importNode));
  allImports->AppendElement(mcir, PR_FALSE);
}

nsresult
mozCellML::FetchImports(nsIMutableArray** xallImports)
{
  *xallImports = allImports;
  NS_ADDREF(*xallImports);

  allImports = nsnull;

  NS_NewArray(getter_AddRefs(allImports));
  return NS_OK;
}

PRUint32
mozCellML::SearchDeeplyForEncapsulations(
                                         nsIDOMElement* searchIn,
                                         nsInterfaceHashtable<nsStringHashKey,
                                                              nsIDOMElement>
                                           &encapsulations
                                        )
{
  nsCOMPtr<nsIDOMNodeList> cl;
  searchIn->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> allComponents;
  FilterNodeListByNameAndTypeNS2(
                                 cl,
                                 getter_AddRefs(allComponents),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("component_ref")
                                );
  PRUint32 i, l;
  allComponents->GetLength(&l);

  nsCOMPtr<nsISimpleEnumerator> allComponentsE;
  allComponents->Enumerate(getter_AddRefs(allComponentsE));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> componentSupports;
    allComponentsE->GetNext(getter_AddRefs(componentSupports));
    nsCOMPtr<nsIDOMElement> componentEl(do_QueryInterface(componentSupports));
    nsAutoString compName;
    componentEl->GetAttribute(NS_LITERAL_STRING("component"), compName);
    if (SearchDeeplyForEncapsulations(componentEl, encapsulations) != 0)
      encapsulations.Put(compName, componentEl);
  }
  return l;
}

PRBool
mozCellML::ComponentIsImported(nsAString& compName,
                               nsAString& impName,
                               nsIDOMElement* root)
{
  nsCOMPtr<nsIDOMNodeList> nl;
  root->GetChildNodes(getter_AddRefs(nl));
  nsCOMPtr<nsIArray> imports;
  FilterNodeListByNameAndTypeNS(
                                nl, getter_AddRefs(imports),
                                nsIDOMNode::ELEMENT_NODE,
                                CELLML_NS_1_1,
                                NS_LITERAL_STRING("import")
                               );
  PRUint32 i, l, j, m;
  imports->GetLength(&l);
  nsCOMPtr<nsISimpleEnumerator> e;
  imports->Enumerate(getter_AddRefs(e));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> impSupports;
    e->GetNext(getter_AddRefs(impSupports));
    nsCOMPtr<nsIDOMElement> impEl(do_QueryInterface(impSupports));
    impEl->GetChildNodes(getter_AddRefs(nl));
    nsCOMPtr<nsIArray> components;
    FilterNodeListByNameAndTypeNS(nl,
                                  getter_AddRefs(components),
                                  nsIDOMNode::ELEMENT_NODE,
                                  CELLML_NS_1_1,
                                  NS_LITERAL_STRING("component"));
    components->GetLength(&m);
    nsCOMPtr<nsISimpleEnumerator> f;
    components->Enumerate(getter_AddRefs(f));
    for (j = 0; j < m; j++)
    {
      nsCOMPtr<nsISupports> compSupports;
      f->GetNext(getter_AddRefs(compSupports));
      nsCOMPtr<nsIDOMElement> compEl(do_QueryInterface(compSupports));
      nsAutoString xcompName;
      compEl->GetAttribute(NS_LITERAL_STRING("name"), xcompName);
      if (!compName.Equals(xcompName))
        continue;
      compEl->GetAttribute(NS_LITERAL_STRING("component_ref"), impName);
      return PR_TRUE;
    }
  }
  return PR_FALSE;
}

nsresult
mozCellML::ExpandEncapsulatedList(
                                  nsAString& inp,
                                  nsCOMArray<nsIDOMElement>& ComponentStack,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       nsIDOMElement>& AllComponents,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       nsIDOMElement>& imported,
                                  nsClassHashtable<nsStringHashKey,nsString>& deeperImported,
                                  nsIDOMElement* importedRoot
                                 )
{
  // We have to do a deep search for component_refs and add them to
  // a hashtable by component name...
  nsInterfaceHashtable<nsStringHashKey,nsIDOMElement> encapsulations;
  encapsulations.Init();
  nsCOMPtr<nsIDOMNodeList> cl;
  importedRoot->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> allGroups;
  FilterNodeListByNameAndTypeNS2(
                                 cl,
                                 getter_AddRefs(allGroups),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("group")
                                );
  nsCOMPtr<nsISimpleEnumerator> allGroupsE;
  allGroups->Enumerate(getter_AddRefs(allGroupsE));
  PRUint32 i, l;
  allGroups->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> groupSupports;
    allGroupsE->GetNext(getter_AddRefs(groupSupports));
    nsCOMPtr<nsIDOMElement> groupEl(do_QueryInterface(groupSupports));
    PRBool encapsulation = PR_FALSE;
    groupEl->GetChildNodes(getter_AddRefs(cl));
    nsCOMPtr<nsIArray> allRelationships;
    FilterNodeListByNameAndTypeNS2(
                                   cl,
                                   getter_AddRefs(allRelationships),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("relationship_ref")
                                  );
    nsCOMPtr<nsISimpleEnumerator> allRelationshipsE;
    allRelationships->Enumerate(getter_AddRefs(allRelationshipsE));
    PRUint32 m, j;
    allRelationships->GetLength(&m);
    for (j = 0; j < m; j++)
    {
      nsCOMPtr<nsISupports> relSupports;
      allRelationshipsE->GetNext(getter_AddRefs(relSupports));
      nsCOMPtr<nsIDOMElement> relEl(do_QueryInterface(relSupports));
      nsAutoString relationship;
      relEl->GetAttribute(NS_LITERAL_STRING("relationship"), relationship);
      if (relationship.Equals(NS_LITERAL_STRING("encapsulation")))
      {
        encapsulation = PR_TRUE;
        break;
      }
    }
    if (!encapsulation)
      continue;
    SearchDeeplyForEncapsulations(groupEl, encapsulations);
  }

  nsCOMPtr<nsIAtom> seenAtom = NS_NewAtom("seen_for_import");

  // Keep processing things on the stack...
  while ((l = ComponentStack.Count()) > 0)
  {
    nsIDOMElement* componentEl = ComponentStack[l - 1];
    ComponentStack.RemoveObjectAt(l - 1);
    // Get the name...
    nsAutoString compName;
    componentEl->GetAttribute(NS_LITERAL_STRING("name"), compName);

    // Check the hashtable...
    nsCOMPtr<nsIDOMElement> compRefEl;
    encapsulations.Get(compName, getter_AddRefs(compRefEl));
    if (compRefEl == nsnull)
      continue;
    // We have other components encapsulated in this one...
    compRefEl->GetChildNodes(getter_AddRefs(cl));
    nsCOMPtr<nsIArray> allSubComponents;
    FilterNodeListByNameAndTypeNS2(
                                   cl,
                                   getter_AddRefs(allSubComponents),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("component_ref")
                                  );
    nsCOMPtr<nsISimpleEnumerator> allSubsE;
    allSubComponents->Enumerate(getter_AddRefs(allSubsE));
    allSubComponents->GetLength(&l);
    PRUint32 i;
    for (i = 0; i < l; i++)
    {
      nsCOMPtr<nsISupports> compSupports;
      allSubsE->GetNext(getter_AddRefs(compSupports));
      nsCOMPtr<nsIDOMElement> compRefEl(do_QueryInterface(compSupports));
      compRefEl->GetAttribute(NS_LITERAL_STRING("component"), compName);

      // Find the real component...
      nsCOMPtr<nsIDOMElement> component;
      AllComponents.Get(compName, getter_AddRefs(component));
      // We have already requested a recursive import. Check to see if this
      // component will be imported. If it is, we put it on the import
      // connection list.
      nsAutoString impName;
      if (component == nsnull &&
          ComponentIsImported(compName, impName, importedRoot))
      {
        deeperImported.Put(compName, new nsString(impName));
        continue;
      }

      if (component == nsnull)
      {
        nsCAutoString ccompName;
        CopyUTF16toUTF8(compName, ccompName);
        return SetEarlyError(new mozCellMLLoadError
                             (NS_LITERAL_STRING("missingComponent"),
                              ccompName));
      }
      imported.Put(compName, component);

      // Set the component as seen...
#if 0 // The API isn't stable. For now, trust authors not to make loops...
      nsCOMPtr<nsIContent> componentNode(do_QueryInterface(component));
      nsresult rv;
      if (componentNode->GetProperty(seenAtom) != nsnull)
        continue;
      componentNode->SetProperty(seenAtom, this, nsnull);
#endif
      // Add it to the queue...
      ComponentStack.AppendObject(component);
      // Install it...
      InstallComponent(component, inp, compName);
    }
  }
  return NS_OK;
}

nsresult
mozCellML::HashAvailableComponents(nsInterfaceHashtable<nsStringHashKey,
                                                        nsIDOMElement>& addTo,
                                   nsIDOMElement* importedRoot)
{
  nsCOMPtr<nsIDOMNodeList> cl;
  importedRoot->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> gotComponents, gotImports;
  FilterNodeListByNameAndTypeNS2(
                                 cl,
                                 getter_AddRefs(gotComponents),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("component")
                                );
  nsCOMPtr<nsISimpleEnumerator> gotComponentsE;
  gotComponents->Enumerate(getter_AddRefs(gotComponentsE));
  PRUint32 i, l;
  gotComponents->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> componentSupports;
    gotComponentsE->GetNext(getter_AddRefs(componentSupports));
    nsCOMPtr<nsIDOMElement> gotComponent(do_QueryInterface(componentSupports));
    nsAutoString compName;
    gotComponent->GetAttribute(NS_LITERAL_STRING("name"), compName);
    addTo.Put(compName, gotComponent);
  }

  // Next, we look at import nodes and the components inside them...
  FilterNodeListByNameAndTypeNS(
                                cl,
                                getter_AddRefs(gotImports),
                                nsIDOMNode::ELEMENT_NODE,
                                CELLML_NS_1_1,
                                NS_LITERAL_STRING("import")
                               );
  nsCOMPtr<nsISimpleEnumerator> gotImportsE;
  gotImports->Enumerate(getter_AddRefs(gotImportsE));
  gotImports->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> importSupports;
    gotImportsE->GetNext(getter_AddRefs(importSupports));
    nsCOMPtr<nsIDOMElement> gotImport(do_QueryInterface(importSupports));
    nsCOMPtr<nsIDOMNodeList> icl;
    gotImport->GetChildNodes(getter_AddRefs(icl));
    FilterNodeListByNameAndTypeNS2(
                                   icl,
                                   getter_AddRefs(gotComponents),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("component")
                                  );
    PRUint32 j, m;
    gotComponents->Enumerate(getter_AddRefs(gotComponentsE));
    gotComponents->GetLength(&m);
    for (j = 0; j < m; j++)
    {
      nsCOMPtr<nsISupports> componentSupports;
      gotComponentsE->GetNext(getter_AddRefs(componentSupports));
      nsCOMPtr<nsIDOMElement> gotComponent(do_QueryInterface(componentSupports));
      nsAutoString compName, realName;
      gotComponent->GetAttribute(NS_LITERAL_STRING("name"), compName);
      addTo.Put(compName, gotComponent);
    }
  }
  return NS_OK;
}

nsresult
mozCellML::AddDirectImports(
                            nsInterfaceHashtable<nsStringHashKey,
                                                 nsIDOMElement>& available,
                            nsInterfaceHashtable<nsStringHashKey,
                                                 nsIDOMElement>& imported,
                            nsIDOMElement* importNode,
                            nsCOMArray<nsIDOMElement>& importStack
                           )
{
  nsCOMPtr<nsIDOMNodeList> cl;
  importNode->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> wantComponents;

  // Get all the CellML components listed inside the import...
  FilterNodeListByNameAndTypeNS(
                                cl,
                                getter_AddRefs(wantComponents),
                                nsIDOMNode::ELEMENT_NODE,
                                CELLML_NS_1_1,
                                NS_LITERAL_STRING("component")
                               );

  PRUint32 i, l;
  nsCOMPtr<nsISimpleEnumerator> wantComponentsE;
  wantComponents->Enumerate(getter_AddRefs(wantComponentsE));
  for (i = 0, wantComponents->GetLength(&l); i < l; i++)
  {
    nsCOMPtr<nsISupports> componentSupports;
    wantComponentsE->GetNext(getter_AddRefs(componentSupports));
    nsCOMPtr<nsIDOMElement> wantComponent(do_QueryInterface(componentSupports));
    nsAutoString wantName, wantActualName, wantImportPrefix;
    wantComponent->GetAttribute(NS_LITERAL_STRING("name"), wantName);
    GetImportNumberPrefix(wantComponent, nsnull, wantImportPrefix);

    wantComponent->GetAttribute(NS_LITERAL_STRING("component_ref"),
                                wantActualName);

    nsCOMPtr<nsIDOMElement> gotComponent;
    available.Get(wantActualName, getter_AddRefs(gotComponent));
    imported.Put(wantActualName, gotComponent);
    if (gotComponent != nsnull)
    {
      nsCOMPtr<nsIDOMNode> gotComponentParent;
      gotComponent->GetParentNode(getter_AddRefs(gotComponentParent));
      nsAutoString pln;
      gotComponentParent->GetLocalName(pln);
      if (pln.Equals(NS_LITERAL_STRING("model")))
        InstallComponent(gotComponent, wantImportPrefix, wantName);
    }
    else
    {
      nsCAutoString cwantActualName, cwantName;
      CopyUTF16toUTF8(wantActualName, cwantActualName);
      CopyUTF16toUTF8(wantName, cwantName);
      return SetEarlyError(new mozCellMLLoadError(
                           NS_LITERAL_STRING("importedMissingComponent"),
                           cwantActualName,
                           cwantName));
    }
    importStack.AppendObject(gotComponent);
  }
  return NS_OK;
}

nsresult
mozCellML::AddRelevantConnections(nsIDOMElement* importedRoot,
                                  nsInterfaceHashtable<nsStringHashKey,
                                                       nsIDOMElement>
                                    &imported,
                                  nsClassHashtable<nsStringHashKey,nsString>&
                                    deeper)
{
  nsCOMPtr<nsIDOMNodeList> cl;
  importedRoot->GetChildNodes(getter_AddRefs(cl));
  nsCOMPtr<nsIArray> connections;
  // Get all the CellML components listed inside the import...
  FilterNodeListByNameAndTypeNS2(
                                 cl,
                                 getter_AddRefs(connections),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("connection")
                                );

  PRUint32 i, l;
  nsCOMPtr<nsISimpleEnumerator> connectionsE;
  connections->GetLength(&l);
  connections->Enumerate(getter_AddRefs(connectionsE));

  for (
       i = 0;
       i < l;
       i++
      )
  {
    nsCOMPtr<nsISupports> connSupports;
    connectionsE->GetNext(getter_AddRefs(connSupports));
    nsCOMPtr<nsIDOMElement> conn(do_QueryInterface(connSupports));

    nsCOMPtr<nsIArray> mapcomps;
    conn->GetChildNodes(getter_AddRefs(cl));
    FilterNodeListByNameAndTypeNS2(
                                   cl,
                                   getter_AddRefs(mapcomps),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("map_components")
                                  );
    PRUint32 nmc;
    mapcomps->GetLength(&nmc);
    if (nmc == 0)
      continue;
    nsCOMPtr<nsIDOMElement> mapel;
    mapcomps->QueryElementAt(0, NS_GET_IID(nsIDOMElement),
                             getter_AddRefs(mapel));
    nsAutoString comp1, comp2;
    mapel->GetAttribute(NS_LITERAL_STRING("component_1"), comp1);
    mapel->GetAttribute(NS_LITERAL_STRING("component_2"), comp2);
    
    nsCOMPtr<nsIDOMElement> importedComponent;
    imported.Get(comp1, getter_AddRefs(importedComponent));
    nsString* realname;
    if (importedComponent == nsnull &&
        (deeper.Get(comp1, &realname), realname == nsnull))
      continue;

    nsAutoString varName;

    importedComponent = nsnull;
    imported.Get(comp2, getter_AddRefs(importedComponent));
    if (importedComponent == nsnull &&
        (deeper.Get(comp2, &realname), realname == nsnull))
      continue;

    allConnections.AppendObject(conn);
  }
  return NS_OK;
}

void
mozCellML::ExtractUnitsFrom(
                            PRUint32 importNo,
                            nsIDOMElement* model
                           )
{
  nsCOMPtr<nsIDOMNodeList> nl;
  nsCOMPtr<nsISimpleEnumerator> e, ec;
  nsCOMPtr<nsIArray> units, components;
  PRUint32 i, j, l, m;

  model->GetChildNodes(getter_AddRefs(nl));
  FilterNodeListByNameAndTypeNS2(
                                 nl, getter_AddRefs(units),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("units")
                                );
  units->GetLength(&l);
  units->Enumerate(getter_AddRefs(e));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> unitsSupports;
    e->GetNext(getter_AddRefs(unitsSupports));
    nsCOMPtr<nsIDOMElement> unitsEl(do_QueryInterface(unitsSupports));
    nsCOMPtr<mozCellMLUnits> u =
      new mozCellMLUnits(importNo, NS_LITERAL_STRING(""), unitsEl);
    unitsList.AppendObject(u);
    nsAutoString fqname;
    u->GetFQName(fqname);
    
#ifdef DEBUG_UNITS
    nsCAutoString cfqname;
    CopyUTF16toUTF8(fqname, cfqname);
    printf("fqname: %s\n", cfqname.get());
#endif

    allUnits.Put(fqname, u);
  }

  // Next we have to try every component...
  FilterNodeListByNameAndTypeNS2(
                                 nl, getter_AddRefs(components),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("component")
                                );
  components->GetLength(&l);
  components->Enumerate(getter_AddRefs(ec));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> compSupports;
    ec->GetNext(getter_AddRefs(compSupports));
    nsCOMPtr<nsIDOMElement> component(do_QueryInterface(compSupports));
    nsAutoString compname;
    nsresult rv = component->GetAttribute(NS_LITERAL_STRING("name"), compname);
    if (rv != NS_OK)
      continue;
    component->GetChildNodes(getter_AddRefs(nl));
    FilterNodeListByNameAndTypeNS2(
                                   nl, getter_AddRefs(units),
                                   nsIDOMNode::ELEMENT_NODE,
                                   CELLML_NS_1_0,
                                   CELLML_NS_1_1,
                                   NS_LITERAL_STRING("units")
                                  );
    units->GetLength(&m);
    units->Enumerate(getter_AddRefs(e));
    for (j = 0; j < m; j++)
    {
      nsCOMPtr<nsISupports> unitsSupports;
      e->GetNext(getter_AddRefs(unitsSupports));
      nsCOMPtr<nsIDOMElement> unitsEl(do_QueryInterface(unitsSupports));
      nsCOMPtr<mozCellMLUnits> u =
        new mozCellMLUnits(importNo, compname, unitsEl);
      unitsList.AppendObject(u);
      nsAutoString fqname;
      u->GetFQName(fqname);

#ifdef DEBUG_UNITS
      nsCAutoString cfqname;
      CopyUTF16toUTF8(fqname, cfqname);
      printf("fqname: %s\n", cfqname.get());
#endif

      allUnits.Put(fqname, u);
    }
  }
}

nsresult
mozCellML::LinkImportedUnits(PRUint32 impNo,
                             nsIDOMElement* impReq)
{
  nsresult rv;
  nsCOMPtr<nsIDOMNodeList> nl;
  impReq->GetChildNodes(getter_AddRefs(nl));
  nsCOMPtr<nsIArray> unitsList;
  FilterNodeListByNameAndTypeNS(
                                nl, getter_AddRefs(unitsList), nsIDOMNode::ELEMENT_NODE,
                                CELLML_NS_1_1,
                                NS_LITERAL_STRING("units")
                               );
  nsCOMPtr<nsISimpleEnumerator> e;
  unitsList->Enumerate(getter_AddRefs(e));
  PRUint32 i, l;
  unitsList->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> unitsSupports;
    e->GetNext(getter_AddRefs(unitsSupports));
    nsCOMPtr<nsIDOMElement> unitsEl(do_QueryInterface(unitsSupports));
    nsAutoString localName, remoteName;
    rv = unitsEl->GetAttribute(NS_LITERAL_STRING("name"), localName);
    if (rv != NS_OK)
      return
        SetEarlyError(new mozCellMLLoadError(
                            NS_LITERAL_STRING("nameAttributeMandatory")));
    rv = unitsEl->GetAttribute(NS_LITERAL_STRING("units_ref"), remoteName);
    if (rv != NS_OK)
      return SetEarlyError(new mozCellMLLoadError(
                                NS_LITERAL_STRING("unitsRefAttributeMandatory")));
    nsPrintfCString ir("%u::", impNo);
    nsAutoString compremote;
    CopyUTF8toUTF16(ir, compremote);
    compremote += remoteName;
    // Try to fetch the unit...
    nsCOMPtr<mozCellMLUnits> remUnit;
    allUnits.Get(compremote, getter_AddRefs(remUnit));
    if (remUnit == nsnull)
    {
      nsCAutoString uname;
      CopyUTF16toUTF8(remoteName, uname);
      return SetEarlyError(new mozCellMLLoadError(
                                        NS_LITERAL_STRING("importedUnitNotFound"),
                                        uname));
    }
    nsCOMPtr<nsIDOMNode> impReqPar;
    impReq->GetParentNode(getter_AddRefs(impReqPar));
    nsCOMPtr<nsIDOMElement> impReqParEl(do_QueryInterface(impReqPar));
    PRUint32 reqImpNo;
    importNumbers.Get(impReqParEl, &reqImpNo);
    nsPrintfCString il("%u::", reqImpNo);
    nsAutoString complocal;
    CopyUTF8toUTF16(il, complocal);
    complocal += localName;
    allUnits.Put(complocal, remUnit);
  }
  return NS_OK;
}

void
mozCellML::ImportAll(nsIDOMElement* root)
{
  nsCOMPtr<nsIDOMNodeList> nl;
  root->GetChildNodes(getter_AddRefs(nl));
  nsCOMPtr<nsIArray> imports;
  FilterNodeListByNameAndTypeNS(
                                nl, getter_AddRefs(imports),
                                nsIDOMNode::ELEMENT_NODE,
                                CELLML_NS_1_1,
                                NS_LITERAL_STRING("import")
                               );
  PRUint32 i, l;
  imports->GetLength(&l);
  nsCOMPtr<nsISimpleEnumerator> e;
  imports->Enumerate(getter_AddRefs(e));
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> impSupports;
    e->GetNext(getter_AddRefs(impSupports));
    nsCOMPtr<nsIDOMElement> impEl(do_QueryInterface(impSupports));
    RequestLoadImport(impEl);
  }
}

void
mozCellML::AddComponentAliases(nsIDOMElement* importNode,
                               nsAString& destINP)
{
  nsAutoString sourceINP;
  nsCOMPtr<nsIDOMNodeList> nl;

  GetImportNumberPrefix(importNode, nsnull, sourceINP);

  importNode->GetChildNodes(getter_AddRefs(nl));
  nsCOMPtr<nsIArray> componentList;
  FilterNodeListByNameAndTypeNS2(
                                 nl, getter_AddRefs(componentList),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("component")
                                );
  nsCOMPtr<nsISimpleEnumerator> componentListE;
  componentList->Enumerate(getter_AddRefs(componentListE));
  PRUint32 i, l;
  componentList->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<nsISupports> compSupports;
    componentListE->GetNext(getter_AddRefs(compSupports));
    nsCOMPtr<nsIDOMElement> compEl(do_QueryInterface(compSupports));

    nsAutoString componentName, realName;
    compEl->GetAttribute(NS_LITERAL_STRING("name"), componentName);
    compEl->GetAttribute(NS_LITERAL_STRING("component_ref"), realName);

    ComponentAlias* ca = new ComponentAlias;
    ca->importPrefix.Assign(destINP);
    ca->componentName.Assign(realName);
    
#ifdef DEBUG_MODE
    nsCAutoString cinp1, cname1, cinp2, cname2;
    CopyUTF16toUTF8(destINP, cinp1);
    CopyUTF16toUTF8(realName, cname1);
    CopyUTF16toUTF8(sourceINP, cinp2);
    CopyUTF16toUTF8(componentName, cname2);

    printf("Adding alias mapping... %s in import %s => %s in import %s\n",
           cname2.get(), cinp2.get(), cname1.get(), cinp1.get());
#endif

    componentAliases.Put(sourceINP + componentName, ca);
  }
}

nsresult
mozCellML::ImportLoaded(nsIDOMElement* importNode,
                        nsIDOMElement* importedRoot)
{
  nsresult rv;

  if (importedRoot == nsnull)
  {
#ifdef DEBUG_MODE
    printf("Imported root element is null!\n");
#endif
    return NS_ERROR_FAILURE;
  }

  /* Set up the model */
  importedRootList.AppendObject(importedRoot);
#ifdef DEBUG_MODE
  printf("ImportLoaded called.\n");
#endif
  
  nsAutoString url;
  importNode->GetAttributeNS(
                             NS_LITERAL_STRING("http://www.w3.org/1999/xlink"),
                             NS_LITERAL_STRING("href"),
                             url
                            );

  importedRoot->SetAttributeNS(
                               NS_LITERAL_STRING("http://www.w3.org/1999/xlink"),
                               NS_LITERAL_STRING("href"),
                               url
                              );

  // Assign a new import number for the hashtables...
  importURLTable.AppendString(url);
  importNumbers.Put(importedRoot, nextHashNumber++);

  nsAutoString myINP;
  GetImportNumberPrefix(importedRoot, nsnull, myINP);
  AddComponentAliases(importNode, myINP);

  // Search for units...
  ExtractUnitsFrom(nextHashNumber - 1, importedRoot);

  // Search for imported units. We always load models providing units
  // irregardless of whether or not we need that unit...
  ImportAll(importedRoot);

  // Link the imported units to the model importing them...
  rv = LinkImportedUnits(nextHashNumber - 1, importNode);
  NS_ENSURE_SUCCESS(rv, rv);

  // Build a hashtable of all available components...
  nsInterfaceHashtable<nsStringHashKey, nsIDOMElement> AllComponents;
  nsInterfaceHashtable<nsStringHashKey,nsIDOMElement> ImportComponents;
  AllComponents.Init();
  ImportComponents.Init();
  HashAvailableComponents(AllComponents, importedRoot);

  nsCOMArray<nsIDOMElement> ComponentStack;

  // Seed the list of components we want with those requested directly, and
  // import the componets on the fly...
  rv = AddDirectImports(AllComponents, ImportComponents, importNode, ComponentStack);
  NS_ENSURE_SUCCESS(rv, rv);

  // Now, we expand the list to include encapsulated components.
  nsClassHashtable<nsStringHashKey,nsString> DeeperImported;
  DeeperImported.Init();

  nsAutoString inp;
  GetImportNumberPrefix(importNode, nsnull, inp);

  ExpandEncapsulatedList(inp, ComponentStack, AllComponents, ImportComponents,
                         DeeperImported, importedRoot);

  // We also need to add the connections involving the imported components...
  AddRelevantConnections(importedRoot, ImportComponents, DeeperImported);

  return NS_OK;
}

void
mozCellML::GetImportNumberPrefix(nsIDOMElement* el, PRUint32* asnumber, nsAString& inp)
{
  nsCOMPtr<nsIDOMNode> el1 = el, el2;
  PRUint32 inint;
  do
  {
    if (importNumbers.Get(el1, &inint) == PR_TRUE)
    {
#ifdef DEBUG_IMPORTS
      printf("Import number is %u\n", inint);
#endif
      if (asnumber != nsnull)
        *asnumber = inint;
      nsPrintfCString pc("%u", inint);
      CopyUTF8toUTF16(pc, inp);
      inp.Append(NS_LITERAL_STRING(":"));
      return;
    }

    /* We need to use two nsCOMPtr<nsIDOMNode> here, otherwise this crashes
     * on Win32...
     */
    if (el1->GetParentNode(getter_AddRefs(el2)) != NS_OK ||
        el2 == nsnull)
    {
#ifdef DEBUG_IMPORTS
      printf("GetImportNumberPrefix couldn't find root(assuming doc 0).\n");
#endif
      inint = 0;
      if (asnumber != nsnull)
        *asnumber = 0;
      nsPrintfCString pc("%u", inint);
      CopyUTF8toUTF16(pc, inp);
      inp.Append(NS_LITERAL_STRING(":"));
      return;
    }
    el1 = el2;
  }
  while (PR_TRUE);
}

nsresult
mozCellML::InstallComponent(nsIDOMElement* componentEl,
                            nsAString& importPrefix,
                            nsAString& componentName)
{
  allComponents->AppendElement(componentEl, PR_FALSE);
  nsCOMPtr<nsIDOMNodeList> vl;
  componentEl->GetChildNodes(getter_AddRefs(vl));
  nsCOMPtr<nsIArray> variables;
  nsCOMPtr<nsISimpleEnumerator> variablesE;

  nsAutoString inp;
  GetImportNumberPrefix(componentEl, nsnull, inp);

  nsAutoString realName;
  componentEl->GetAttribute(NS_LITERAL_STRING("name"), realName);

  importPrefix = inp;
  componentName = realName;

#ifdef DEBUG_MODE
  nsCAutoString ccomponentName;
  CopyUTF16toUTF8(componentName, ccomponentName);
  printf("Installing component %s\n", ccomponentName.get());
#endif

  FilterNodeListByNameAndTypeNS2(
                                 vl,
                                 getter_AddRefs(variables),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("variable")
                                );

  size_t j, n;
  variables->GetLength(&n);
  variables->Enumerate(getter_AddRefs(variablesE));
  
  for (
       j = 0;
       j < n;
       j++
      )
  {
    // They will own us, because we must outlive them to protect doc...
    nsCOMPtr<nsISupports> variableSupports;
    variablesE->GetNext(getter_AddRefs(variableSupports));
    nsCOMPtr<nsIDOMElement> variableEl(do_QueryInterface(variableSupports));
    nsCOMPtr<mozCellMLVariable> v = new mozCellMLVariable(variableEl, this);
    PRBool isflagged;
    v->IsFlagged(mozICellML::NO_INTERFACE_IN, &isflagged);
    nsXPIDLString name;
    v->GetName(getter_Copies(name));
    nsAutoString fqname(importPrefix + componentName +
                        NS_LITERAL_STRING(":") + name);
    if (!isflagged)
    {
      nonMainVariables.Put(fqname, v);
      nonMainList.AppendObject(v);
      continue;
    }
    variableList->AppendElement(v, PR_FALSE);
    allVariables.Put(fqname, v);

    nsCOMPtr<mozCellMLVariableResolution> res
      = new mozCellMLVariableResolution(v, v);
    allResolutions.AppendObject(res);

    allVariableRes.Put(fqname, res);
  }
  return NS_OK;
}

nsresult
mozCellML::FindVariableElements(nsIDOMElement* root)
{
  nsCOMPtr<nsIDOMNodeList> dnl;
  root->GetChildNodes(getter_AddRefs(dnl));
  nsCOMPtr<nsIArray> components;
  nsCOMPtr<nsISimpleEnumerator> componentsE;
  FilterNodeListByNameAndTypeNS2(
                                 dnl,
                                 getter_AddRefs(components),
                                 nsIDOMNode::ELEMENT_NODE,
                                 CELLML_NS_1_0,
                                 CELLML_NS_1_1,
                                 NS_LITERAL_STRING("component")
                                );
  PRUint32 i, l;
  components->GetLength(&l);
  components->Enumerate(getter_AddRefs(componentsE));
  for (
       i = 0;
       i < l;
       i++
      )
  {
    nsCOMPtr<nsISupports> componentSupports;
    componentsE->GetNext(getter_AddRefs(componentSupports));
    nsCOMPtr<nsIDOMElement> componentEl(do_QueryInterface(componentSupports));

    nsAutoString compName, compPrefix;
    InstallComponent(componentEl, compPrefix, compName);
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::GetAllVariables(
                           PRUint32 flags,
                           mozICellMLVariable*** variables,
                           PRUint32* count
                          )
{
  if (docRoot == NULL)
    return NS_ERROR_NOT_INITIALIZED;

  nsCOMArray<mozICellMLVariable> allMatchingVars;

  nsCOMPtr<nsISimpleEnumerator> allEnum;
  variableList->Enumerate(getter_AddRefs(allEnum));
  PRUint32 i, l;
  variableList->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    PRBool isflagged;
    nsCOMPtr<nsISupports> vSupports;
    allEnum->GetNext(getter_AddRefs(vSupports));
    nsCOMPtr<mozCellMLVariable> v(do_QueryInterface(vSupports));
    v->IsFlagged(flags, &isflagged);
    if (!isflagged)
      continue;
    allMatchingVars.AppendObject(v);
  }

  *variables = (mozICellMLVariable**)nsMemory::Alloc(sizeof(mozICellMLVariable*) *
                                                     allMatchingVars.Count());
  *count = allMatchingVars.Count();

  for (i = 0; i < (PRUint32)allMatchingVars.Count(); i++)
  {
    (*variables)[i] = allMatchingVars[i];
    NS_ADDREF((*variables)[i]);
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::GetVariable(
                       PRUint32 iimport,
                       const PRUnichar* icomponent,
                       const PRUnichar* iname,
                       mozICellMLVariable** var
                      )
{
  if (docRoot == NULL)
    return NS_ERROR_NOT_INITIALIZED;

  nsAutoString name(iname);
  nsAutoString component(icomponent);

  nsCOMPtr<nsISimpleEnumerator> allEnum;
  variableList->Enumerate(getter_AddRefs(allEnum));
  PRUint32 i, l;
  variableList->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsXPIDLString vname;
    nsCOMPtr<nsISupports> vSupports;
    allEnum->GetNext(getter_AddRefs(vSupports));
    nsCOMPtr<mozCellMLVariable> v(do_QueryInterface(vSupports));

    v->GetName(getter_Copies(vname));
    if (!name.Equals(vname))
      continue;

    v->GetComponentName(getter_Copies(vname));
    if (!component.Equals(vname))
      continue;

    PRUint32 vimport;
    v->GetImportNumber(&vimport);
    if (vimport != iimport)
      continue;

    *var = v;
    NS_ADDREF(*var);
    return NS_OK;
  }
  return NS_ERROR_FAILURE;
}

NS_IMETHODIMP
mozCellML::CheckValidity(
                         PRUnichar** validityError
                        )
{
  if (localValidationError)
  {
    *validityError = ToNewUnicode(validationErrorMessage);
    return NS_OK;
  }
  *validityError = ToNewUnicode(NS_LITERAL_STRING(""));
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::SetStringBundle(mozITranslationSet *ibundle)
{
  bundle = ibundle;

  if (bundle == nsnull)
  {
#ifdef DEBUG_MODE
    printf("mozCellML::SetStringBundle: Bundle is null!\n");
#endif
    return NS_ERROR_FAILURE;
  }
  return NS_OK;
}

const char*
mozCellML::Translate(const char* msg)
{
  if (bundle == nsnull)
  {
#ifdef DEBUG_MODE
    printf("Warning: mozCellML::Translate called prior to setting bundle.\n");
#endif
    nsCAutoString msgCode(msg);
    ctrans = NS_LITERAL_CSTRING("Error message unavailable(no bundle). Message Code ") + msgCode;
    return ctrans.get();
  }
  nsCAutoString cname(msg);
  PRUnichar* umsg = ToNewUnicode(cname);
  nsXPIDLString wtrans;
  nsresult rv = bundle->GetStringFromName(umsg, getter_Copies(wtrans));
  nsCRT::free(umsg);
  if (rv != NS_OK)
  {
    nsCAutoString msgCode(msg);
    ctrans = NS_LITERAL_CSTRING("Error message unavailable"
                              "(translation missing). Message Code "
                               ) + msgCode;
    return ctrans.get();
  }
  CopyUTF16toUTF8(wtrans, ctrans);
  return ctrans.get();
}

nsresult
mozCellML::ValidationError(const nsAString& msg)
{
  if (localValidationError)
    return NS_ERROR_INVALID_ARG;
  localValidationError = PR_TRUE;
#if 0
  if (!validationErrorMessage.Equals(NS_LITERAL_STRING("")))
    validationErrorMessage.Append(NS_LITERAL_STRING(" and "));
  validationErrorMessage.Append(msg);
#else
  validationErrorMessage.Assign(msg);
#endif
#ifdef DEBUG_MODE
  nsCAutoString ve;
  CopyUTF16toUTF8(msg, ve);
  printf("Got a validation error: %s\n", ve.get());
#endif
  return NS_ERROR_INVALID_ARG;
}

NS_IMETHODIMP
mozCellML::ParseCommandLine(const char* data, char** args, PRUint32 maxargs)
{
  PRUint32 n = 0;
  PRBool quoted = PR_FALSE, escaped = PR_FALSE;
  nsCAutoString tmp;

  maxargs -= 2;

  while (*data != 0)
  {
    if (escaped)
    {
      tmp += *data;
      escaped = PR_FALSE;
      data++;
      continue;
    }
#if 0
    else if (*data == '\\')
    {
      escaped = PR_TRUE;
      data++;
      continue;
    }
#endif
    else if (*data == '\"')
      quoted = !quoted;
    else if (!quoted && *data == ' ')
    {
      if (n < maxargs)
        args[n++] = ToNewCString(tmp);
      tmp = "";
    }
    else
      tmp += *data;
    data++;
  }

  args[n++] = ToNewCString(tmp);
  args[n++] = NULL;

  return NS_OK;
}

NS_IMETHODIMP
mozCellML::FreeCommandLine(char **args)
{
  while (*args)
  {
    PR_Free(*args);
    args++;
  }
  return NS_OK;
}

mozCellMLVariable*
mozCellML::FindVariable(nsAString& inp, nsAString& compName,
                        nsAString& varName)
{
  nsAutoString vr(inp + compName + NS_LITERAL_STRING(":") + varName);
  nsCOMPtr<mozCellMLVariable> var;
  allVariables.Get(vr, getter_AddRefs(var));
  return var;
}

NS_IMETHODIMP
mozCellML::NewIntegrator(
                         const char* cfile,
                         const char* compile,
                         const char* execute,
                         mozICellMLIntegrator** result
                        )
{
#ifdef DEBUG_MODE
  printf("NewIntegrator called.\n");
#endif
  // At this stage, we actually generate code. The first step is to open a
  // C file to write the code to.

  PRFileDesc* prfd = PR_Open(cfile, PR_WRONLY | PR_CREATE_FILE, 0700);
#ifdef DEBUG_MODE
  printf("Creating ServerSocket...\n");
#endif

  PRUint32 i, l;

  if (prfd == NULL)
    return NS_ERROR_FAILURE;

  nsCOMPtr<nsIServerSocket> serverSocket = do_CreateInstance(NS_SERVERSOCKET_CONTRACTID);

  serverSocket->Init(0, PR_TRUE, 1);
  PRInt32 local_port;
  serverSocket->GetPort(&local_port);

#ifdef DEBUG_MODE
  printf("Writing integrator code...\n");
#endif

  nsCStringArray functionStack;

  PR_fprintf(
             prfd,
             "#define ND %u\n"
             "#define NS %u\n"
             "#define NC %u\n\n"
             "#define CONNECT_PORT %u\n\n"
             "#include <DoIntegration.I>\n\n"
             "static void init_helper() {\n"
             "#ifdef DEBUG_MATHEMATICS\n"
             "  disable_debug = 0;\n"
             "#endif\n"
             "}\n",
             differentials.Count() + mHiddenHigher,
             simples.Count(), constants.Count(),
             local_port
            );

  nsCAutoString diffcode, oncesimpcode, repeatsimpcode;
  nsresult ret;

  for (i = 0, l = differentialEvaluations.Count();
       i < l;
       i++)
  {
    ret =
      differentialEvaluations[i]->ProduceCode
      (
       this, allVariableRes, importNumbers, allUnits, diffcode, functionStack
      );
    if (ret != NS_OK)
    {
      PR_Close(prfd);
      return ret;
    }
  }

  for (i = 0, l = computedConstantEvaluations.Count();
       i < l;
       i++)
  {
    ret =
      computedConstantEvaluations[i]->ProduceCode
      (
       this, allVariableRes, importNumbers, allUnits, oncesimpcode,
       functionStack
      );
    if (ret != NS_OK)
    {
      PR_Close(prfd);
      return ret;
    }
  }
  // Next, append initial value computations...
  for (i = 0, l = constants.Count();
       i < l;
       i++)
    constants[i]->WriteRuntimeIVAssignment(oncesimpcode);
  for (i = 0, l = differentials.Count();
       i < l;
       i++)
    differentials[i]->WriteRuntimeIVAssignment(oncesimpcode);

  for (i = 0, l = perStepEvaluations.Count();
       i < l;
       i++)
  {
    ret =
      perStepEvaluations[i]->ProduceCode
      (
       this, allVariableRes, importNumbers, allUnits, repeatsimpcode,
       functionStack
      );
    if (ret != NS_OK)
    {
      PR_Close(prfd);
      return ret;
    }
  }

  for (i = 0; i < (PRUint32)functionStack.Count(); i++)
    PR_Write(prfd, functionStack[i]->get(), functionStack[i]->Length());

  PR_fprintf(
             prfd,
             "static void\n"
             "ComputeDifferentials(double y, const double *x, double *dx)\n"
             "{\n"
            );
  PR_Write(prfd, diffcode.get(), diffcode.Length());
  PR_fprintf(
             prfd,
             "}\n\n"
             "static void\n"
             "EvaluateSimpleOnce(double *x)\n"
             "{\n"
            );
  PR_Write(prfd, oncesimpcode.get(), oncesimpcode.Length());
  PR_fprintf(
             prfd,
             "}\n"
             "static void\n"
             "EvaluateSimple(double y, double *x)\n"
             "{\n"
            );
  PR_Write(prfd, repeatsimpcode.get(), repeatsimpcode.Length());
  PR_fprintf(
             prfd,
             "}\n"
            );
#ifdef DEBUG_MODE
  printf("Closing C file...\n");
#endif
  PR_Close(prfd);

#ifdef DEBUG_MODE
  printf("Parsing compile command line...\n");
#endif
  char* argv[100];
  ParseCommandLine(compile, argv, 100);
  if (argv[0] == NULL)
    return NS_ERROR_FAILURE;
#ifdef DEBUG_MODE
  printf("Creating compiler process...\n");
#endif
  PRProcess* p = PR_CreateProcess(argv[0], argv, NULL, NULL);
  FreeCommandLine(argv);
  if (p == NULL)
    return NS_ERROR_FAILURE;
#ifdef DEBUG_MODE
  printf("Waiting for compilation to complete...\n");
#endif

  PRInt32 excode;
  PR_WaitProcess(p, &excode);
  if (excode != 0)
  {
#ifdef DEBUG_MODE
    printf("Compilation failed.\n");
#endif
    return NS_ERROR_FAILURE;
  }

/* #define RUN_VALGRIND_INTEGRATOR */

#ifdef RUN_VALGRIND_INTEGRATOR
  ParseCommandLine(execute, argv + 1, 100);
  argv[0] = "/home/amil082/usr/bin/valgrind";
  if (argv[1] == NULL)
#else
  ParseCommandLine(execute, argv, 100);
  if (argv[0] == NULL)
#endif
  {
#ifdef DEBUG_MODE
    printf("Bad execute command line.\n");
#endif
    return NS_ERROR_FAILURE;
  }

#ifdef DEBUG_MODE
  printf("Compiled the program. Executing it(%s)...\n", argv[0]);
#endif

#ifndef MANUAL_RUN
  p = PR_CreateProcess(argv[0], argv, NULL, NULL);
#endif
  nsCAutoString exec(argv[0]);

  if (p == NULL)
  {
#ifdef DEBUG_MODE
    printf("Can't create process %s\n", argv[0]);
#endif
    FreeCommandLine(argv);
    return NS_ERROR_FAILURE;
  }
#ifdef RUN_VALGRIND_INTEGRATOR
  FreeCommandLine(argv + 1);
#else
  FreeCommandLine(argv);
#endif

#ifdef DEBUG_MODE
  printf("Detaching from the program...\n");
#endif

#ifndef MANUAL_RUN
  PR_DetachProcess(p);
  fflush(stdout);
#endif

#ifdef DEBUG_MODE
  printf("Creating the integrator...\n");
#endif

  // We now have to create the integrator...
  *result = new mozCellMLIntegrator(
                                    serverSocket,
                                    differentials,
                                    simples,
                                    constants,
                                    exec,
                                    mHiddenHigher
                                   );
  // Result has already been addref'd.
#ifdef DEBUG_MODE
  printf("Everything worked.\n");
#endif

  return NS_OK;
}

NS_IMETHODIMP
mozCellML::NewFitter(
                     const char* cfile,
                     const char* compile,
                     const char* execute,
                     mozICellMLFitter** result
                    )
{
#ifdef DEBUG_MODE
  printf("NewFitter called.\n");
#endif
  // At this stage, we actually generate code. The first step is to open a
  // C file to write the code to.

  PRFileDesc* prfd = PR_Open(cfile, PR_WRONLY | PR_CREATE_FILE, 0700);
#ifdef DEBUG_MODE
  printf("Creating ServerSocket...\n");
#endif

  PRUint32 i, l;

  if (prfd == NULL)
    return NS_ERROR_FAILURE;

  nsCOMPtr<nsIServerSocket> serverSocket = do_CreateInstance(NS_SERVERSOCKET_CONTRACTID);

  serverSocket->Init(0, PR_TRUE, 1);
  PRInt32 local_port;
  serverSocket->GetPort(&local_port);

#ifdef DEBUG_MODE
  printf("Writing integrator code...\n");
#endif

  nsCStringArray functionStack;

  PR_fprintf(
             prfd,
             "#define ND %u\n"
             "#define NS %u\n"
             "#define NC %u\n\n"
             "#define CONNECT_PORT %u\n\n"
             "#include <DoFitting.I>\n\n"
             "static void init_helper() {\n"
             "#ifdef DEBUG_MATHEMATICS\n"
             "  disable_debug = 0;\n"
             "#endif\n"
             "}\n",
             differentials.Count() + mHiddenHigher,
             simples.Count(), constants.Count(),
             local_port
            );

  nsCAutoString diffcode, oncesimpcode, repeatsimpcode;
  nsresult ret;

  for (i = 0, l = differentialEvaluations.Count();
       i < l;
       i++)
  {
    ret =
      differentialEvaluations[i]->ProduceCode
      (
       this, allVariableRes, importNumbers, allUnits, diffcode, functionStack
      );
    if (ret != NS_OK)
    {
      PR_Close(prfd);
      return ret;
    }
  }

  for (i = 0, l = computedConstantEvaluations.Count();
       i < l;
       i++)
  {
    ret =
      computedConstantEvaluations[i]->ProduceCode
      (
       this, allVariableRes, importNumbers, allUnits, oncesimpcode,
       functionStack
      );
    if (ret != NS_OK)
    {
      PR_Close(prfd);
      return ret;
    }
  }

  for (i = 0, l = perStepEvaluations.Count();
       i < l;
       i++)
  {
    ret =
      perStepEvaluations[i]->ProduceCode
      (
       this, allVariableRes, importNumbers, allUnits, repeatsimpcode,
       functionStack
      );
    if (ret != NS_OK)
    {
      PR_Close(prfd);
      return ret;
    }
  }

  for (i = 0; i < (PRUint32)functionStack.Count(); i++)
    PR_Write(prfd, functionStack[i]->get(), functionStack[i]->Length());


  PR_fprintf(
             prfd,
             "static void\n"
             "ComputeDifferentials(double y, const double *x, double *dx)\n"
             "{\n"
            );
  PR_Write(prfd, diffcode.get(), diffcode.Length());
  PR_fprintf(
             prfd,
             "}\n\n"
             "static void\n"
             "EvaluateSimpleOnce(double *x)\n"
             "{\n"
            );
  PR_Write(prfd, oncesimpcode.get(), oncesimpcode.Length());
  PR_fprintf(
             prfd,
             "}\n"
             "static void\n"
             "EvaluateSimple(double y, double *x)\n"
             "{\n"
            );
#ifdef DEBUG_MODE
  printf("Closing C file...\n");
#endif
  PR_Close(prfd);

#ifdef DEBUG_MODE
  printf("Parsing compile command line...\n");
#endif
  char* argv[100];
  ParseCommandLine(compile, argv, 100);
  if (argv[0] == NULL)
    return NS_ERROR_FAILURE;
#ifdef DEBUG_MODE
  printf("Creating compiler process...\n");
#endif
  PRProcess* p = PR_CreateProcess(argv[0], argv, NULL, NULL);
  FreeCommandLine(argv);
  if (p == NULL)
    return NS_ERROR_FAILURE;
#ifdef DEBUG_MODE
  printf("Waiting for compilation to complete...\n");
#endif

  PRInt32 excode;
  PR_WaitProcess(p, &excode);
  if (excode != 0)
  {
#ifdef DEBUG_MODE
    printf("Compilation failed.\n");
#endif
    return NS_ERROR_FAILURE;
  }
#ifdef DEBUG_MODE
  printf("Compiled the program. Executing it(%s)...\n", execute);
#endif

  ParseCommandLine(execute, argv, 100);
  if (argv[0] == NULL)
  {
#ifdef DEBUG_MODE
    printf("Bad execute command line.\n");
#endif
    return NS_ERROR_FAILURE;
  }

#ifndef MANUAL_RUN
  p = PR_CreateProcess(argv[0], argv, nsnull, nsnull);
#endif
  nsCAutoString exec(argv[0]);

  if (p == nsnull)
  {
#ifdef DEBUG_MODE
    printf("Can't create process %s\n", argv[0]);
#endif
    FreeCommandLine(argv);
    return NS_ERROR_FAILURE;
  }
  FreeCommandLine(argv);

#ifdef DEBUG_MODE
  printf("Detaching from the program...\n");
#endif

#ifndef MANUAL_RUN
  PR_DetachProcess(p);
#endif

#ifdef DEBUG_MODE
  printf("Creating the fitter...\n");
#endif

  // We now have to create the fitter...
  *result = new mozCellMLFitter(
                                serverSocket,
                                differentials,
                                simples,
                                constants,
                                exec,
                                mHiddenHigher
                               );
#ifdef DEBUG_MODE
  printf("Everything worked.\n");
#endif

  return NS_OK;
}

NS_IMETHODIMP
mozCellML::StartCmgui(CmguiIApplicationBootstrap* appb)
{
  // Find the components directory...
  nsCOMPtr<nsIProperties>
    dsp(do_GetService(NS_DIRECTORY_SERVICE_CONTRACTID));
  nsCOMPtr<nsIFile> comdir;
  nsresult rv =
    dsp->Get("ComsD", NS_GET_IID(nsIFile), getter_AddRefs(comdir));
  if (rv != NS_OK)
  {
    printf("Can't find ComsD.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }
#ifdef XP_WIN
  rv = comdir->Append(NS_LITERAL_STRING("ccbridge.exe"));
#else
  rv = comdir->Append(NS_LITERAL_STRING("ccbridge"));
#endif
  
  PRBool exists;
  comdir->Exists(&exists);
  if (!exists)
  {
    nsresult rv =
      dsp->Get("ProfD", NS_GET_IID(nsIFile), getter_AddRefs(comdir));
    if (rv != NS_OK)
    {
      printf("Can't find ProfD.\n");
      fflush(stdout);
      return NS_ERROR_FAILURE;
    }
    comdir->Append(NS_LITERAL_STRING("extensions"));
    comdir->Append(NS_LITERAL_STRING("{a0d0a9af-bb3d-46c1-a470-8800b0b3f3e2}"));
    comdir->Append(NS_LITERAL_STRING("components"));
#ifdef XP_WIN
    rv = comdir->Append(NS_LITERAL_STRING("ccbridge.exe"));
#else
    rv = comdir->Append(NS_LITERAL_STRING("ccbridge"));
#endif
  }

  nsCAutoString bridgeCPath;
  rv = comdir->GetNativePath(bridgeCPath);
  if (rv != NS_OK)
  {
    printf("Can't get native path.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }

#ifdef DEBUG_MODE
  printf("Creating ORB instance...\n");
  fflush(stdout);
#endif
  // Convert appb to a string...
  nsCOMPtr<CORBA_IORBProxy> orbp =
    do_CreateInstance("@xpcom2corba.sunsite.dk/CORBA/ORBProxy;1");
  if (orbp == nsnull)
  {
    printf("Can't get an ORB.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }

  PRBool success;
  if (orbp->Load_module(
                        "{a0d0a9af-bb3d-46c1-a470-8800b0b3f3e2}",
                        "mozcellml",
                        &success) != NS_OK)
  {
    printf("Warning: Load_module for mozcellml failed.\n");
    fflush(stdout);
  }

  char* objstr;
#ifdef DEBUG_MODE
  printf("Stringfying an object...\n");
  fflush(stdout);
#endif
  rv = orbp->Object_to_string(NS_STATIC_CAST(nsISupports*,appb), &objstr);
  if (rv != NS_OK)
  {
    printf("Object_to_string failed.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }

#ifdef DEBUG_MODE
  printf("Getting bridgeCPath...\n");
  fflush(stdout);
#endif
  const char* bcp = bridgeCPath.get();
  const char* bridge_argv[] =
    {
      /* #define VALGRIND_THE_CCBRIDGE */
#ifdef VALGRIND_THE_CCBRIDGE
      "/home/amil082/usr/bin/valgrind", "--leak-check=full",
#endif
      bcp,
      objstr,
      NULL
    };

#ifdef DEBUG_MODE
#endif

#ifdef DEBUG_MODE_WIN32
  printf("Creating process...\n");
  fflush(stdout);
  char buf[1024];
  sprintf(buf, "c:/cygwin/bin/bash.exe -c '/bin/gdb.exe --args \"%s\" \"%s\"'",
           bcp, objstr);

  PROCESS_INFORMATION pri;
  STARTUPINFO si;
  memset(&si, 0, sizeof(si));
  si.cb = sizeof(si);

  CreateProcess
    (
     "c:/cygwin/bin/bash.exe", buf, NULL, NULL, FALSE,
     CREATE_NEW_CONSOLE | CREATE_NEW_PROCESS_GROUP, NULL, NULL, &si, &pri
    );
#else
  PRProcess* bridgeProcess =
    PR_CreateProcess(
#ifdef DEBUG_MODE
                     "c:/cygwin/bin/gdb.exe",
#elif defined(VALGRIND_THE_CCBRIDGE)
                     "/home/amil082/usr/bin/valgrind",
#else
                     bcp,
#endif
                     (char**)bridge_argv, NULL, NULL);
#ifdef DEBUG_MODE
  printf("Freeing string argument..\n");
  fflush(stdout);
#endif
  nsMemory::Free(objstr);
  if (bridgeProcess == nsnull)
  {
    printf("Can't fork bridge process.\n");
    fflush(stdout);
    return NS_ERROR_FAILURE;
  }
#ifdef DEBUG_MODE
  printf("Detaching the process...\n");
  fflush(stdout);
#endif
  PR_DetachProcess(bridgeProcess);
#endif

  return NS_OK;
}

NS_IMPL_ISUPPORTS1(WeakServerSocketListenerAdaptor, nsIServerSocketListener)
NS_IMPL_ISUPPORTS1(WeakInputCallbackAdaptor, nsIInputStreamCallback)
NS_IMPL_ISUPPORTS1(WeakOutputCallbackAdaptor, nsIOutputStreamCallback)

NS_IMETHODIMP
WeakServerSocketListenerAdaptor::OnSocketAccepted(nsIServerSocket* aServ,
                                                  nsISocketTransport* aTransport)
{
  nsCOMPtr<nsIServerSocketListener> ss(do_QueryReferent(weakSock));
  if (ss != nsnull)
    return ss->OnSocketAccepted(aServ, aTransport);
  return NS_OK;
}

NS_IMETHODIMP
WeakServerSocketListenerAdaptor::OnStopListening(nsIServerSocket* aServ,
                                                 nsresult aStatus)
{
  nsCOMPtr<nsIServerSocketListener> ss(do_QueryReferent(weakSock));
  if (ss != nsnull)
    return ss->OnStopListening(aServ, aStatus);
  return NS_OK;
}

NS_IMETHODIMP
WeakInputCallbackAdaptor::OnInputStreamReady(nsIAsyncInputStream* aStrm)
{
  nsCOMPtr<nsIInputStreamCallback> ss(do_QueryReferent(weakSock));
  if (ss != nsnull)
    return ss->OnInputStreamReady(aStrm);
  return NS_OK;
}

NS_IMETHODIMP
WeakOutputCallbackAdaptor::OnOutputStreamReady(nsIAsyncOutputStream* aStrm)
{
  nsCOMPtr<nsIOutputStreamCallback> ss(do_QueryReferent(weakSock));
  if (ss != nsnull)
    return ss->OnOutputStreamReady(aStrm);
  return NS_OK;
}

NS_IMETHODIMP
mozCellML::UiSafeRun(nsIRunnable* aRunnable)
{
#ifndef XP_WIN
  aRunnable->Run();
#else
  static PRUint32 depth = 0;

  if (corbaThread != nsnull)
  {
    corbaThread->RunOnThread(aRunnable);
  }
  else
  {
    fflush(stderr);
    printf("corba thread is null!\n");
    fflush(stdout);
  }
#endif
  return NS_OK;
}

#ifdef XP_WIN32
NS_IMPL_ISUPPORTS1(mozCellMLCORBAThread, nsIRunnable);

mozCellMLCORBAThread::mozCellMLCORBAThread()
  : mAutoDestruct(false)
{
  mCORBASendSemaphore = CreateSemaphore(NULL, 0, 1, NULL);
  mCORBARecvSemaphore = CreateSemaphore(NULL, 0, 1, NULL);

  // While both semaphores are down, launch the thread...
  nsCOMPtr<nsIThread> thread;
  NS_ADDREF(this);
  NS_NewThread(getter_AddRefs(thread), this);
}

mozCellMLCORBAThread::~mozCellMLCORBAThread()
{
  // This is only supposed to be called *after* the thread has noticed that
  // mAutoDestruct is on...
  CloseHandle(mCORBASendSemaphore);
  CloseHandle(mCORBARecvSemaphore);
}

NS_IMETHODIMP
mozCellMLCORBAThread::Run()
{
  while (true)
  {
    // Wait for the send semaphore to go up...
    WaitForSingleObject(mCORBASendSemaphore, INFINITE);

    // If we are to be destroyed, do it now...
    if (mAutoDestruct)
    {
      mozCellMLCORBAThread* ct = this;
      NS_RELEASE(ct);
      return NS_OK;
    }
    // Do our work...
    mDoRun->Run();

    // Signal back that we are done...
    ReleaseSemaphore(mCORBARecvSemaphore, 1, NULL);
  }
}

void
mozCellMLCORBAThread::RunOnThread(nsIRunnable* aDoRun)
{
  // We can safely touch this until we signal mCORBASendSemaphore(assuming we
  // were called on the main thread).
  mDoRun = aDoRun;
  
  // Raise the send semaphore...
  ReleaseSemaphore(mCORBASendSemaphore, 1, NULL);

  // Wait for the recv semaphore to go up, and process windows messages at
  // the same time...
  while
  (
   MsgWaitForMultipleObjects(1, &mCORBARecvSemaphore, FALSE, INFINITE,
                             QS_SENDMESSAGE | QS_PAINT | QS_POSTMESSAGE) ==
   WAIT_OBJECT_0 + 1
  )
  {
    MSG msg;
    GetMessage(&msg, NULL, 0, 0);
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
}

void
mozCellMLCORBAThread::AutoDestruct()
{
  // We can safely touch this until we signal mCORBASendSemaphore(assuming we
  // were called on the main thread).
  mAutoDestruct = true;
  
  // Raise the send semaphore...
  ReleaseSemaphore(mCORBASendSemaphore, 1, NULL);
}

#endif
