/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "mozCellML.h"
#include "mozCellMLVariable.h"
#include "mozCellMLVariableResolution.h"
#include "MathMLUtilities.h"
#include <nsPrintfCString.h>

#define T(model, s) (model->Translate(s))
#define COMPLAIN(args) \
{\
  char* result = PR_smprintf args; \
  nsCAutoString complainCString(result); \
  nsAutoString complainString; \
  CopyUTF8toUTF16(complainCString, complainString); \
  PR_Free(result); \
  return model->ValidationError(complainString); \
}

mozCellMLEquation::mozCellMLEquation
(
 mozCellML* imodel,
 nsIDOMElement* iapplyNode,
 nsInterfaceHashtable<nsStringHashKey,mozCellMLVariableResolution>& iallVariableRes,
 nsAString& aComponent
)
  : applyNode(iapplyNode), model(imodel), allVariableRes(iallVariableRes),
    mComponent(aComponent)
{
}

NS_IMPL_ISUPPORTS0(mozCellMLEquation)

PRUint32
mozCellMLEquation::CountInvolving(nsCOMArray<mozCellMLVariable>& varInvolving)
{
  PRUint32 i;
  PRUint32 count = 0;
  for (i = 0; i < (PRUint32)varInvolving.Count(); i++)
  {
    // Figure out if v is mentioned here...
    mozCellMLVariable* v = varInvolving[i];
    EquationVariableRelation evr;
    nsCOMPtr<mozCellMLVariableResolution> res;
    GetVariableRelation(v, evr, getter_AddRefs(res));
    if (evr != NOT_MENTIONED && evr != NOT_FREE)
      count++;
  }
  return count;
}

nsresult
mozCellMLEquation::GetIdentifier(nsAString& id)
{
  id = NS_LITERAL_STRING("unidentified");
  applyNode->GetAttribute(NS_LITERAL_STRING("id"), id);
  return NS_OK;
}

mozCellMLVariable*
mozCellMLEquation::GetInvolving(nsCOMArray<mozCellMLVariable>& varInvolving)
{
  PRUint32 i;
  for (i = 0; i < (PRUint32)varInvolving.Count(); i++)
  {
    // Figure out if v is mentioned here...
    mozCellMLVariable* v = varInvolving[i];
    nsCOMPtr<mozCellMLVariableResolution> res;
    EquationVariableRelation evr;
    GetVariableRelation(v, evr, getter_AddRefs(res));
    if (evr != NOT_MENTIONED && evr != NOT_FREE)
      return v;
  }
  return NULL;
}

nsresult
mozCellMLEquation::GetVariableRelation
(
 mozCellMLVariable* var,
 EquationVariableRelation &evr,
 mozCellMLVariableResolution** res
)
{
  // Firstly see if this node is mentioned...
  nsCOMPtr<nsIDOMNodeList> applyChildrenAll;
  nsCOMPtr<nsIArray> applyChildren;
  applyNode->GetChildNodes(getter_AddRefs(applyChildrenAll));
  FilterNodeListByTypeNS(applyChildrenAll, getter_AddRefs(applyChildren),
                         nsIDOMNode::ELEMENT_NODE,
                         MATHML_NS);
  PRUint32 numApplyChildren;
  applyChildren->GetLength(&numApplyChildren);
  if (numApplyChildren != 3)
  {
    COMPLAIN((T(model, "wrongTopLevelApplySubnodes"), numApplyChildren))
  }
  nsCOMPtr<nsISimpleEnumerator> applyChildEnum;
  applyChildren->Enumerate(getter_AddRefs(applyChildEnum));
  nsCOMPtr<nsISupports> opNodeSupports;
  applyChildEnum->GetNext(getter_AddRefs(opNodeSupports));
  nsCOMPtr<nsIDOMNode> opNode(do_QueryInterface(opNodeSupports));

  nsAutoString name;
  opNode->GetLocalName(name);
  if (!name.Equals(NS_LITERAL_STRING("eq")))
  {
    nsCAutoString cname;
    CopyUTF16toUTF8(name, cname);
    COMPLAIN((T(model, "unsupportedTopLevelApply"), cname.get()));
  }

  nsCOMPtr<nsISupports> lhsSupports, rhsSupports;
  applyChildEnum->GetNext(getter_AddRefs(lhsSupports));
  applyChildEnum->GetNext(getter_AddRefs(rhsSupports));
  nsCOMPtr<nsIDOMElement> lhs(do_QueryInterface(lhsSupports));
  nsCOMPtr<nsIDOMElement> rhs(do_QueryInterface(rhsSupports));
  
  // We now have two nodes, left and right.
  VariablePosition lhsPosition, rhsPosition;
  nsCOMPtr<mozCellMLVariableResolution> lres, rres;
  nsresult ret = GetPosition(lhs, var, lhsPosition, getter_AddRefs(lres));
  NS_ENSURE_SUCCESS(ret, ret);

  ret = GetPosition(rhs, var, rhsPosition, getter_AddRefs(rres));
  NS_ENSURE_SUCCESS(ret, ret);

  if (lhsPosition == NOT_INVOLVED || lhsPosition == NOT_FREELY_INVOLVED)
  {
    if (rhsPosition == NOT_INVOLVED || rhsPosition == NOT_FREELY_INVOLVED)
    {
      if ((lhsPosition == NOT_FREELY_INVOLVED ||
           rhsPosition == NOT_FREELY_INVOLVED))
      {
        evr = NOT_FREE;
      }
      else
        evr = NOT_MENTIONED;
    }
    else if (rhsPosition == DIRECT_AS_CI)
      evr = SUBJECT;
    else if (rhsPosition == INVOLVED_IN_EQUATION)
      evr = MIXED;
    else if (rhsPosition == INVOLVED_AS_BOUND)
      evr = BOUND;
    else if (rhsPosition == INVOLVED_AS_DIFFVAR)
      evr = DIFF_VAR;
    *res = rres;
    if (*res != nsnull)
      NS_ADDREF(*res);
    return NS_OK;
  }
  if (rhsPosition == NOT_INVOLVED || rhsPosition == NOT_FREELY_INVOLVED)
  {
    if (lhsPosition == DIRECT_AS_CI)
      evr = SUBJECT;
    else if (lhsPosition == INVOLVED_IN_EQUATION)
      evr = MIXED;
    else if (lhsPosition == INVOLVED_AS_BOUND)
      evr = BOUND;
    else if (lhsPosition == INVOLVED_AS_DIFFVAR)
      evr = DIFF_VAR;
    *res = lres;
    if (*res != nsnull)
      NS_ADDREF(*res);
    return NS_OK;
  }

  // Eliminate diffs on both sides...
  if ((lhsPosition == INVOLVED_AS_BOUND || lhsPosition == INVOLVED_AS_DIFFVAR) &&
      (rhsPosition == INVOLVED_AS_BOUND || rhsPosition == INVOLVED_AS_DIFFVAR))
    COMPLAIN((T(model, "unsupportedDiffBothSides")));

  *res = lres;
  if (*res != nsnull)
    NS_ADDREF(*res);

  // Diffs on one side...
  if (lhsPosition == INVOLVED_AS_BOUND || rhsPosition == INVOLVED_AS_BOUND)
  {
    evr = BOUND;
    return NS_OK;
  }
  if (lhsPosition == INVOLVED_AS_DIFFVAR || rhsPosition == INVOLVED_AS_DIFFVAR)
  {
    evr = DIFF_VAR;
    return NS_OK;
  }

  evr = MIXED;
  return NS_OK;
}

mozCellMLEquation::EquationType
mozCellMLEquation::GetEquationType()
{
  return LookForDiff(applyNode);
}

mozCellMLEquation::EquationType
mozCellMLEquation::LookForDiff(nsIDOMElement* el)
{
  nsCOMPtr<nsIArray> elChildren;
  nsCOMPtr<nsIDOMNodeList> elChildrenAll;
  el->GetChildNodes(getter_AddRefs(elChildrenAll));
  FilterNodeListByTypeNS(elChildrenAll, getter_AddRefs(elChildren),
                         nsIDOMNode::ELEMENT_NODE,
                         MATHML_NS);
  PRUint32 length, i;
  elChildren->GetLength(&length);
  if (length == 0)
    return MATHEMATICAL;
  
  nsCOMPtr<nsISimpleEnumerator> enChildren;
  elChildren->Enumerate(getter_AddRefs(enChildren));
  for (i = 0; i < length; i++)
  {
    nsCOMPtr<nsISupports> childSupports;
    enChildren->GetNext(getter_AddRefs(childSupports));
    nsCOMPtr<nsIDOMElement> childEl(do_QueryInterface(childSupports));
    if (i == 0)
    {
      nsAutoString ln;
      childEl->GetLocalName(ln);
      if (ln.Equals(NS_LITERAL_STRING("diff")))
      {
        // It is a differential. Look for the degree...
        return DIFFERENTIAL;
      }
    }
    if (LookForDiff(childEl) == DIFFERENTIAL)
      return DIFFERENTIAL;
  }
  return MATHEMATICAL;
}

nsresult
mozCellMLEquation::GetPosition(nsIDOMElement* el,
                               mozCellMLVariable* var,
                               VariablePosition& vp,
                               mozCellMLVariableResolution** res)
{
  nsAutoString tn;
  el->GetLocalName(tn);

  if (tn.Equals(NS_LITERAL_STRING("ci")))
  {
    nsCOMPtr<nsIDOMNodeList> subNodes;
    el->GetChildNodes(getter_AddRefs(subNodes));
    nsCOMPtr<nsIArray> subText;
    FilterNodeListByType(subNodes, getter_AddRefs(subText),
                         nsIDOMNode::TEXT_NODE);
    PRUint32 numText;
    subText->GetLength(&numText);
    if (numText == 0)
      return NS_ERROR_FAILURE;
    nsCOMPtr<nsISimpleEnumerator> subTextE;
    subText->Enumerate(getter_AddRefs(subTextE));
    nsCOMPtr<nsISupports> txtNodeSupports;
    subTextE->GetNext(getter_AddRefs(txtNodeSupports));
    nsCOMPtr<nsIDOMText> txt(do_QueryInterface(txtNodeSupports));
    nsAutoString nameStr;
    nsCAutoString nameCStr;
    txt->GetData(nameStr);
    nameStr.Trim(" \t\r\n\f");
    CopyUTF16toUTF8(nameStr, nameCStr);
    // We need to map the name to the variable...
    nsCOMPtr<mozCellMLVariableResolution> vr;
    allVariableRes.Get(mComponent + NS_LITERAL_STRING(":") + nameStr,
                       getter_AddRefs(vr));
    // We have hit an undefined variable...
    if (vr == nsnull)
    {
      nsCAutoString cnameStr, cComponent;
      CopyUTF16toUTF8(nameStr, cnameStr);
      CopyUTF16toUTF8(mComponent, cComponent);
      COMPLAIN((T(model, "variableUsedNotDef"), cnameStr.get(),
                cComponent.get()));
    }
    if (vr->GetVariable() == var)
    {
      vp = DIRECT_AS_CI;
      *res = vr;
      NS_ADDREF(*res);
    }
    else
    {
      vp = NOT_INVOLVED;
      *res = nsnull;
    }
    return NS_OK;
  }
  else if (tn.Equals(NS_LITERAL_STRING("cn")))
  {
    vp = NOT_INVOLVED;
    *res = nsnull;
    return NS_OK;
  }

  if (tn.Equals(NS_LITERAL_STRING("piecewise")))
  {
    nsCOMPtr<nsIDOMNodeList> childrenAll;
    nsCOMPtr<nsIArray> children;
    el->GetChildNodes(getter_AddRefs(childrenAll));
    FilterNodeListByTypeNS(childrenAll, getter_AddRefs(children),
                           nsIDOMNode::ELEMENT_NODE,
                           MATHML_NS);
    PRUint32 numChildren, i;
    children->GetLength(&numChildren);
    nsCOMPtr<nsISimpleEnumerator> enChildren;
    children->Enumerate(getter_AddRefs(enChildren));
    nsCOMPtr<mozCellMLVariableResolution> possibleRes;
    for (i = 0; i < numChildren; i++)
    {
      nsCOMPtr<nsISupports> pSupports;
      enChildren->GetNext(getter_AddRefs(pSupports));
      nsCOMPtr<nsIDOMElement> pEl(do_QueryInterface(pSupports));

      // Check pEl is either piece or otherwise...
      nsAutoString pElName;
      pEl->GetLocalName(pElName);
      if (!pElName.Equals(NS_LITERAL_STRING("piece")) &&
          !pElName.Equals(NS_LITERAL_STRING("otherwise")))
      {
        nsCAutoString cpElName;
        CopyUTF16toUTF8(pElName, cpElName);
        COMPLAIN((T(model, "piecewiseChildNotPieceOrOtherwise"), cpElName.get()))
      }

      // Get the child of pEl...
      pEl->GetChildNodes(getter_AddRefs(childrenAll));
      FilterNodeListByTypeNS(childrenAll, getter_AddRefs(children),
                             nsIDOMNode::ELEMENT_NODE,
                             MATHML_NS);
      PRUint32 numPieceChildren, j;
      children->GetLength(&numPieceChildren);
      nsCOMPtr<nsISimpleEnumerator> enPieceChildren;
      children->Enumerate(getter_AddRefs(enPieceChildren));
      for (j = 0; j < numPieceChildren; j++)
      {
        enPieceChildren->GetNext(getter_AddRefs(pSupports));
        pEl = do_QueryInterface(pSupports);

        nsresult ret = GetPosition(pEl, var, vp, res);
        NS_ENSURE_SUCCESS(ret, ret);
        if (vp == DIRECT_AS_CI)
        {
          vp = INVOLVED_IN_EQUATION;
          return NS_OK;
        }
        if (vp == INVOLVED_IN_EQUATION)
          return NS_OK;
        if (*res != nsnull)
        {
          if (vp == NOT_FREELY_INVOLVED)
            possibleRes = *res;
          NS_RELEASE(*res);
        }
        
        if (vp == INVOLVED_AS_BOUND || vp == INVOLVED_AS_DIFFVAR)
          COMPLAIN((T(model, "mixedDiffsNotSupp")));
      }
    }
    if (possibleRes != nsnull)
    {
      *res = possibleRes;
      vp = NOT_FREELY_INVOLVED;
    }
    else
    {
      *res = nsnull;
      vp = NOT_INVOLVED;
    }
    return NS_OK;
  }

  if (!tn.Equals(NS_LITERAL_STRING("apply")))
  {
    nsCAutoString cname;
    CopyUTF16toUTF8(tn, cname);
    COMPLAIN((T(model, "mathNotApply"), cname.get()))
  }

  // Firstly see if this node is mentioned...
  nsCOMPtr<nsIDOMNodeList> childrenAll;
  nsCOMPtr<nsIArray> children;
  el->GetChildNodes(getter_AddRefs(childrenAll));
  FilterNodeListByTypeNS(childrenAll, getter_AddRefs(children),
                         nsIDOMNode::ELEMENT_NODE,
                         MATHML_NS);
  PRUint32 numChildren, i;
  children->GetLength(&numChildren);

  if (numChildren == 0)
    COMPLAIN((T(model, "applyWrongChildren2"), "undefined", numChildren));

  // Fetch the operator...
  nsCOMPtr<nsISimpleEnumerator> enChildren;
  children->Enumerate(getter_AddRefs(enChildren));
  nsCOMPtr<nsISupports> nSupports;
  enChildren->GetNext(getter_AddRefs(nSupports));
  nsCOMPtr<nsIDOMElement> elOp(do_QueryInterface(nSupports));

  elOp->GetLocalName(tn);
  if (tn.Equals(NS_LITERAL_STRING("diff")) ||
      tn.Equals(NS_LITERAL_STRING("partialdiff")))
  {
    // There should be two more children, or 3 if one is a degree...
    if (numChildren != 3)
    {
      nsCAutoString cname;
      CopyUTF16toUTF8(tn, cname);
      COMPLAIN((T(model, "applyWrongChildren3"), cname.get(), numChildren));
    }
    nsCOMPtr<nsIDOMElement> bvarEl, ciEl, degreeEl;

    enChildren->GetNext(getter_AddRefs(nSupports));
    nsCOMPtr<nsIDOMElement> el1(do_QueryInterface(nSupports));
    enChildren->GetNext(getter_AddRefs(nSupports));
    nsCOMPtr<nsIDOMElement> el2(do_QueryInterface(nSupports));
    
    el1->GetLocalName(tn);
    if (tn.Equals(NS_LITERAL_STRING("bvar")))
    {
      bvarEl = el1;
      ciEl = el2;
    }
    else
    {
      bvarEl = el2;
      ciEl = el1;
    }
    
    // Now reach inside bvarEl to get the degree...
    bvarEl->GetChildNodes(getter_AddRefs(childrenAll));
    FilterNodeListByNameAndTypeNS(childrenAll, getter_AddRefs(children),
                                  nsIDOMNode::ELEMENT_NODE,
                                  MATHML_NS, NS_LITERAL_STRING("degree"));
    childrenAll->GetLength(&numChildren);
    if (numChildren > 0)
    {
      nsCOMPtr<nsISupports> degSupports;
      children->Enumerate(getter_AddRefs(enChildren));
      enChildren->GetNext(getter_AddRefs(degSupports));
      nsCOMPtr<nsIDOMElement> degEl(do_QueryInterface(degSupports));
    }

    // Check that the elements are what we think they are...
    bvarEl->GetLocalName(tn);
    if (!tn.Equals(NS_LITERAL_STRING("bvar")))
      COMPLAIN((T(model, "expectedBvarInDiff")));
    ciEl->GetLocalName(tn);
    if (!tn.Equals(NS_LITERAL_STRING("ci")))
      COMPLAIN((T(model, "expectedCIInDiff")));
    if (degreeEl != nsnull)
    {
      degreeEl->GetLocalName(tn);
      if (!tn.Equals(NS_LITERAL_STRING("degree")))
        COMPLAIN((T(model, "expectedDegreeInDiff")));
    }

    // Now check if we are in the bound variable...
    nsCOMPtr<nsIDOMNodeList> bvarNodesAll;
    bvarEl->GetChildNodes(getter_AddRefs(bvarNodesAll));
    nsCOMPtr<nsIArray> bvarEls;
    FilterNodeListByTypeNS(bvarNodesAll, getter_AddRefs(bvarEls),
                           nsIDOMNode::ELEMENT_NODE,
                           MATHML_NS);
    PRUint32 bvarElsLength;
    bvarEls->GetLength(&bvarElsLength);
    if (bvarElsLength != 1)
      COMPLAIN((T(model, "wrongBvarChildren"), bvarElsLength));
    
    nsCOMPtr<nsIDOMElement> bvarCiEl;
    bvarEls->QueryElementAt(0, NS_GET_IID(nsIDOMElement), getter_AddRefs(bvarCiEl));

    // Make sure it really is a <ci>...
    bvarCiEl->GetLocalName(tn);
    if (!tn.Equals(NS_LITERAL_STRING("ci")))
    {
      nsCAutoString ctn;
      CopyUTF16toUTF8(tn, ctn);
      COMPLAIN((T(model, "bvarChildNotCi"), ctn.get()));
    }

    // We now have two ci elements, see if either match...
    nsresult ret = GetPosition(bvarCiEl, var, vp, res);
    NS_ENSURE_SUCCESS(ret, ret);

    if (vp == DIRECT_AS_CI)
    {
      vp = INVOLVED_AS_BOUND;
      return NS_OK;
    }

    if (*res != nsnull)
    {
      NS_RELEASE(*res);
      *res = nsnull;
    }

    ret = GetPosition(ciEl, var, vp, res);
    NS_ENSURE_SUCCESS(ret, ret);

    if (vp == DIRECT_AS_CI)
    {
      if (degreeEl == nsnull)
        mDegree = 1;
      else
      {
        nsCOMPtr<nsIDOMNodeList> subNodes;
        degreeEl->GetChildNodes(getter_AddRefs(subNodes));
        nsCOMPtr<nsIArray> subText;
        FilterNodeListByType(subNodes, getter_AddRefs(subText),
                             nsIDOMNode::TEXT_NODE);
        PRUint32 numText;
        subText->GetLength(&numText);
        if (numText == 0)
        {
          COMPLAIN((T(model, "degreeWithNoText")))
        }
        nsCOMPtr<nsISimpleEnumerator> subTextE;
        subText->Enumerate(getter_AddRefs(subTextE));
        nsCOMPtr<nsISupports> txtNodeSupports;
        subTextE->GetNext(getter_AddRefs(txtNodeSupports));
        nsCOMPtr<nsIDOMText> txt(do_QueryInterface(txtNodeSupports));
        nsAutoString valueStr;
        nsCAutoString valueCStr;
        txt->GetData(valueStr);
        valueStr.Trim(" \t\r\n\f");
        CopyUTF16toUTF8(valueStr, valueCStr);
        mDegree = strtoul(valueCStr.get(), nsnull, 10);
        if (mDegree == 0 || mDegree > 1000)
        {
          COMPLAIN((T(model, "invalidDegree"), mDegree))
        }
      }
      vp = INVOLVED_AS_DIFFVAR;
    }
    else
    {
      vp = NOT_INVOLVED;
      if (*res != nsnull)
        NS_RELEASE(*res);
      *res = nsnull;
    }

    return NS_OK;
  }

  PRBool isInvolved = PR_FALSE, isBound = PR_FALSE;
  *res = nsnull;
  for (i = 1; i < numChildren; i++)
  {
    enChildren->GetNext(getter_AddRefs(nSupports));
    nsCOMPtr<nsIDOMElement> subel(do_QueryInterface(nSupports));

    subel->GetLocalName(tn);
    // Skip qualifiers...
#define SKIP_QUAL(x) if (tn.Equals(NS_LITERAL_STRING(x))) \
                       continue;

    SKIP_QUAL("interval");
    SKIP_QUAL("condition");
    SKIP_QUAL("domainofapplication");
    SKIP_QUAL("momentabout");
    // Also some constants...
    SKIP_QUAL("pi");
    SKIP_QUAL("exponentiale");
    SKIP_QUAL("eulergamma");
    SKIP_QUAL("infinity");
    SKIP_QUAL("notanumber");
    SKIP_QUAL("true");
    SKIP_QUAL("false");

    // If we find a variable in a bvar for an int, then it is automatically
    // treated as being absent, because it isn't free.
    if (tn.Equals(NS_LITERAL_STRING("bvar")))
    {
      nsCOMPtr<nsIDOMNodeList> bvarNodesAll;
      subel->GetChildNodes(getter_AddRefs(bvarNodesAll));
      nsCOMPtr<nsIArray> bvarEls;
      FilterNodeListByTypeNS(bvarNodesAll, getter_AddRefs(bvarEls),
                             nsIDOMNode::ELEMENT_NODE,
                             MATHML_NS);
      PRUint32 bvarElsLength;
      bvarEls->GetLength(&bvarElsLength);
      if (bvarElsLength != 1)
        COMPLAIN((T(model, "wrongBvarChildren"), bvarElsLength));
      
      nsCOMPtr<nsIDOMElement> bvarCiEl;
      bvarEls->QueryElementAt(0, NS_GET_IID(nsIDOMElement), getter_AddRefs(bvarCiEl));
      
      // Make sure it really is a <ci>...
      bvarCiEl->GetLocalName(tn);
      if (!tn.Equals(NS_LITERAL_STRING("ci")))
      {
        nsCAutoString ctn;
        CopyUTF16toUTF8(tn, ctn);
        COMPLAIN((T(model, "bvarChildNotCi"), ctn.get()));
      }
      mozCellMLVariableResolution *tres2;
      nsresult ret = GetPosition(bvarCiEl, var, vp, &tres2);
      NS_ENSURE_SUCCESS(ret, ret);

      if (tres2 != nsnull)
        NS_RELEASE(tres2);
      
      if (vp == DIRECT_AS_CI)
      {
        nsXPIDLString varname;
        var->GetName(getter_Copies(varname));
        nsCAutoString cname;
        CopyUTF16toUTF8(varname, cname);
        vp = NOT_FREELY_INVOLVED;
        return NS_OK;
      }
      continue;
    }

    // Other qualifiers get unwrapped...
    if (tn.Equals(NS_LITERAL_STRING("lowlimit")) ||
        tn.Equals(NS_LITERAL_STRING("uplimit")) ||
        tn.Equals(NS_LITERAL_STRING("degree")) ||
        tn.Equals(NS_LITERAL_STRING("logbase")))
    {
      nsCOMPtr<nsIDOMNodeList> allQualContents;
      subel->GetChildNodes(getter_AddRefs(allQualContents));
      nsCOMPtr<nsIArray> qualContents;
      FilterNodeListByTypeNS(allQualContents, getter_AddRefs(qualContents),
                             nsIDOMNode::ELEMENT_NODE,
                             MATHML_NS);
      PRUint32 qcl;
      qualContents->GetLength(&qcl);
      if (qcl < 1)
        continue;
      nsCOMPtr<nsISimpleEnumerator> enQualContents;
      qualContents->Enumerate(getter_AddRefs(enQualContents));
      nsCOMPtr<nsISupports> qc1Supports;
      enQualContents->GetNext(getter_AddRefs(qc1Supports));
      subel = do_QueryInterface(qc1Supports);
    }

    mozCellMLVariableResolution *tres;
    nsresult ret = GetPosition(subel, var, vp, &tres);
    NS_ENSURE_SUCCESS(ret, ret);

    if (vp == DIRECT_AS_CI || vp == INVOLVED_IN_EQUATION)
    {
      if (!isInvolved)
        *res = tres;
      else
      {
        NS_RELEASE(tres);
      }
      isInvolved = PR_TRUE;
    }
    else if (vp == INVOLVED_AS_BOUND || vp == INVOLVED_AS_DIFFVAR)
    {
      if (*res != nsnull)
        NS_RELEASE(*res);
      if (tres != nsnull)
        NS_RELEASE(tres);
      COMPLAIN((T(model, "mixedDiffsNotSupp")));
    }
    else
    {
      if (vp == NOT_FREELY_INVOLVED)
        isBound = PR_TRUE;
      if (tres != nsnull)
        NS_RELEASE(tres);
    }
  }

  if (isInvolved)
    vp = INVOLVED_IN_EQUATION;
  else if (isBound)
    vp = NOT_FREELY_INVOLVED;
  else
    vp = NOT_INVOLVED;
  return NS_OK;
}

nsresult
mozCellMLEquation::ProduceEvaluation
(
 mozCellMLVariable* var, mozCellMLEvaluation** eval
)
{
  EquationType et = GetEquationType();

  nsCOMPtr<nsIDOMNode> mathn, componentn;
  applyNode->GetParentNode(getter_AddRefs(mathn));
  mathn->GetParentNode(getter_AddRefs(componentn));
  nsCOMPtr<nsIDOMElement> component(do_QueryInterface(componentn));

  EquationVariableRelation evr;
  nsCOMPtr<mozCellMLVariableResolution> vr;
  GetVariableRelation(var, evr, getter_AddRefs(vr));
  if (vr == nsnull)
    printf("Warning: ProduceEvaluation failed to resolve variable.\n");
  if (evr == NOT_MENTIONED)
    printf("Warning: Asked to produce an evaluation for a variable using an "
           "equation which doesn't mention the variable.\n");

  nsCOMPtr<nsIDOMNodeList> applyChildrenAll;
  nsCOMPtr<nsIArray> applyChildren;
  applyNode->GetChildNodes(getter_AddRefs(applyChildrenAll));
  FilterNodeListByTypeNS(applyChildrenAll, getter_AddRefs(applyChildren),
                         nsIDOMNode::ELEMENT_NODE,
                         MATHML_NS);
  nsCOMPtr<nsISimpleEnumerator> applyChildEnum;
  applyChildren->Enumerate(getter_AddRefs(applyChildEnum));
  nsCOMPtr<nsISupports> opNodeSupports;
  applyChildEnum->GetNext(getter_AddRefs(opNodeSupports));
  nsCOMPtr<nsISupports> lhsSupports, rhsSupports;
  applyChildEnum->GetNext(getter_AddRefs(lhsSupports));
  nsCOMPtr<nsIDOMElement> lhs(do_QueryInterface(lhsSupports));
  applyChildEnum->GetNext(getter_AddRefs(rhsSupports));
  nsCOMPtr<nsIDOMElement> rhs(do_QueryInterface(rhsSupports));
  
  // We now have two nodes, left and right.
  VariablePosition lhsPosition, rhsPosition;
  nsCOMPtr<mozCellMLVariableResolution> tvr;
  nsresult ret = GetPosition(lhs, var, lhsPosition, getter_AddRefs(tvr));
  NS_ENSURE_SUCCESS(ret, ret);
  ret = GetPosition(rhs, var, rhsPosition, getter_AddRefs(tvr));
  NS_ENSURE_SUCCESS(ret, ret);

  if (et == DIFFERENTIAL)
  {
    if (lhsPosition == INVOLVED_AS_DIFFVAR)
    {
      *eval = new mozCellMLEvaluation(vr, rhs,
                                      mozCellMLEvaluation::EVALUATE_AND_ASSIGN_RATE,
                                      component, mDegree);
    }
    else
    {
      *eval = new mozCellMLEvaluation(vr, lhs,
                                      mozCellMLEvaluation::EVALUATE_AND_ASSIGN_RATE,
                                      component, mDegree);
    }
    NS_ADDREF(*eval);
    return NS_OK;
  }
  
  // If it is the subject of the equation, we don't need to perform the
  // Newton-Raphson step.
  if (evr == SUBJECT)
  {
    if (lhsPosition == DIRECT_AS_CI)
      *eval = new mozCellMLEvaluation(vr, rhs,
                                      mozCellMLEvaluation::EVALUATE_AND_ASSIGN,
                                      component);
    else
      *eval = new mozCellMLEvaluation(vr, lhs,
                                      mozCellMLEvaluation::EVALUATE_AND_ASSIGN,
                                      component);
    NS_ADDREF(*eval);
    return NS_OK;
  }

  nsCOMPtr<nsIDOMNode> lhsClone, rhsClone;
  lhs->CloneNode(PR_TRUE, getter_AddRefs(lhsClone));
  rhs->CloneNode(PR_TRUE, getter_AddRefs(rhsClone));
  nsCOMPtr<nsIDOMDocument> ownerDocument;
  lhs->GetOwnerDocument(getter_AddRefs(ownerDocument));
  nsCOMPtr<nsIDOMElement> opEl;
  nsCOMPtr<nsIDOMNode> unusedNode;
  ownerDocument->CreateElementNS(MATHML_NS, NS_LITERAL_STRING("minus"),
                                 getter_AddRefs(opEl));
  nsCOMPtr<nsIDOMElement> innerApplyEl;
  ownerDocument->CreateElementNS(MATHML_NS, NS_LITERAL_STRING("apply"),
                                 getter_AddRefs(innerApplyEl));
  innerApplyEl->AppendChild(opEl, getter_AddRefs(unusedNode));
  innerApplyEl->AppendChild(lhsClone, getter_AddRefs(unusedNode));
  innerApplyEl->AppendChild(rhsClone, getter_AddRefs(unusedNode));
  ownerDocument->CreateElementNS(MATHML_NS, NS_LITERAL_STRING("abs"),
                                 getter_AddRefs(opEl));
  nsCOMPtr<nsIDOMElement> outerApplyEl;
  ownerDocument->CreateElementNS(MATHML_NS, NS_LITERAL_STRING("apply"),
                                 getter_AddRefs(outerApplyEl));
  outerApplyEl->AppendChild(opEl, getter_AddRefs(unusedNode));
  outerApplyEl->AppendChild(innerApplyEl, getter_AddRefs(unusedNode));

  *eval = new mozCellMLEvaluation(vr, outerApplyEl,
                                  mozCellMLEvaluation::MINIMISE_OVER_VARIABLE,
                                  component);
  NS_ADDREF(*eval);
  return NS_OK;
}

mozCellMLEvaluation::mozCellMLEvaluation
(
 mozCellMLVariableResolution* aEvalVar, nsIDOMElement* aMathML,
 EvaluationType aEvalType,
 nsIDOMElement* aComponent,
 PRUint32 aDegree
)
  : mEvalVar(aEvalVar), mMathML(aMathML), mEvalType(aEvalType),
    mComponent(aComponent), mDegree(aDegree)
{
}

NS_IMPL_ISUPPORTS0(mozCellMLEvaluation)

nsresult
mozCellMLEvaluation::ProduceCode
(
 mozCellML* model,
 nsInterfaceHashtable<nsStringHashKey,mozCellMLVariableResolution>&
   allVariables,
 nsDataHashtable<nsISupportsHashKey,PRUint32>&
   importNumbers,
 nsInterfaceHashtable<nsStringHashKey,mozCellMLUnits>&
   allUnits,
 nsACString& code, nsCStringArray& functionStack
)
{
  // Firstly write the maths debug code...
  code += "#ifdef DEBUG_MATHEMATICS\n  active_cmeta_id = \"";
  nsXPIDLString cmeta_id;
  nsresult ret = mEvalVar->GetVariable()->GetName(getter_Copies(cmeta_id));

  if (ret != NS_OK)
    cmeta_id = NS_LITERAL_STRING("unidentified");

  nsCAutoString ccmeta_id;
  CopyUTF16toUTF8(cmeta_id, ccmeta_id);
  code += ccmeta_id;
  code += "\";\n";
  code += "#endif\n";

  code += NS_LITERAL_CSTRING("  ");
  // See what type it is...
  if (mEvalType == EVALUATE_AND_ASSIGN)
  {
      nsCAutoString expressionRHS;
      nsresult ret =
        MathMLExpressionAsString(model, mMathML, allVariables, importNumbers,
                                 allUnits, mComponent, expressionRHS,
                                 functionStack);
      NS_ENSURE_SUCCESS(ret, ret);
      
      // Finally, we need to compute the highest differential. At this point,
      // we may need units conversion...
      mEvalVar->GetAssignCode(
                              mEvalVar->GetVariable()->GetCode(),
                              expressionRHS, code
                             );
      code += "\n";
  }
  else if (mEvalType == EVALUATE_AND_ASSIGN_RATE)
  {
    // See if we have to add extra derivatives...
    if (mDegree > 1)
    {
      PRUint32 i;
      for (i = 0; i < mDegree - 2; i++)
      {
        // No conversions needed, because we do it at the highest differential.
        nsPrintfCString doHigher("dx[%u] = x[%u];\n  ",
                                 mHigherDifferentialSpaceDX + i,
                                 mHigherDifferentialSpaceX + i + 1);
        code += doHigher;
      }
      nsCAutoString vn;
      mEvalVar->GetVariable()->GetVariableRateName(vn);
      code += vn;
      nsPrintfCString lowestRHS(" = x[%u];\n", mHigherDifferentialSpaceX);
      code += lowestRHS;

      nsCAutoString highestRHS;
      nsresult ret =
        MathMLExpressionAsString(model, mMathML, allVariables, importNumbers,
                                 allUnits, mComponent, highestRHS,
                                 functionStack);
      NS_ENSURE_SUCCESS(ret, ret);

      // Finally, we need to compute the highest differential. At this point,
      // we may need units conversion...
      mEvalVar->GetDiffAssignCode(
                                  mHigherDifferentialSpaceDX + mDegree - 1,
                                  highestRHS, code, mDegree
                                 );
      code += "\n";
    }
    else
    {
      nsCAutoString expressionRHS;
      nsresult ret =
        MathMLExpressionAsString(model, mMathML, allVariables, importNumbers,
                                 allUnits, mComponent, expressionRHS,
                                 functionStack);
      NS_ENSURE_SUCCESS(ret, ret);

      // Finally, we need to compute the highest differential. At this point,
      // we may need units conversion...
      mEvalVar->GetDiffAssignCode(
                                  -2 - mEvalVar->GetVariable()->GetCode(),
                                  expressionRHS, code
                                 );
      code += "\n";
    }
  }
  else
  {
    // We now have to minimise a function over a variable to evaluate that
    // variable. Firstly get it as a string...
    nsCAutoString expression("double function");
    PRUint32 solveno = mEvalVar->GetVariable()->GetCode();
    PRUint32 funcno = model->GetFunctionNumber();
    nsPrintfCString funcn("%u", funcno);
    nsPrintfCString solven("%u", solveno);
    expression += funcn;
    expression += "(double y, double* x)\n{\n  return \n";
    nsresult ret =
      MathMLExpressionAsString(model, mMathML, allVariables, importNumbers,
                               allUnits, mComponent, expression,
                               functionStack);
    NS_ENSURE_SUCCESS(ret, ret);

    expression += ";\n}\n";
    functionStack.AppendCString(expression);
    
    code += NS_LITERAL_CSTRING("  solve_by_minimisation(y, x, &function") + funcn +
            NS_LITERAL_CSTRING(", ") + solven;
    code += ");\n";
  }
  return NS_OK;
}

void
mozCellMLEvaluation::AllocateDifferentialSpace
(
 PRUint32& aHigherDifferentialSpaceDX,
 PRUint32& aHigherDifferentialSpaceX
)
{
  mHigherDifferentialSpaceDX = aHigherDifferentialSpaceDX;
  mHigherDifferentialSpaceX = aHigherDifferentialSpaceX;
  aHigherDifferentialSpaceDX += (mDegree - 1);
  aHigherDifferentialSpaceX += (mDegree - 1);
}
