/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <nsInterfaceHashtable.h>
#include <nsCOMArray.h>

class mozCellML;
class mozCellMLVariable;
class mozCellMLVariableResolution;
class nsIDOMElement;
class mozCellMLEvaluation;

class mozCellMLEquation
  : public nsISupports
{
 public:
  NS_DECL_ISUPPORTS

  mozCellMLEquation(mozCellML* imodel, nsIDOMElement* iapplyNode,
                    nsInterfaceHashtable<nsStringHashKey,mozCellMLVariableResolution>& iallVariableRes,
                    nsAString& aComponents);

  enum EquationType
  {
    DIFFERENTIAL, // The equation contains a diff operator.
    MATHEMATICAL  // The equation does not contain a diff operator.
  };

  enum EquationVariableRelation
  {
    NOT_MENTIONED, // The equation does not contain this variable.
    NOT_FREE,      // The equation is mentioned only as a bound variable.
    BOUND,         // The (differential) equation is across this bound variable.
    DIFF_VAR,      // The (differential) equation differentiates this variable.
    SUBJECT,       // This variable is appears by itself on one side of the
                   // equation, is not at all on the other side.
    MIXED          // This variable appears more than once in the equation, or
                   // appears once, but in an expression with other variables.
  };

  enum VariablePosition
  {
    DIRECT_AS_CI,
    INVOLVED_IN_EQUATION,
    INVOLVED_AS_BOUND,
    INVOLVED_AS_DIFFVAR,
    NOT_INVOLVED,
    NOT_FREELY_INVOLVED
  };

  PRUint32 CountInvolving(nsCOMArray<mozCellMLVariable>& varInvolve);

  // Note: Does NOT AddRef the mozCellMLVariable, so do not use once varInvolve
  // goes out of scope, unless you AddRef manually.
  mozCellMLVariable* GetInvolving(nsCOMArray<mozCellMLVariable>& varInvolve);
  nsresult GetVariableRelation(mozCellMLVariable* var,
                               EquationVariableRelation &evr,
                               mozCellMLVariableResolution** res);
  nsresult ProduceEvaluation(mozCellMLVariable* var, mozCellMLEvaluation** eval);
  EquationType GetEquationType();

  nsresult GetIdentifier(nsAString& id);
private:
  PRUint32 mDegree;
  nsCOMPtr<nsIDOMElement> applyNode;
  nsCOMPtr<mozCellML> model;
  nsInterfaceHashtable<nsStringHashKey,mozCellMLVariable> variableList;
  nsInterfaceHashtable<nsStringHashKey,mozCellMLVariableResolution>& allVariableRes;
  nsString mComponent;

  nsresult GetPosition(nsIDOMElement* el, mozCellMLVariable* var,
                       VariablePosition &vp,
                       mozCellMLVariableResolution** res);
  EquationType LookForDiff(nsIDOMElement* el);
};

class mozCellMLEvaluation
  : public nsISupports
{
 public:
  NS_DECL_ISUPPORTS

  enum EvaluationType
  {
    EVALUATE_AND_ASSIGN,
    MINIMISE_OVER_VARIABLE,
    EVALUATE_AND_ASSIGN_RATE
  };

  mozCellMLEvaluation(mozCellMLVariableResolution* aEvalVar, nsIDOMElement* aMathML,
                      EvaluationType aEvalType, nsIDOMElement* component,
                      PRUint32 aDegree = 1);

  /**
   * Generates C code suitable to compute the variable(or in the case of
   * EVALUATE_AND_ASSIGN_RATE, the rate of the variable). When the degree of
   * differentiation is >1, will compute all derivatives.
   */
  nsresult ProduceCode(mozCellML* model,
                       nsInterfaceHashtable<nsStringHashKey,mozCellMLVariableResolution>&
                         allVariables,
                       nsDataHashtable<nsISupportsHashKey,PRUint32>&
                         importNumbers,
                       nsInterfaceHashtable<nsStringHashKey,mozCellMLUnits>&
                         allUnits,
                       nsACString& code, nsCStringArray& functionStack);

  void AllocateDifferentialSpace(PRUint32& aHigherDifferentialSpaceDX,
                                 PRUint32& aHigherDifferentialSpaceX);
private:
  nsCOMPtr<mozCellMLVariableResolution> mEvalVar;
  nsCOMPtr<nsIDOMElement> mMathML;
  EvaluationType mEvalType;
  nsCOMPtr<nsIDOMElement> mComponent;
  PRUint32 mDegree;
  PRUint32 mHigherDifferentialSpaceDX, mHigherDifferentialSpaceX;
};
