/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz> (original developer)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "mozCellMLPointToZinc.h"
#include "../../../../../mozGeneralPlugin/mozIGeneralPlugin.h"

#ifdef XP_WIN
#include <nsIThread.h>
#include "mozWinFixup.h"
#endif

#define BATCH_SIZE 100

nsCOMArray<mozIGeneralPlugin> mozCellMLPointToZinc::plugins;

static double
abs_d(double a1)
{
  if (a1 < 0)
    return -a1;
  return a1;
}

static int
compare_with_tolerance(double a1, double a2)
{
  double nf;
  /* This covers not a number etc... too. */
  if (a1 == a2)
  {
    return 1;
  }
  if (a1 == 0.0 || a2 == 0.0)
    return 0;
  if (a1 < 0.0)
  {
    if (a2 > 0.0)
      return 0;
    a1 = -a1;
    a2 = -a2;
  }
  else if (a2 < 0.0)
    return 0;
  nf = a1;
  if (nf > a2)
    nf = a2;
  nf = abs_d((a1-a2)/nf);
  if (nf < 2E-16)
    return 1;
  return 0;
}

static double
floor_with_tolerance(double v)
{
  double fl;
  fl = (PRUint32)v;
  if (((fl + 1.0) / v) < 1.00000000000001)
  {
    return (fl + 1.0);
  }
  return fl;
}

mozCellMLPointToZinc::mozCellMLPointToZinc
(
 mozICellML* aCellML,
 mozIScaleWatcher* iscaleWatcher,
 double istartx, double istopx, double igridx,
 PRUint32 inumvariables,
 mozICellMLVariable** ivariables,
 CmguiIFunction** ifunctions,
 CmguiICommandData* icmddata,
 CmguiITimeSequencePackage* iseqPackage,
 CmguiINodeFieldCreator* infc,
 CmguiINode* ipointNode
)
  : mCellML(aCellML), startx(istartx), stopx(istopx), gridx(igridx),
    numvariables(inumvariables), unsentPoints(0), scaleWatcher(iscaleWatcher),
    cmddata(icmddata), seqPackage(iseqPackage), nfc(infc),
    pointNode(ipointNode), point(0), mapIntgToUser(nsnull), mapUserToIntg(nsnull),
    data(nsnull), scales(nsnull)
{
  PRUint32 i;
  double v = (stopx - startx) / gridx;
  maxpoint = (PRUint32)floor_with_tolerance(v);
  v = startx + maxpoint * gridx;
  if (compare_with_tolerance(v, stopx))
    maxpoint++;

  for (i = 0; i < inumvariables; i++)
    functions.AppendObject(ifunctions[i]);
  for (i = 0; i < inumvariables; i++)
    variables.AppendObject(ivariables[i]);
}

mozCellMLPointToZinc::~mozCellMLPointToZinc()
{
  if (mapIntgToUser)
    delete [] mapIntgToUser;
  if (mapUserToIntg)
    delete [] mapUserToIntg;
  if (data)
    delete [] data;
  if (scales)
    delete [] scales;
}

NS_IMETHODIMP
mozCellMLPointToZinc::AddPlugin(mozIGeneralPlugin* iplug)
{
  plugins.AppendObject(iplug);
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLPointToZinc::RemovePlugin(mozIGeneralPlugin* iplug)
{
  plugins.RemoveObject(iplug);
  return NS_OK;
}

NS_IMPL_ISUPPORTS5(mozCellMLPointToZinc, mozIPointWatcher,
                   mozIPointToZinc, nsISupportsWeakReference,
                   mozIPointToZincLocal, nsIRunnable)

#define dabs(x) ((x>=0) ? x : -x)

NS_IMETHODIMP
mozCellMLPointToZinc::Run()
{
  nsresult ret;
  if (!unsentPoints)
    return NS_OK;

  nsCOMPtr<CmguiIRegion> region;
  ret = cmddata->GetRootRegion(getter_AddRefs(region));
  NS_ENSURE_SUCCESS(ret, ret);

  ret = region->BeginChange();
  NS_ENSURE_SUCCESS(ret, ret);

  nsCOMPtr<CmguiITimeSequence> tseq;

  ret =
    seqPackage->GetSimpleTimeSequence(unsentPoints,
                                      startx + (point - unsentPoints) * gridx,
                                      gridx,
                                      getter_AddRefs(tseq));
  NS_ENSURE_SUCCESS(ret, ret);

  PRUint32 i;
  for (i = 0; i < numvariables; i++)
  {
    ret = functions[i]->DefineOnNode(pointNode, tseq, nfc);
    NS_ENSURE_SUCCESS(ret, ret);

    nsCOMPtr<CmguiIFunction> matrix;
    ret = cmddata->CreateMatrix
          (
           1, data + i * maxpoint + point - unsentPoints, unsentPoints,
           getter_AddRefs(matrix)
          );
    NS_ENSURE_SUCCESS(ret, ret);

    nsCOMPtr<CmguiIFunctionVariable> funcVar;
    ret = functions[i]->GetNodalValues
          (
           "", 0, pointNode,
           CmguiIFunction::NODAL_VALUE,
           0, tseq, getter_AddRefs(funcVar)
          );
    NS_ENSURE_SUCCESS(ret, ret);

    ret = funcVar->SetValue(matrix);
    NS_ENSURE_SUCCESS(ret, ret);
  }

  ret = region->EndChange();
  NS_ENSURE_SUCCESS(ret, ret);

  unsentPoints = 0;
  return NS_OK;
}

nsresult
mozCellMLPointToZinc::SendPointsToZinc()
{
  if (!unsentPoints)
    return NS_OK;

  mozICellML* mCellMLRaw = mCellML;
  mozICellMLLocal* mCellMLLocalRaw;
  mCellMLLocalRaw = NS_STATIC_CAST(mozICellMLLocal*, mCellMLRaw);
  
  return mCellMLLocalRaw->UiSafeRun(this);
}

NS_IMETHODIMP
mozCellMLPointToZinc::PointAdded(double x, mozICellMLVariable** variable,
                                 PRUint32 numvariables,double* values,
                                 PRUint32 numvariables_2)
{
  PRUint32 i, j;
  PRUint32 lpoint = point;
  if (lpoint >= maxpoint)
    return NS_ERROR_FAILURE;
  point++;

  if (!mapIntgToUser)
  {
    mapIntgToUser = new PRUint32[numvariables];
    if (!mapIntgToUser)
      return NS_ERROR_OUT_OF_MEMORY;
    mapUserToIntg = new PRUint32[numvariables];
    if (!mapUserToIntg)
      return NS_ERROR_OUT_OF_MEMORY;
    for (i = 0; i < numvariables; i++)
      for (j = 0; j < numvariables; j++)
      {
        PRBool b;
        variable[i]->Equals(variables[j], &b);
        if (b)
        {
          mapIntgToUser[i] = j;
          mapUserToIntg[j] = i;
        }
      }
  }
  if (!data)
  {
    data = new double[numvariables * maxpoint];
    if (!data)
      return NS_ERROR_OUT_OF_MEMORY;
  }
  if (!scales)
  {
    scales = new double[numvariables];
    if (!scales)
      return NS_ERROR_OUT_OF_MEMORY;
    memset(scales, 0, sizeof(double) * numvariables);
  }

  for (i = 0; i < numvariables; i++)
  {
    data[mapIntgToUser[i] * maxpoint + lpoint] = values[i];
    if (scales[mapIntgToUser[i]] < dabs(values[i]))
    {
      scales[mapIntgToUser[i]] = dabs(values[i]) * 1.5;
      scaleWatcher->ScaleChanged(mapIntgToUser[i], scales[mapIntgToUser[i]]);
    }
  }

  if (point == maxpoint)
  {
    scaleWatcher->Progress(x);
  }
  else if (lpoint % BATCH_SIZE == 0)
    scaleWatcher->Progress(x);

  /* Finally, inform Zinc... */
  if (unsentPoints++ > BATCH_SIZE)
    SendPointsToZinc();

  return NS_OK;
}

NS_IMETHODIMP
mozCellMLPointToZinc::Done()
{
  if (unsentPoints > 0)
    SendPointsToZinc();
  scaleWatcher->Done();
  scaleWatcher = nsnull;
  mCellML = nsnull;
  cmddata = nsnull;
  seqPackage = nsnull;
  nfc = nsnull;
  pointNode = nsnull;
  variables.Clear();
  functions.Clear();
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLPointToZinc::Halted()
{
  if (unsentPoints > 0)
    SendPointsToZinc();
  scaleWatcher->Halted();
  scaleWatcher = nsnull;
  mCellML = nsnull;
  cmddata = nsnull;
  seqPackage = nsnull;
  nfc = nsnull;
  pointNode = nsnull;
  variables.Clear();
  functions.Clear();
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLPointToZinc::Failed(const char* emsg)
{
  if (unsentPoints > 0)
    SendPointsToZinc();
  scaleWatcher->Failed(emsg);
  scaleWatcher = nsnull;
  mCellML = nsnull;
  cmddata = nsnull;
  seqPackage = nsnull;
  nfc = nsnull;
  pointNode = nsnull;
  variables.Clear();
  functions.Clear();
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLPointToZinc::GetProgress(double* aProgress)
{
  if (point >= maxpoint)
    *aProgress = 100.0;
  else
    *aProgress = (point * 100.0 / maxpoint);
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLPointToZinc::GetValue(PRUint32 variable, PRUint32 timepoint,
                               double* result)
{
  if (variable >= numvariables ||
      timepoint >= point ||
      data == nsnull)
    return NS_ERROR_FAILURE;
  *result = data[variable * maxpoint + timepoint];
  return NS_OK;
}
